import sys
import time
import os
import optparse
import logging

# hacky sys.path maipulations when running from source-code.
if not hasattr(sys, "frozen"):
    try:
        import tbzrlib
    except ImportError:
        # Need to put my parent dir on sys.path.
        sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from bzrlib.trace import mutter, note, push_log_file
from tbzrlib.app import get_app, StartupError

is_windows = sys.platform.startswith("win")

if is_windows:
    import win32api, win32con
    have_console = False
    try:
        win32api.GetConsoleTitle()
        have_console = True
    except win32api.error, exc:
        have_console = exc[0] == 0

if is_windows and not have_console:
    if getattr(sys, "frozen", None) == "windows_exe":
        # boot_common.py sets sys.stdout to a "Blackhole" object.  These
        # assertions are to catch when py2exe changes its implementation, so
        # we can see if this still makes sense.
        assert sys.stderr.__class__.__name__ == "Stderr", sys.stderr
        assert sys.stdout.__class__.__name__ == "Blackhole", sys.stdout
        # It needs more thought if we really do want to throw this away...
        sys.stderr = sys.stdout
    else:
        class _NullStream(object):
            def write(self, x): pass
            def flush(self): pass
            def read(self, x, size=0): return ""
            def close(self): pass
        sys.stderr = sys.stdout = _NullStream()

def console_ctrl_handler(evt):
    # NOTE that this is called on its own thread making it difficult to do
    # things like break into pdb etc.
    stop_reason = None
    # False means "run the next/default handler", True means don't call others.
    rc = False
    # Leave Ctrl+Break alone so it can kill the app in case things are serious!
    if evt in (win32con.CTRL_C_EVENT,): # win32con.CTRL_BREAK_EVENT):
        stop_reason = "Ctrl+C"
        rc = True
    # During a log-off etc process the console is also closed, so this works
    # OK here.
    elif evt == win32con.CTRL_CLOSE_EVENT:
        stop_reason = "Console closing"
        rc = True

    if stop_reason is not None:
        note("%s - stopping application...", stop_reason)
        get_app().stop()

    return rc

def set_signal_handlers():
    # special handling for Ctrl+C etc to cleanly shutdown.

    # We gave consideration to hooking standard bzr ctrl+break break behaviour,
    # via bzrlib.breakin, but that isn't useful as we are called on a new
    # thread, so pdb has absolutely nothing of interest...
    if is_windows:
        win32api.SetConsoleCtrlHandler(console_ctrl_handler, True)

def reset_signal_handlers():
    if is_windows:
        win32api.SetConsoleCtrlHandler(console_ctrl_handler, False)

def setup_logging(options):
    # Note that by using the bzr log, we will prevent bzr from being able to
    # roll the log file over.  So we probably need to incorporate that into
    # our idle loop somehow.
    from bzrlib.trace import _open_bzr_log, enable_default_logging

    # bzr will add a logging handler that writes to stderr - and py2exe
    # assumes that a GUI program which has written to sys.stderr is in error,
    # so writes a log and displays a message box on exit.  See boot_common.py
    # in py2exe for more.
    if getattr(sys, "frozen", None) == "windows_exe":
        # boot_common.py sets sys.stdout to a "Blackhole" object.  These
        # assertions are to catch when py2exe changes its implementation, so
        # we can see if this still makes sense.
        assert sys.stderr.__class__.__name__ == "Stderr", sys.stderr
        assert sys.stdout.__class__.__name__ == "Blackhole", sys.stdout
        # It needs more thought if we really do want to throw this away...
        sys.stderr = sys.stdout

    enable_default_logging()
    bzr_log_file = _open_bzr_log()
    bzr_log = logging.getLogger("bzr")

    # convert a level name to a value.
    try:
        val = int(getattr(logging, options.log_level.upper()))
    except (ValueError, AttributeError):
        # not a level name from the logging module - maybe a literal.
        try:
            val = int(options.log_level)
        except ValueError:
            bzr_log.error("Invalid log-level '%s' ignored" % (options.log_level,))
            val = None

    if val is not None:
        bzr_log.setLevel(val)

    if options.verbose:
        # treat 'mutter' calls as DEBUG log messages.
        push_log_file(bzr_log_file)

        # If we have a windows console, send output there too.
        if is_windows and have_console:
            push_log_file(sys.stdout)

def parse_options():
    parser = optparse.OptionParser()
    parser.add_option("-v", "--verbose", action="store_true")
    parser.add_option("", "--log-level", action="store",
                      default=logging.INFO)
    parser.add_option("", "--show-config", action="store_true",
                      help="Dump all available configuration options and "
                           "values")
    parser.add_option("-r", "--replace", action="store_true",
                      help="Attempt to replace (ie, close) an existing cache "
                           "process before starting")

    (options, args) = parser.parse_args()
    return parser, options, args

def selftest(options, args):
    from bzrlib.tests import run_suite
    from bzrlib.tests.TestUtil import TestLoader
    testmod_names = [
        'tbzrlib.tests.test_actions',
        'tbzrlib.tests.test_dispatcher',
        'tbzrlib.tests.test_vcspath',
        'tbzrlib.tests.test_watcher',
        'tbzrlib.tests.cache.test_wtcache',
        'tbzrlib.tests.shellext.test_copyhook',
        'tbzrlib.tests.shellext.test_contextmenu',
        'tbzrlib.tests.shellext.test_iconoverlay',
    ]
    pattern = ".*"
    if args:
        pattern = args[0]

    loader = TestLoader()
    suite = loader.suiteClass()
    suite.addTest(loader.loadTestsFromModuleNames(testmod_names))
    ok = run_suite(suite, pattern=pattern, verbose=options.verbose)
    # we return the process exit code.
    if not ok:
        return 1
    return 0

def run():
    parser, options, args = parse_options()
    # setup logger early
    setup_logging(options)

    if args:
        if args[0] == "selftest":
            sys.exit(selftest(options, args[1:]))
        else:
            parser.error("Unrecognized arguments '%s'" % " ".join(args))

    # setup a ctrl+c handler.
    set_signal_handlers()
    try:
        try:
            try:
                a = get_app()
                if options.replace:
                    import win32gui, win32con
                    try:
                        hwnd = win32gui.FindWindow("TBZRCache_Taskbar", None)
                    except win32gui.error:
                        # no cache running - that's OK
                        hwnd = 0
                    if not hwnd:
                        print "--replace specified, but can't seem to find an existing process..."
                    else:
                        try:
                            win32gui.PostMessage(hwnd, win32con.WM_QUIT, 0, 0)
                            # wait a few seconds for the old app to die.
                            for i in xrange(20):
                                if a.create_app_mutex():
                                    a.hmutex.Close()
                                    a.hmutex = None
                                    print "Existing process found and closed."
                                    break
                                time.sleep(0.1)
                            else:
                                print "Existing process failed to terminate."
                        except win32gui.error, exc:
                            logger.error("Failed to send quit message: %s", exc)

                if options.show_config:
                    print "## Dumping all config options and current values"
                    print "## Config file is:", a.get_config_filename()
                    a.config.dump(sys.stdout)
                    print "## end of config"
                a.run()
            except StartupError, exc:
                msg = exc[0]
                logging.getLogger("bzr.tbzr").error(msg)
                sys.exit(1)
        finally:
            reset_signal_handlers()
    except (SystemExit, KeyboardInterrupt):
        # generally we will never see these exceptions - Ctrl+C handling
        # will prevent KeyboardInterrupt, and nothing else stops the server
        # via SystemExit - but its still good form to let these ones go...
        raise
    except:
        logging.getLogger("bzr.tbzr").exception("TBZRCache process failed")
        sys.exit(1)

if __name__=='__main__':
    run()
    # tbzrcache sometimes hungs while shutting down.
    # So force exit until I get what's the matter.
    os._exit(0)
