# Copyright (C) 2007 by Alexander Haro (amduser) <amduser29@gmail.com>
# Copyright (C) 2008 by Canonical
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
import sys
import os
import optparse
import logging
import traceback
import tbzrlib.i18n as i18n

is_windows = sys.platform.startswith("win")
is_frozen = False

if is_windows:
    import win32api, win32con

    # The py2exe/stderr pain...
    if getattr(sys, "frozen", None) == "windows_exe":
        assert sys.stderr.__class__.__name__ == "Stderr", sys.stderr
        assert sys.stdout.__class__.__name__ == "Blackhole", sys.stdout
        sys.stderr = sys.stdout
        is_frozen = True

    have_console = False
    try:
        win32api.GetConsoleTitle()
        have_console = True
    except win32api.error, exc:
        have_console = exc[0] == 0
        if not is_frozen:
            sys.stdout = sys.stderr = open('nul', 'wb')

from bzrlib.trace import mutter, note, warning, push_log_file
from tbzrcommands import error_dialog

# Almost identical to tbzrcache.py, but no doubt will change...
# (eg, our lifetime is much shorter than tbzrcache.)
def setup_logging(options):
    from bzrlib.trace import _open_bzr_log, enable_default_logging

    enable_default_logging()
    bzr_log_file = _open_bzr_log()
    bzr_log = logging.getLogger("bzr")

    # convert a level name to a value.
    try:
        val = int(getattr(logging, options.log_level.upper()))
    except (ValueError, AttributeError):
        # not a level name from the logging module - maybe a literal.
        try:
            val = int(options.log_level)
        except ValueError:
            bzr_log.error("Invalid log-level '%s' ignored" % (options.log_level,))
            val = None

    if val is not None:
        bzr_log.setLevel(val)

    if options.verbose:
        # treat 'mutter' calls as DEBUG log messages.
        push_log_file(bzr_log_file)

        if is_windows and have_console:
            # If we have a windows console, send output there too.
            push_log_file(sys.stdout)

    # Use a default exception handler so exceptions thrown by Qt events, such
    # as a dialog 'accept' slots, are logged appropriately.
    sys.excepthook = excepthook

def excepthook(t, v, tb):
    logging.getLogger("bzr").error("Unhandled tbzrcommand exception",
                                   exc_info=(t, v, tb))

class TBZRCommandError(Exception):
    pass

class UnrecognizedCommandError(TBZRCommandError):
    pass

def chdir_to_path(path):
    if os.path.isdir(path):
        os.chdir(path)
    else:
        os.chdir(os.path.dirname(path))

QBZR_CMDS = 'bzrlib.plugins.qbzr.lib.commands'

def invoke_about(command, argname, files, directory, kw):
    """Invoke command takes no arguments"""
    # some commands expects current directory is in working tree.
    if not directory and files:
        chdir_to_path(files[0])
    command.run(**kw)

def invoke_add(command, argname, files, directory, kw):
    """Invoke command takes file list as arguments."""
    if files:
        kw[argname] = files
    command.run(**kw)

def invoke_annotate(command, argname, files, directory, kw):
    """Invoke command takes a file name."""
    try:
        kw[argname] = files[0]
        command.run(**kw)
    except Exception:
        pass

def invoke_explorer(command, argname, files, directory, kw):
    #assert argname == 'location'
    if not directory:
        for p in files:
            if os.path.isdir(p):
                directory = p
                break
        else:
            return command.run(**kw)
    kw[argname] = directory
    command.run(**kw)

commands_qt = {
    # cmd-name    : (module_name,
    #                class_name,          filenames_arg,
    #                invoker, refresh cache)

    'about'       : ('tbzrcommands.cmd_about',
                     'cmd_about', None,
                     invoke_about, False),

    'add'         : (QBZR_CMDS,
                     'cmd_qadd', 'selected_list',
                     invoke_add, True),

    'annotate'    : (QBZR_CMDS,
                     'cmd_qannotate', 'filename',
                     invoke_annotate, False),

    'branch'      : (QBZR_CMDS,
                     'cmd_qbranch', None,
                     invoke_about, True),

    'browse'      : (QBZR_CMDS,
                     'cmd_qbrowse', 'location',
                     invoke_about, False),

    'cat'         : (QBZR_CMDS,
                     'cmd_qcat', 'filename',
                     invoke_annotate, False),

    'commit'      : (QBZR_CMDS,
                     'cmd_qcommit', 'selected_list',
                     invoke_add, True),

    'diff'        : (QBZR_CMDS,
                     'cmd_qdiff', 'file_list',
                     invoke_add, False),

    'getnew'      : (QBZR_CMDS,
                     'cmd_qgetnew', 'location',
                     invoke_about, True),

    'getupdates'  :(QBZR_CMDS,
                    'cmd_qgetupdates', 'location',
                    invoke_about, True),

    'push'        : (QBZR_CMDS,
                     'cmd_qpush', 'directory',
                     invoke_about, False),

    'help'        : ('tbzrcommands.cmd_help',
                     'cmd_help', None,
                     invoke_about, False),

    'init'        : (QBZR_CMDS,
                     'cmd_qinit', 'location',
                     invoke_about, True),

    'log'         : (QBZR_CMDS,
                     'cmd_qlog', 'locations_list',
                     invoke_add, False),

    'remove'      : ('tbzrcommands.cmd_remove',
                     'cmd_remove', 'file_list',
                     invoke_add, False),

    'revert'      : (QBZR_CMDS,
                     'cmd_qrevert', 'selected_list',
                     invoke_add, True),

    'settings'    : ('tbzrcommands.cmd_settings',
                     'cmd_settings', None,
                     invoke_about, False),

    'refresh'     : ('tbzrcommands.cmd_refresh',
                     'cmd_refresh', None,
                     invoke_about, True),

    'pull'        : (QBZR_CMDS,
                     'cmd_qpull', 'directory',
                     invoke_about, True),

    'send'        : (QBZR_CMDS,
                     'cmd_qsend', 'submit_branch',
                     invoke_about, False),

    'merge'       : (QBZR_CMDS,
                     'cmd_qmerge', 'directory',
                     invoke_about, False),

    'export'      : (QBZR_CMDS,
                     'cmd_qexport', 'branch_or_subdir',
                     invoke_about, False),

    'uncommit'    : (QBZR_CMDS,
                     'cmd_quncommit', 'location',
                     invoke_about, True),

    'tag'         : (QBZR_CMDS,
                     'cmd_qtag', 'directory',
                     invoke_about, False),

    'conflicts'   : (QBZR_CMDS,
                     'cmd_qconflicts', None,
                     invoke_about, True),

    'bind'        : (QBZR_CMDS,
                     'cmd_qbind', 'location',
                     invoke_about, True), # Really needs refresh?

    'unbind'      : (QBZR_CMDS,
                     'cmd_qunbind', None,
                     invoke_about, True), # Really needs refresh?

    'switch'      : (QBZR_CMDS,
                     'cmd_qswitch', None,
                     invoke_about, True), # Really needs refresh?

    'info'        : (QBZR_CMDS,
                     'cmd_qinfo', 'location',
                     invoke_about, False),

    'run'         : (QBZR_CMDS,
                     'cmd_qrun', 'directory',
                     invoke_about, True),

    'explorer'    : ('bzrlib.plugins.explorer.lib.commands',
                     'cmd_explorer', 'location',
                     invoke_explorer, False),
}

commands = commands_qt

def invoke_command(command_name, filenames, directory=None):
    try:
        mod_name, klass_name, arg_name, invoker, need_refresh = commands[command_name]
    except KeyError:
        raise UnrecognizedCommandError(command_name)

    # before attempting the import ensure bzrlib has loaded plugins and
    # modified the plugin path accordingly in binary builds.
    try:
        from bzrlib.commands import _register_builtin_commands
        _register_builtin_commands()
        del _register_builtin_commands
    except ImportError:
        pass
    from bzrlib.commands import install_bzr_command_hooks
    install_bzr_command_hooks()
    from bzrlib.plugin import load_plugins
    load_plugins()

    __import__(mod_name)
    klass = getattr(sys.modules[mod_name], klass_name)
    arg = None
    kw = {}
    takes_options = getattr(klass, 'takes_options', [])
    for opt in takes_options:
        # options will generally be an 'Options' object or a string.  ui-mode
        # is *always* setup as an Option instance.
        if getattr(opt, 'name', None) == 'ui-mode':
            kw['ui_mode'] = True
            break
    
    try:
        # finally get down to business...        
        instance = klass()
        invoker(instance, arg_name, filenames, directory, kw)
    finally:
        if need_refresh:
            refresh_cache(filenames, directory)

def refresh_cache(filenames, directory):
    path = directory
    if path is None and filenames:
        path = filenames[0]

    if path is not None:
        import tbzrlib.pipe.client as client
        link = client.get_cache_link()
        link.request_refresh_status(path)
    else:
        import tbzrlib.app
        app = tbzrlib.app.get_app()
        app.do_refresh_cache()

# option parser that is smarter when its in a Windows GUI.
class OptionParser(optparse.OptionParser):
    def exit(self, status=0, msg=None):
        if msg:
            if have_console:
                sys.stderr.write(msg)
            else:
                error_dialog(msg)
        sys.exit(status)

    def print_help(self, file=None):
        if have_console:
            optparse.OptionParser.print_help(self, file)
        else:
            error_dialog(self.format_help())


def get_unicode_argv():
    from bzrlib import osutils
    if getattr(osutils, 'get_unicode_argv', None):
        return osutils.get_unicode_argv()   # it should do the right thing
    else:
        try:
            user_encoding = get_user_encoding()
            return [a.decode(user_encoding) for a in sys.argv[1:]]
        except UnicodeDecodeError:
            # we can't produce valid unicode argv
            warning(("Command-line parameter '%r' is unsupported "
                "by the current encoding." % a))
            return sys.argv[1:]


def main(argv=None):
    i18n.install()
    if argv is None:
        argv = get_unicode_argv()

    parser = OptionParser()
    parser.add_option("-v", "--verbose",
                  action="store_true",
                  help="Increase verbosity.")

    parser.add_option("", "--log-level", action="store",
                      default=logging.INFO)

    parser.add_option("", "--file-list",
                  action="store",
                  help="The name of a file containing the filenames to operate on")

    parser.add_option("", "--file",
                  action="append", dest="files", default=[],
                  help="The name of a file to operate on")

    parser.add_option("", "--dir",
                  action="store",
                  help="Working Directory.")

    all_commands = ", ".join(sorted(commands.keys()))
    parser.add_option("", "--command",
                      action="store",
                      help="The name of the command to execute, one of: " +
                           all_commands)

    try:
        options, args = parser.parse_args(argv)
        if args:
            parser.error("This program only accepts options - see --help for details")
        if not options.command:
            parser.error("You must specify the --command option")
    except optparse.OptParseError, exc:
        # command-line option failure.
        # ack - haven't setup logging yet :(  This is a problem with py2exe
        msg = "\n".join(traceback.format_exception_only(*sys.exc_info()[:2]))
        if have_console:
            print msg
        else:
            error_dialog(msg)
        sys.exit(1)

    try:
        setup_logging(options)

        if options.dir:
            os.chdir(options.dir)

        filenames = options.files[:]
        if options.file_list:
            try:
               f = open(options.file_list, 'r')
            except EnvironmentError, exc:
                parser.error("Failed to open --file-list file %r" % options.file_list)
            filenames += [l.strip() for l in f.readlines()]
            f.close()

        invoke_command(options.command, filenames, options.dir)
    except:
        warning("Failed to execute command %r", options.command, exc_info=True)
        if not have_console:
            msg = "\n".join(traceback.format_exception_only(*sys.exc_info()[:2]))
            error_dialog(msg)


if __name__=='__main__':
    main()
