# This is a utility for tracing debug output from Tortoise BZR.

# We collect output from 2 places:
# * From the win32trace collector, primarily for the (old) Python implemented
#   shell extension.  This will be rarely used, but we continue to supply the
#   DLL for diagnostic purposes, should the C++ implemented one fail.
# * From any process that calls OutputDebugString() - such as our C++
#   implemented shell extension.

# It dumps the output to the console, and also writes it to '.tbzrtrace.log'
# in the same directory as .bzr.log (the 'my documents' folder)

# XXX - todo - consider giving tbzrcache an option to write its output
# here too?  Or maybe even have this program log what it gets to .bzr.log, so
# all "evidence" is in the same place?

import sys
import os
import mmap
import struct
import time
import win32api
import win32con
import win32event
import win32trace
from win32com.shell import shell, shellcon

stop_event = win32event.CreateEvent(None, 0, 0, None)
output_file = None

def console_ctrl_handler(evt):
    # NOTE that this is called on its own thread making it difficult to do
    # things like break into pdb etc.
    stop_reason = None
    # False means "run the next/default handler", True means don't call others.
    rc = False
    # Leave Ctrl+Break alone so it can kill the app in case things are serious!
    if evt in (win32con.CTRL_C_EVENT,): # win32con.CTRL_BREAK_EVENT):
        stop_reason = "Ctrl+C"
        rc = True
    # During a log-off etc process the console is also closed, so this works
    # OK here.
    elif evt == win32con.CTRL_CLOSE_EVENT:
        stop_reason = "Console closing"
        rc = True

    if stop_reason is not None:
        #print "%s - stopping application..." % stop_reason
        win32event.SetEvent(stop_event)

    return rc

class ODSReader:
    """A reader of the system's OutputDebugString stream."""
    # Via http://timgolden.me.uk/python/win32_how_do_i/capture-OutputDebugString.html
    #
    # Original code by Dan Brotherston
    #
    def __init__(self):
        self.buffer_ready = win32event.CreateEvent(None, 0, 0, "DBWIN_BUFFER_READY")
        self.event = win32event.CreateEvent (None, 0, 0,  "DBWIN_DATA_READY")
        self.buffer = mmap.mmap (0, 4096, "DBWIN_BUFFER", mmap.ACCESS_WRITE)
        # Signal that we're ready to accept debug output
        win32event.SetEvent(self.buffer_ready)

    def iter_chunks(self):
        self.buffer.seek (0)
        # The first DWORD is the process id which generated the output
        process_id, = struct.unpack ("L", self.buffer.read (4))
        data = self.buffer.read(4092)
        if "\0" in data:
          got = data[:data.index ("\0")]
        else:
          got = data
        yield "%04d:" % (process_id,)
        yield got
        # Signal that we're ready to accept more
        win32event.SetEvent(self.buffer_ready)

class Win32TraceReader:
    def __init__(self):
        win32trace.InitRead()
        self.event = win32trace.GetHandle()

    def iter_chunks(self):
        got = win32trace.read()
        yield "????:"
        yield got

def run():
    readers = ODSReader(), Win32TraceReader()
    events = [r.event for r in readers] + [stop_event]
    while 1:
        rc = win32event.WaitForMultipleObjects(events, False, win32event.INFINITE)
        offset = rc - win32event.WAIT_OBJECT_0
        if offset < len(readers):
            reader = readers[offset]
            for chunk in reader.iter_chunks():
                print chunk,
                output_file.write(chunk)
        else:
            # must be out stop event!
            break
    print "Stopped..."


def main():
    global output_file
    # open a log file that also gets the output.
    dir = shell.SHGetFolderPath(0, shellcon.CSIDL_PERSONAL, None, 0)
    fname = os.path.join(dir, ".tbzrtrace.log")
    # don't append to the log - the hope is they start this process, repro
    # a fault, then take a copy with the minimum necessary recorded.
    output_file = open(fname, "w", 0)
    output_file.write("tbzrtrace output log opened at %s\n" % time.ctime(time.time()))

    # a banner
    print "tbzrtrace - This program will collect debugging output from TortoiseBZR"
    print "A copy of this output is being written to:"
    print "  ", fname
    print "Please include this file in any bug reports or mailing list queries."
    print "Press Ctrl+C to finish..."

    # Run the app.
    win32api.SetConsoleCtrlHandler(console_ctrl_handler, True)
    run()

    # and done!
    output_file.write("tbzrtrace output log closed at %s\n" % time.ctime(time.time()))
    output_file.close()

    
if __name__=='__main__':
    main()
