# setup.py for Tortoise Bazaar.
# Note that this doesn't make binaries - that is done by Bazaar's main setup

# UI building support stolen from QBzr.

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from distutils import errors, log
from distutils.core import Command
from distutils.command.build import build as std_build
from distutils.dep_util import newer
from StringIO import StringIO
import glob
import os
import re
import sys
import subprocess
import extras


_translate_re = re.compile(r'QtGui\.QApplication.translate\(.*?, (.*?), None, QtGui\.QApplication\.UnicodeUTF8\)')
_import_re = re.compile(r'(from PyQt4 import QtCore, QtGui)')


class build_ui(Command):
    description = "build TBzr UI files"
    user_options = [
        ('force', 'f', 'Force creation of ui files'),
        ]
    boolean_options = ['force']

    def initialize_options(self):
        self.force = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('force', 'force'))

    def run(self):
        from PyQt4 import uic
        # Unlike qbzr, we just put them in the same dir with the same base name
        for uifile in glob.glob("tbzrcommands/ui/*.ui"):
            uifile = uifile.replace('\\', '/')
            pyfile = "tbzrcommands/ui/" + os.path.splitext(os.path.basename(uifile))[0] + ".py"
            if self.force or newer(uifile, pyfile):
                log.info("compiling %s -> %s", uifile, pyfile)
                tmp = StringIO()
                uic.compileUi(uifile, tmp)
                source = _translate_re.sub(r'gettext(\1)', tmp.getvalue())
                source = source.replace("from PyQt4 import QtCore, QtGui",
                    "from PyQt4 import QtCore, QtGui\n"
                    "from tbzrlib.i18n import gettext\n")
                f = open(pyfile, "wb")
                f.write(source)
                f.close()

class build_docs(Command):
    description = "build TBzr docs"
    user_options = [
        ('force', 'f', 'Force creation of ui files'),
        ]
    boolean_options = ['force']

    def initialize_options(self):
        self.force = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('force', 'force'))

    def run(self):
        import bzrlib
        bzr_src = os.path.abspath(os.path.join(bzrlib.__path__[0], '..'))
        # the 'tools' dir is in the parent of bzrlib.
        script = os.path.join(bzr_src, 'tools', 'rst2html.py')
        if not os.path.isfile(script):
            raise RuntimeError("can't find rst2html.py")
        css = os.path.join(bzr_src, 'doc', 'default.css')
        for check in (script, css):
            if not os.path.isfile(check):
                raise RuntimeError("can't find %r" % check)

        for in_name in glob.glob("doc/*.txt"):

            out_name = os.path.splitext(in_name)[0] + ".html"
            if self.force or newer(in_name, out_name):
                log.info("generating %s -> %s", in_name, out_name)
                args = [sys.executable, script,
                        '--link-stylesheet',
                        '--footnote-references=superscript',
                        '--halt=warning',
                        '--stylesheet=%s' % css,
                        in_name, out_name]
                rc = subprocess.call(args)
                if rc:
                    raise RuntimeError("doc build failed with exit code %s" % rc)

class build_shellext(Command):
    description = "build c++ implemented tbzr shell extension"
    user_options = [
        ('force', 'f', 'Force creation of ui files'),
        ]
    boolean_options = ['force']
    def initialize_options(self):
        self.force = None

    def finalize_options(self):
        self.set_undefined_options('build',
                                   ('force', 'force'))

    def find_devenv(self):
        "Find devenv.com - a tool able to build project files."
        import _winreg
        # We always look in the 32bit hive of the registry, even when we
        # happen to be a 64bit executable.
        KEY_WOW64_32KEY = 512 # in _winreg module in 2.6 and later.
        try:
            hk = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                                 r"Software\Microsoft\VisualStudio\9.0\Setup\VS",
                                 _winreg.KEY_READ|KEY_WOW64_32KEY)
            edir, typ = _winreg.QueryValueEx(hk, "EnvironmentDirectory")
        except _winreg.error, exc:
            self.debug_print(
                "VS2008 is not installed due to registry error: %s" % exc)
            return None
        ret = os.path.join(edir, "devenv.com")
        if not os.path.isfile(ret):
            raise RuntimeError("Can't find devenv.com in '%s'" % dir)
        return ret

    def run(self):
        # For now we need VS2008, but over time, it is expected this can
        # be generalized to "for all python versions other than 2.5, the
        # compiler used to build Python will be acceptable."
        proj_dir = os.path.join(os.path.dirname(__file__), "shellext")
        sln = os.path.join(proj_dir, "tbzrshellext.sln")
        out_dir = os.path.join(proj_dir, "build")
        driver = self.find_devenv()
        if driver: # Using VC++
            if self.force:
                cmd = "/rebuild"
            else:
                cmd = "/build"
            for cfg, target in [
                    ("Release|Win32", "tbzrshellext_x86.dll"),
                    ("Release|x64", "tbzrshellext_x64.dll"),
                     ]:
                argv = [driver, sln, cmd, cfg, "/nologo"]
                self.spawn(argv, search_path=False)
                # The build process also did a 'per-user' registration of the
                # extension, so just 'undo' that now.
                # using 'search_path=True means we must pass a string :(
                fqtarget = os.path.join(out_dir, target)
                if not os.path.isfile(fqtarget):
                    raise RuntimeError("Target '%s' failed to build" % fqtarget)
        else: # Using msbuild.
            vcproj = os.path.join(proj_dir, 'tbzrshellext.vcproj')
            cmd = ['msbuild', '/nologo', vcproj]
            if self.force:
                cmd.append('/t:Rebuild')
            for cfg, target in [
                    ("Win32", "tbzrshellext_x86.dll"),
                    ("x64", "tbzrshellext_x64.dll"),
                    ]:
                argv = list(cmd)
                argv.append('/p:Configuration="Release";Platform="%s"' % (cfg,))
                try:
                    self.spawn(argv)
                except StandardError:
                    self.debug_print("Windows SDK is needed to build shell extension.")
                    raise
                fqtarget = os.path.join(out_dir, target)
                if not os.path.isfile(fqtarget):
                    raise RuntimeError("Target '%s' failed to build" % fqtarget)


class build(std_build):
    sub_commands = std_build.sub_commands + [
        ('build_ui', lambda(me): True),
        ('build_docs', lambda(me): True),
        ('build_shellext', lambda(me): True),
        ('build_mo', lambda(me): True),
        ]

cmdclass = {
    'build_ui': build_ui,
    'build_docs': build_docs,
    'build_shellext': build_shellext,
    'build': build,
}
cmdclass.update(extras.cmdclass)

setup(name='tortoisebzr',
      description='Toroise Bazaar',
      keywords='bzr tortoise',
      url='http://wiki.bazaar.canonical.com/TortoiseBzr',
      license='GPL',
      author='TortoiseBzr Developers',
      author_email='tortoisebzr-developers@lists.launchpad.net',
      packages=['tbzrlib',
                'tbzrcommands',
                ],
      cmdclass=cmdclass,
      )
