#include "stdafx.h"
#include "Config.h"
#include "RemoteLink.h"
#include <process.h>

CConfig *CConfig::m_singleton = NULL;
CriticalSection CConfig::m_lock("CConfig::lock");

#define IS_VALID_HANDLE(h)	(h != NULL && h != INVALID_HANDLE_VALUE)

CConfig::CConfig(void)
{
	m_thread_stop_event = INVALID_HANDLE_VALUE;
	m_config_changed_event = INVALID_HANDLE_VALUE;
}

CConfig::~CConfig(void)
{
	if (!IS_VALID_HANDLE(m_thread_stop_event)) {
		ResetEvent(m_thread_stop_event);
		CloseHandle(m_thread_stop_event);
	}

	if (!IS_VALID_HANDLE(m_config_changed_event)) {
		CloseHandle(m_config_changed_event);
	}
}

/* static */
CConfig *CConfig::GetConfig()
{
	if (!m_singleton) {
		AutoCritical _lock(m_lock);
		if (!m_singleton)
			m_singleton = new CConfig();
	}
	return m_singleton;
}

void CConfig::AddListener(CConfig::IChangeListener *listener)
{
	AutoCritical _lock(m_lock);
	for (std::vector<IChangeListener *>::iterator itr = m_listeners.begin() ; itr != m_listeners.end(); ++itr) {
		if (*itr == listener)
			return;
	}
	m_listeners.push_back(listener);
}

void CConfig::RemoveListener(CConfig::IChangeListener *listener)
{
	AutoCritical _lock(m_lock);
	for (std::vector<IChangeListener *>::iterator itr = m_listeners.begin() ; itr != m_listeners.end(); ++itr) {
		if (*itr == listener)
		{
			m_listeners.erase(itr);
			return;
		}
	}
}


std::wstring CConfig::GetString(std::wstring config_name)
{
	{
		AutoCritical _lock(m_lock);
		std::map<std::wstring, std::wstring>::iterator itr = m_cache.find(config_name);
		if (itr != m_cache.end())
			return itr->second;
	}

	return this->RequestGetConfig(config_name);
}

BOOL CConfig::GetBool(std::wstring config_name)
{
	std::wstring config_value = this->GetString(config_name);
	if (_wcsicmp(config_value.c_str(), L"true") == 0 || _wtoi(config_value.c_str()) != 0)
		return TRUE;
	else
		return FALSE;
}

std::wstring CConfig::RequestGetConfig(std::wstring config_name)
{
	if (!IS_VALID_HANDLE(m_config_changed_event))
	{
		// Create event which will notify changing of configurations.
		std::wstring event_name = CRemoteLink::GetLink()->RequestGetParam(L"system:config_event_name");
		m_config_changed_event = CreateEventW(NULL, TRUE, FALSE, event_name.c_str());
		if (!IS_VALID_HANDLE(m_config_changed_event)) {
			DWORD err = ::GetLastError();
			throw TBzrWin32Exception(err, "Failed to create config changed event");
		}

		// Begin thread which will watch changinc of configurations.
		m_thread_stop_event = CreateEventW(NULL, TRUE, FALSE, NULL);
		if (!IS_VALID_HANDLE(m_thread_stop_event)) {
			DWORD err = ::GetLastError();
			throw TBzrWin32Exception(err, "Failed to create thread stop event");
		}
		_beginthread(CConfig::WatchConfigChange, 0, this);
	}

	// Get config value from tbzrcache process.
	{ // Scope for lock
		AutoCritical _lock(m_lock);
		std::wstring param_name(L"config:" + config_name);
		std::wstring config_value = CRemoteLink::GetLink()->RequestGetParam(param_name.c_str());
		m_cache.insert(std::pair<std::wstring, std::wstring>(config_name, config_value));
		return config_value;
	}
}

/* static */
void CConfig::WatchConfigChange(void *param)
{
	CConfig *conf = (CConfig *)param;
	HANDLE events[2] = {conf->m_thread_stop_event, conf->m_config_changed_event};
	for ( ; ; ) {
		DWORD ret = WaitForMultipleObjects(2, events, FALSE, INFINITE);
		if (ret == WAIT_OBJECT_0 + 1) {
			// Clear value cache when config is changed.
			AutoCritical _lock(m_lock);
			conf->m_cache.clear();
			for (std::vector<IChangeListener *>::iterator itr = conf->m_listeners.begin() ; itr != conf->m_listeners.end() ; ++itr) {
				(*itr)->ConfigChanged();
			}
			continue;
		}
		else
			break;
	}
}

