// ContextMenu.cpp : Implementation of CContextMenu

#include "stdafx.h"
#include "ContextMenu.h"
#include "RemoteLink.h"
#include "Config.h"

#define GetPIDLFolder(pida) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[0])
#define GetPIDLItem(pida, i) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[i+1])

#define MAYBE_THROW(exp) if (!SUCCEEDED(hr=(exp))) {\
    TBZR_ERROR("Calling " #exp " failed (res=%x)", hr);\
    throw TBzrHresultException(hr);}

static const char* TBZR_ID_STRING = "TortoiseBZR";

UINT CContextMenu::m_CF_SHELLIDLIST=::RegisterClipboardFormat(CFSTR_SHELLIDLIST);

static std::wstring string_from_strret(STRRET &strret, PCUITEMID_CHILD pidl)
{
	std::wstring ret;
	WCHAR *s = 0;
	HRESULT hr = StrRetToStr (&strret, pidl, &s);
	if (SUCCEEDED(hr))
		ret = s;
	CoTaskMemFree(s);
	return ret;
}


static void InitBitmapInfo(__out_bcount(cbInfo) BITMAPINFO *pbmi, ULONG cbInfo, LONG cx, LONG cy, WORD bpp)
{
    ZeroMemory(pbmi, cbInfo);
    pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    pbmi->bmiHeader.biPlanes = 1;
    pbmi->bmiHeader.biCompression = BI_RGB;

    pbmi->bmiHeader.biWidth = cx;
    pbmi->bmiHeader.biHeight = cy;
    pbmi->bmiHeader.biBitCount = bpp;
}

static HRESULT Create32BitHBITMAP(HDC hdc, const SIZE *psize, __deref_opt_out void **ppvBits, __out HBITMAP* phBmp)
{
    *phBmp = NULL;

    BITMAPINFO bmi;
    InitBitmapInfo(&bmi, sizeof(bmi), psize->cx, psize->cy, 32);

    HDC hdcUsed = hdc ? hdc : GetDC(NULL);
    if (hdcUsed)
    {
        *phBmp = CreateDIBSection(hdcUsed, &bmi, DIB_RGB_COLORS, ppvBits, NULL, 0);
        if (hdc != hdcUsed)
        {
            ReleaseDC(NULL, hdcUsed);
        }
    }
    return (NULL == *phBmp) ? E_OUTOFMEMORY : S_OK;
}

static bool HasAlpha(__in ARGB *pargb, SIZE& sizImage, int cxRow)
{
	ULONG cxDelta = cxRow - sizImage.cx;
	for (ULONG y = sizImage.cy; y; --y)
	{
		for (ULONG x = sizImage.cx; x; --x)
		{
			if (*pargb++ & 0xFF000000)
			{
				return true;
			}
		}

		pargb += cxDelta;
	}

	return false;
}

static bool IsIllegalFolder(std::wstring folder)
{
    static const int csidlarray[] = {
        CSIDL_BITBUCKET,
        CSIDL_CDBURN_AREA,
        CSIDL_COMMON_STARTMENU,
        CSIDL_COMPUTERSNEARME,
        CSIDL_CONNECTIONS,
        CSIDL_CONTROLS,
        CSIDL_COOKIES,
        CSIDL_FONTS,
        CSIDL_HISTORY,
        CSIDL_INTERNET,
        CSIDL_INTERNET_CACHE,
        CSIDL_NETHOOD,
        CSIDL_NETWORK,
        CSIDL_PRINTERS,
        CSIDL_PRINTHOOD,
        CSIDL_RECENT,
        CSIDL_SENDTO,
        CSIDL_STARTMENU,
        0
    };

    if (folder.compare(0, 3, L"::{") == 0)
        return true;

    TCHAR buf[MAX_PATH];	//MAX_PATH ok, since SHGetSpecialFolderPath doesn't return the required buffer length!
    LPITEMIDLIST pidl = NULL;
    for (int i = 0; csidlarray[i]; i++) {
        pidl = NULL;
        if (SHGetFolderLocation(NULL, csidlarray[i], NULL, 0, &pidl)!=S_OK)
            continue;
        if (!SHGetPathFromIDList(pidl, buf))
        {
            // not a file system path, definitely illegal for our use
            CoTaskMemFree(pidl);
            continue;
        }
        CoTaskMemFree(pidl);
        if (_tcslen(buf)==0)
            continue;
        if (_tcscmp(buf, folder.c_str())==0)
            return true;
    }
    return false;
}


// A couple of helper "owner" classes to worm around the lack
// of C++ 'finally' - the objects will be cleaned up even on
// exception exit.
class STGMEDIUMOwner
{
public:
	STGMEDIUMOwner(STGMEDIUM &medium) : m_medium(&medium) {;}
	~STGMEDIUMOwner() {
		ReleaseStgMedium(m_medium);
		// hrmph - TSVN does this (but somewhat strangely?)
		// but ReleaseStgMedium imply not necessary?
		if (m_medium->pUnkForRelease)
			m_medium->pUnkForRelease->Release();
	}
protected:
	STGMEDIUM *m_medium;
};

class GlobalLockOwner
{
public:
	GlobalLockOwner(void *val) : m_val(val) {;}
	~GlobalLockOwner() {
		::GlobalUnlock(m_val);
	}
protected:
	void *m_val;
};

CContextMenu::CContextMenu() : m_refcnt(0)
{
	OSVERSIONINFOEX inf;
	SecureZeroMemory(&inf, sizeof(OSVERSIONINFOEX));
	inf.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	GetVersionEx((OSVERSIONINFO *)&inf);
	fullver = MAKEWORD(inf.dwMinorVersion, inf.dwMajorVersion);
}

STDMETHODIMP CContextMenu::QueryInterface(const IID &riid, void **ppv)
{
    if (!ppv)
        return E_POINTER;
    *ppv = NULL;

    if (IsEqualIID(riid, IID_IShellExtInit) || IsEqualIID(riid, IID_IUnknown)) {
        *ppv = (LPSHELLEXTINIT) this;
    }
    else if (IsEqualIID(riid, IID_IContextMenu)) {
        *ppv = (LPCONTEXTMENU) this;
    }
    else if (IsEqualIID(riid, IID_IContextMenu2)) {
        *ppv = (LPCONTEXTMENU2) this;
    }
    else if (IsEqualIID(riid, IID_IContextMenu3)) {
        *ppv = (LPCONTEXTMENU3) this;
    }
    if (*ppv) {
        AddRef();
        return NOERROR;
    }

    return E_NOINTERFACE;
}


void CContextMenu::DoInitialize(PCIDLIST_ABSOLUTE pidlFolder, IDataObject *pdtobj, HKEY /*hkeyProgID*/)
{
	HRESULT hr;
	m_fileselected = true;
	// get selected files/folders
	if (pdtobj) {
		STGMEDIUM medium;
		FORMATETC fmte = {(CLIPFORMAT)m_CF_SHELLIDLIST,
			(DVTARGETDEVICE FAR *)NULL, 
			DVASPECT_CONTENT, 
			-1, 
			TYMED_HGLOBAL};
		MAYBE_THROW(pdtobj->GetData(&fmte, &medium));
		STGMEDIUMOwner mowner(medium); // guarantee cleanup
		if (medium.tymed != TYMED_HGLOBAL) {
			TBZR_ERROR("Non hglobal STGMEDIUM fetched");
            throw TBzrHresultException(E_UNEXPECTED, "Non hglobal STGMEDIUM fetched");
		}
		//Enumerate PIDLs which the user has selected
		CIDA* cida = (CIDA*)GlobalLock(medium.hGlobal);
		if (!cida) {
			TBZR_ERROR("No data available!\r\n");
            throw TBzrHresultException(E_UNEXPECTED, "No data available.");
		}
		GlobalLockOwner glo(cida); // guarantee cleanup
		LPCITEMIDLIST parentPIDL = GetPIDLFolder(cida);
		if (!parentPIDL) {
			TBZR_ERROR("No parent PIDL??\r\n");
            throw TBzrHresultException(E_UNEXPECTED, "No parent PIDL?");
		}
        IShellFolder *desktop, *parent;
		MAYBE_THROW(::SHGetDesktopFolder(&desktop));
        AutoComPtr<IShellFolder> _desktop(desktop);
        if (parentPIDL->mkid.cb==0) {// already the desktop!
			parent = desktop;
            parent->AddRef();
        }
		else {
			MAYBE_THROW(desktop->BindToObject(parentPIDL, 0, IID_IShellFolder, (void**)&parent))
		}
        AutoComPtr<IShellFolder> _parent(parent);
		for (UINT i = 0; i < cida->cidl; ++i) {
			LPCITEMIDLIST childPIDL = GetPIDLItem (cida, i);
			if (!childPIDL) {
				TBZR_ERROR("No child PIDL??\r\n");
                throw TBzrHresultException(E_UNEXPECTED);
			}
			STRRET strret;
			MAYBE_THROW(parent->GetDisplayNameOf(childPIDL, SHGDN_NORMAL | SHGDN_FORPARSING, &strret));
			std::wstring this_name = string_from_strret(strret, childPIDL);
			TBZR_MUTTER("Context menu filename %d is '%S'\r\n", i, this_name.c_str());
			m_filenames.push_back(this_name);
		}
	}
	// Now check pidlFolder
	if (pidlFolder) {
		m_fileselected = false;
		// python impl just overwrites any from the data object???
		m_filenames.clear();
		WCHAR path_buf[MAX_PATH+1];
		MAYBE_THROW(SHGetPathFromIDListW(pidlFolder, path_buf));
		m_filenames.push_back(path_buf);
	}
	TBZR_MUTTER("CContextMenu intialization succeeded with %d files\n", m_filenames.size());
	//ATLASSERT(SUCCEEDED(hr)); // error exit is via exceptions... TODO
}

// IShellExtInit
STDMETHODIMP CContextMenu::Initialize(PCIDLIST_ABSOLUTE pidlFolder, IDataObject *pdtobj, HKEY hkeyProgID)
{
	try {
		DoInitialize(pidlFolder, pdtobj, hkeyProgID);
		return S_OK;
	}
    catch (TBzrException &e) {
        TBZR_ERROR("CContextMenu::Initialize failed, %s.\n", e.what());
    }
    return E_FAIL;
}

// IContextMenu

UINT CContextMenu::BuildFromRemoteItems(HMENU hmenu, UINT indexMenu, 
										UINT idCmdFirst, const std::vector<TBZRMenuInfo> &items)
{
	class cur_menu {
	public:
		cur_menu(HMENU h, UINT i) : hmenu(h), indexMenu(i) {;}
		HMENU hmenu;
		UINT indexMenu;
	};
	TBZR_MUTTER("Starting build of context menu\n");
	std::vector<cur_menu> submenus; // a stack
	submenus.push_back(cur_menu(hmenu, indexMenu));
	int current_level = 0;
	int current_id = idCmdFirst;
	for (size_t i=0;i<items.size();i++) {
		const TBZRMenuInfo &item = items[i];
		HMENU hsubmenu = NULL;
		m_handlers[current_id] = item;
		if (item.level > current_level) {
			//ATLASSERT(item.level == current_level + 1);
			current_level++;
			submenus.push_back(cur_menu(CreatePopupMenu(), indexMenu));
			indexMenu = 0; // first item in new sub-menu.
		} else if (item.level < current_level) {
			//ATLASSERT(item.level == current_level - 1);
			current_level--;
			cur_menu last = submenus[submenus.size()-1];
			submenus.pop_back();
			hsubmenu = last.hmenu;
			indexMenu = last.indexMenu;
		}

		MENUITEMINFO menuinfo;
		menuinfo.cbSize = sizeof(menuinfo);
        menuinfo.dwItemData = (ULONG_PTR)TBZR_ID_STRING;
		if (!item.text.empty()) {
			menuinfo.fMask = MIIM_FTYPE | MIIM_ID | MIIM_STRING | MIIM_DATA;
			if (hsubmenu) {
				menuinfo.fMask |= MIIM_SUBMENU;
				menuinfo.hSubMenu = hsubmenu;
			}
			if (fullver > 0x500 && !item.ico.empty()) {
				menuinfo.fMask |= MIIM_BITMAP;
				if (fullver >= 0x600) { // Vista or lator
					// Use PARGB Bitmap
					menuinfo.hbmpItem = IconToBitmapPARGB32(item);
				}
				else {
					// Use owner drawing
					menuinfo.hbmpItem = HBMMENU_CALLBACK;
				}
			}
			menuinfo.fType = MFT_STRING;
			menuinfo.dwTypeData = (LPWSTR)item.text.c_str();
			menuinfo.cch = (UINT)item.text.size();
			menuinfo.wID = current_id;
			++current_id;
			TBZR_MUTTER("Context menu item '%S' at depth %d\n", item.text.c_str(), current_level);
		} else {
			// its a sep!
			menuinfo.fMask = MIIM_FTYPE | MIIM_STATE;
			menuinfo.fType = MFT_SEPARATOR;
			menuinfo.fState = MFS_ENABLED;
			TBZR_MUTTER("Context menu separator at depth %d\n", current_level);
		}
		HMENU hmenu_this = submenus[submenus.size()-1].hmenu;
		if (!InsertMenuItem(hmenu_this, indexMenu, TRUE, &menuinfo)) {
			TBZR_ERROR("InsertMenuItem failed: %d\n", ::GetLastError());
			// but continue on regardless - maybe only 1 failed?
		}
		indexMenu++;
	}

    if (fullver >= 0x600) {
        MENUINFO MenuInfo = {sizeof(MenuInfo)};
        MenuInfo.fMask   = MIM_STYLE | MIM_APPLYTOSUBMENUS;
        MenuInfo.dwStyle = MNS_CHECKORBMP;

        SetMenuInfo(hmenu, &MenuInfo);
    }

	// Return the number of commands we added (hrmph - I don't think this
	// should include child menus?)
	TBZR_MUTTER("Context menu complete - added %d items\n", current_id - idCmdFirst);
	return current_id - idCmdFirst;
}

STDMETHODIMP CContextMenu::QueryContextMenu(HMENU hmenu, UINT indexMenu,
                                            UINT idCmdFirst, UINT /*idCmdLast*/, UINT uFlags)
{
    TBZR_MUTTER("QueryContextMenu()\n");

    try {
		if (!g_isExplorer && CConfig::GetConfig()->GetBool(L"icon_overlays.show_only_in_explorer"))
			return S_OK;
    }
	catch (TBzrException &e) {
		TBZR_ERROR("CConfig::GetBool failed. %s\r\n", e.what());
		return E_FAIL;
	}

    if (uFlags & CMF_DEFAULTONLY)
        return S_OK;

	if (((uFlags & 0x000f)!=CMF_NORMAL)&&(!(uFlags & CMF_EXPLORE))&&(!(uFlags & CMF_VERBSONLY)))
		return S_OK;

    if (m_filenames.empty())
        return S_OK;

    for (size_t i = 0; i < m_filenames.size(); ++i) {
        if (IsIllegalFolder(m_filenames[i]))
            return S_OK;
    }

    // Avoid insert menus twice.
    int count = GetMenuItemCount(hmenu);
    for (int i=0; i<count; ++i) {
        TCHAR menubuf[MAX_PATH];
        MENUITEMINFO miif = {sizeof(MENUITEMINFO)};
        miif.fMask = MIIM_DATA;
        miif.dwTypeData = menubuf;
        miif.cch = sizeof(menubuf)/sizeof(TCHAR);
        GetMenuItemInfo(hmenu, i, TRUE, &miif);
        if (miif.dwItemData == (ULONG_PTR)TBZR_ID_STRING)
            return S_OK;
    }

    try {
        std::vector<TBZRMenuInfo> commands =
            CRemoteLink::GetLink()->RequestMenuCommands(m_filenames, 0);

        m_idfirst = idCmdFirst;
        return BuildFromRemoteItems(hmenu, indexMenu, idCmdFirst, commands);
    }
    catch (TBzrException &e) {
        TBZR_ERROR("RequestMenuCommands failed. %s\n", e.what());
        return E_FAIL;
    }
}

STDMETHODIMP CContextMenu::InvokeCommand(CMINVOKECOMMANDINFO *pici)
{
	TBZR_MUTTER("InvokeCommand()\n");
	if (HIWORD(pici->lpVerb)) {
		return E_FAIL;
	}

	UINT idCmd = LOWORD(pici->lpVerb) + m_idfirst;
	std::map<UINT, TBZRMenuInfo>::iterator it = m_handlers.find(idCmd);
	if (it == m_handlers.end()) {
		// Should we return E_FAIL?
		TBZR_ERROR("InvokeCommand for unknown ID %d\n", idCmd);
		return S_FALSE;
	}
	TBZRMenuInfo &item = (*it).second;

	// TODO: if more than 1 command, check the key-state
	TBZRMenuCommand command = item.commands[0];
	std::wstringstream scmdline;
	scmdline << command.cmdline;
	if (command.exec_type == EXEC_TBZRCOMMAND_STD) {
		// XXX - just the hwnd - which we don't support yet :)
		;
	} else if (command.exec_type==EXEC_TBZRCOMMAND_STD_SELECTION) {
		if (m_fileselected) {
			for (size_t i=0;i<m_filenames.size();i++) {
				scmdline << " \"--file=";
				scmdline << m_filenames[i];
				scmdline << "\"";
			}
		}
		else {
			scmdline << " \"--dir=";
			scmdline << m_filenames[0];
			scmdline << "\"";
		}
	} else {
		TBZR_ERROR("Unexpected exec type %d\n", command.exec_type);
		return E_UNEXPECTED;
	}
	// Now can execute it!
	std::wstring cmdline(scmdline.str());
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);
	TBZR_NOTE("Executing cmdline: %S\n", cmdline.c_str());
	if (!::CreateProcess(NULL, (LPWSTR)cmdline.c_str(), NULL, NULL, FALSE, DETACHED_PROCESS, NULL, NULL, &si, &pi)) {
		TBZR_ERROR("CreateProcess failed with error %d\n", ::GetLastError());
		return E_UNEXPECTED;
	}
	TBZR_MUTTER("CreatedProcess with PID %d\n", pi.dwProcessId);
	::CloseHandle(pi.hProcess);
	::CloseHandle(pi.hThread);
	return S_OK;
}

STDMETHODIMP CContextMenu::GetCommandString(UINT_PTR idCmd, UINT uType, UINT * /*pReserved*/, LPSTR pszName, UINT cchMax)
{
	//TBZR_MUTTER("GetCommandString()\n");
	UINT id = (UINT)idCmd + m_idfirst;

	switch (uType) {
	case GCS_VALIDATEA:
	case GCS_VALIDATEW:
		return m_handlers.find(id) != m_handlers.end() ? S_OK : S_FALSE;

	case GCS_VERBA:
	case GCS_VERBW:
		return S_FALSE;

	case GCS_HELPTEXTW:
		{
			std::map<UINT, TBZRMenuInfo>::iterator it = m_handlers.find(id);
			if (it == m_handlers.end()) {
				TBZR_NOTE("GetCommandString for unknown ID %p\n", idCmd);
				return E_UNEXPECTED;
			}
			TBZRMenuInfo &item = (*it).second;

			wcscpy_s((wchar_t *)pszName, cchMax, item.long_text.c_str());
			return S_OK;
		}
	case GCS_HELPTEXTA:
		TBZR_ERROR("HELPTEXTA - we only do unicode!\n");
		return S_FALSE;
	default:
		TBZR_ERROR("GetCommandString has unrecognised flag 0x%x\n", uType);
		return E_UNEXPECTED;
	}
}

STDMETHODIMP CContextMenu::HandleMenuMsg(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	return HandleMenuMsg2(uMsg, wParam, lParam, NULL);
}

STDMETHODIMP CContextMenu::HandleMenuMsg2(UINT uMsg, WPARAM /*wParam*/, LPARAM lParam, LRESULT *pResult)
{
	LRESULT res;
	if (pResult == NULL)
		pResult = &res;
	*pResult = FALSE;
	switch (uMsg)
	{
	case WM_MEASUREITEM:
		{
			MEASUREITEMSTRUCT* lpmis = (MEASUREITEMSTRUCT*)lParam;
			if (lpmis==NULL)
				break;
			lpmis->itemWidth += 2;
			if (lpmis->itemHeight < 16)
				lpmis->itemHeight = 16;
			*pResult = TRUE;
		}
		break;
	case WM_DRAWITEM:
		{
			DRAWITEMSTRUCT* lpdis = (DRAWITEMSTRUCT*)lParam;
			if ((lpdis==NULL)||(lpdis->CtlType != ODT_MENU))
				return S_OK;		//not for a menu

			// Find the item associated with the ID.
			std::map<UINT, TBZRMenuInfo>::iterator it = m_handlers.find(lpdis->itemID);
			if (it == m_handlers.end()) {
				TBZR_ERROR("HandleMenuMsg for unknown ID %d\n", lpdis->itemID);
				return NOERROR;
			}
			TBZRMenuInfo &item = (*it).second;
			HICON hIcon = (HICON)LoadHICON(item);
			if (hIcon) {
				DrawIconEx(lpdis->hDC,
					lpdis->rcItem.left - 16,
					lpdis->rcItem.top + (lpdis->rcItem.bottom - lpdis->rcItem.top - 16) / 2,
					hIcon, 16, 16,
					0, NULL, DI_NORMAL);
				DestroyIcon(hIcon);
				*pResult = TRUE;
			}
		}
		break;
	// XXX - todo - WM_MENUCHAR
	default:
		;
	}
	return NOERROR;
}

HICON CContextMenu::LoadHICON(const TBZRMenuInfo &item)
{
	size_t off = item.ico.find(L"#");
	HMODULE mod = NULL;
	UINT ico_flags = LR_DEFAULTSIZE;
	LPCTSTR resid;

	if (off!=std::wstring::npos) {
		std::wstring fname(item.ico.substr(0, off));
		std::wstring strid(item.ico.substr(off+1, 50)); // we want to the end
		resid = MAKEINTRESOURCE((WORD)_wtoi(strid.c_str()));
		mod = LoadLibraryEx(fname.c_str(), NULL, LOAD_LIBRARY_AS_DATAFILE);
		if (!mod) {
			TBZR_ERROR("LoadLibraryEx failed for '%S' - %d\n", 
				fname.c_str(), ::GetLastError());
			return NULL;
		}
	} else {
		resid = item.ico.c_str();
		ico_flags |= LR_LOADFROMFILE;
	}
	HICON hico = (HICON)LoadImage(mod, resid, IMAGE_ICON, 0, 0, ico_flags);
	if (hico==NULL)
		TBZR_ERROR("LoadImage for '%S' failed - %d\n", item.ico.c_str(), ::GetLastError());
	return hico;
}

HBITMAP CContextMenu::IconToBitmapPARGB32(const TBZRMenuInfo &menuInfo)
{
	std::map<std::wstring, HBITMAP>::iterator bitmap_it = bitmaps.lower_bound(menuInfo.ico);
	if (bitmap_it != bitmaps.end() && bitmap_it->first == menuInfo.ico)
		return bitmap_it->second;

	HICON hIcon = LoadHICON(menuInfo);
	if (!hIcon)
		return NULL;

	SIZE sizIcon;
	sizIcon.cx = GetSystemMetrics(SM_CXSMICON);
	sizIcon.cy = GetSystemMetrics(SM_CYSMICON);

	RECT rcIcon;
	SetRect(&rcIcon, 0, 0, sizIcon.cx, sizIcon.cy);
	HBITMAP hBmp = NULL;

	ARGB *pbits;
	std::vector<bool> opaque(sizIcon.cx * sizIcon.cy);

	HDC hdcDest = CreateCompatibleDC(NULL);
	if (hdcDest)
	{
		if (SUCCEEDED(Create32BitHBITMAP(hdcDest, &sizIcon, (void**)&pbits, &hBmp)))
		{
			HBITMAP hbmpOld = (HBITMAP)SelectObject(hdcDest, hBmp);
			if (hbmpOld)
			{
				bool use_opaque = false;
				memset((void*)pbits, 0, sizIcon.cx * sizIcon.cy * 4);
				if (DrawIconEx(hdcDest, 0, 0, hIcon, sizIcon.cx, sizIcon.cy, 0, NULL, DI_MASK)) {
					use_opaque = true;
					for (int i = 0; i < sizIcon.cx * sizIcon.cy; ++i) {
						opaque[i] = !pbits[i];
					}
				}
				memset((void*)pbits, 0, sizIcon.cx * sizIcon.cy * 4);
				if (DrawIconEx(hdcDest, 0, 0, hIcon, sizIcon.cx, sizIcon.cy, 0, NULL, DI_NORMAL)) {
					if (!HasAlpha(pbits, sizIcon, sizIcon.cx) && use_opaque) {
						for (int i = 0; i < sizIcon.cx * sizIcon.cy; ++i) {
							if (opaque[i])
								pbits[i] |= 0xFF000000;
							else
								pbits[i] = 0UL;
						}
					}
				}
				SelectObject(hdcDest, hbmpOld);
			}
		}

		DeleteDC(hdcDest);
	}

	DestroyIcon(hIcon);

	if(hBmp)
		bitmaps.insert(bitmap_it, std::make_pair(menuInfo.ico, hBmp));
	return hBmp;
}

/* vim: set ts=4 sw=4 noexpandtab : */
