// Simple TortoiseSVN-like Bazaar plugin for the Windows Shell
// Published under the GNU GPL, v2 or later.
// Copyright (C) 2008 Canonical
//
// Contributors:
//   Mark Hammond
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#include "stdafx.h"
#include "RemoteLink.h"

#include <exception>


CriticalSection CRemoteLink::m_lock("CRemoteLink::lock");
CRemoteLink *CRemoteLink::m_singleton = NULL;
DWORD CRemoteLink::m_timeout_ms = 10000;

// These must be the same as in python's tbzrlib.dispatcher module.
enum REQUEST_ID {
    REQUEST_HELLO,
    REQUEST_GOODBYE,
    REQUEST_RESTART,
    REQUEST_FILE_STATUS,
    REQUEST_MENU_COMMANDS,
    REQUEST_RELEASE_HANDLES,
    REQUEST_REFRESH_STATUS,
    REQUEST_GET_PARAM,

    REQUEST_ID_NUM  // End of enum
};

CRemoteLink::CRemoteLink()
{
	m_hevent = ::CreateEvent(NULL, FALSE, FALSE, NULL);
	m_hpipe = INVALID_HANDLE_VALUE;
	m_last_timeout = 0;
	m_buffer = new CLinkBufferManager();
}

CRemoteLink::~CRemoteLink()
{
	delete m_buffer;
	ClosePipe();
	::CloseHandle(m_hevent);
}

/* static */
CRemoteLink *CRemoteLink::GetLink()
{
	if (m_singleton) return m_singleton;

	AutoCritical _lock(m_lock);
	if (!m_singleton)
		m_singleton = new CRemoteLink();
	return m_singleton;
}

void CRemoteLink::ClosePipe()
{
	if (m_hpipe != INVALID_HANDLE_VALUE) {
		::CloseHandle(m_hpipe);
		m_hpipe = INVALID_HANDLE_VALUE;
	}
}

void CRemoteLink::EnsurePipeOpen()
{
	if (m_hpipe != INVALID_HANDLE_VALUE)
		return;

	std::wstring pipe_name = GetPipeName();
	TBZR_MUTTER("ensure_pipe_open - opening %S\r\n", pipe_name.c_str());
	// See comments in client.py for details about this loop, but note:
	// TSVN doesn't handle the failure of WaitNamedPipe nor that a 
	// subsequent ConnectNamedPipe may still fail with "pipe busy - so
	// this now more resembles transport_pipes.cc from Gdb than it does
	// TSVN
	DWORD pipe_wait_retries_left = 32;
	DWORD pipe_wait_timeout = 50; // wait 50ms for a pipe to become available.
	while (pipe_wait_retries_left > 0) {
		m_hpipe = CreateFile(pipe_name.c_str(),
							 GENERIC_READ | GENERIC_WRITE, // open mode,
							 0, // no shareing
							 NULL, // default security
							 OPEN_EXISTING,
							 FILE_FLAG_OVERLAPPED,
							 NULL); // no template
		if (m_hpipe != INVALID_HANDLE_VALUE) {
			// The pipe connected; change to message-read mode.
			DWORD dwMode = PIPE_READMODE_MESSAGE; 
			if (!SetNamedPipeHandleState(m_hpipe, &dwMode, NULL, NULL)) {
				ClosePipe();
                DWORD err = ::GetLastError();
				TBZR_ERROR("Failed to SetNamedPipeHandleState - %d\r\n", err);
                throw TBzrWin32Exception(err, "Failed to SetNamedPipeHandleState");
			}
			return;
		}
		DWORD err = ::GetLastError();
		if (err == ERROR_FILE_NOT_FOUND) {
			// server might still be starting - BUT - its also possible
			// the server is failing to start - this would mean we
			// *always* take a full timeout hit, which isn't what we
			// want.
			TBZR_MUTTER("EnsurePipeOpen - the pipe does not exist\r\n");
            throw TBzrWin32Exception(err, "EnsurePipeOpen - the pipe does not exist");
		}
		if (err != ERROR_PIPE_BUSY) {
			TBZR_ERROR("Non pipe-busy error: %d\r\n", err);
            throw TBzrWin32Exception(err, "Non pipe-busy error.");
		}
		// TSVNCache is running but is busy connecting a different client.
		// Do not give up immediately but wait for a few milliseconds until
		// the server has created the next pipe instance
		// BUT - see above - ERROR_FILE_NOT_FOUND is possible here.
		// transport_pipes.cc does the "busy-ish" loop here; it seems
		// we could possibly avoid this and just retry from the outer
		// loop again, but it must be done that way for a reason...
		while (pipe_wait_retries_left > 0 && 
			   ::WaitNamedPipe(pipe_name.c_str(), pipe_wait_timeout)) {
			if (::GetLastError() == ERROR_FILE_NOT_FOUND)
				::Sleep(0); // give the server a chance
			pipe_wait_retries_left -= 1;
		}
		// OK - theoretically a pipe instance is available - let's
		// go back around the loop and see if we can get it before
		// someone beats us to it!
		if (pipe_wait_retries_left) // might have already hit zero!
			pipe_wait_retries_left -= 1;
	}
	TBZR_ERROR("Failed to open pipe - hit retry limit\r\n");
    throw TBzrWin32Exception(ERROR_PIPE_BUSY, "Failed to open pipe - hit retry limit.");
}

void CRemoteLink::StartRemoteProcess()
{
	// How do we determine what executable to start, and with what args?
	// In 'binary builds' (ie, there the python code is built into executables)
	// we can deduce the location and name of the executable we should use.
	// But what about source builds, or custom options?
	// We solve this by looking in the tbzr config file for a couple
	// of settings.  We do this simply - we assume the Python code doesn't
	// to anything too fancy.

	// step1 - locate the tbzr config file...
	std::wstring cfg_name;
	WCHAR appdata[MAX_PATH];
	static wchar_t temp_env_path[2048]; // just for one variable!
	if (SUCCEEDED(SHGetFolderPath(0, CSIDL_APPDATA, NULL, SHGFP_TYPE_CURRENT, appdata))) {
		cfg_name = appdata;
		cfg_name.append(L"\\bazaar\\2.0\\tbzr.conf");
	}
	std::wstring new_py_path, old_py_path;
	std::wstring cmdline;
	std::wstring cmdline_extra;
	bool create_new_console = false;
	if (!cfg_name.empty()) {
		// OK - got a config file - see if we have a commandline specified.
		WCHAR buf[MAX_PATH]={0};
		::GetPrivateProfileString(L"tbzrcache", L"command_prefix", L"", buf, sizeof(buf)/sizeof(buf[0]), cfg_name.c_str());
		cmdline = buf;
		// See if we have command_options specified (the tbzr UI will only ever touch
		// the 'command_options' string, leaving exactly how the executable is found
		// alone.
		buf[0] = 0;
		::GetPrivateProfileString(L"tbzrcache", L"command_options", L"", buf, sizeof(buf)/sizeof(buf[0]), cfg_name.c_str());
		if (buf[0]) {
			cmdline_extra += buf;
		}
		// for convenience, particularly to people trying to run tbzrcache.py from
		// source, allow a pythonpath value to be specified.
		buf[0] = 0;
		::GetPrivateProfileString(L"tbzrcache", L"pythonpath", L"", buf, sizeof(buf)/sizeof(buf[0]), cfg_name.c_str());
		new_py_path = buf;
		// and also to optionally create a new console where tbzrcache can be seen.
		buf[0] = 0;
		::GetPrivateProfileString(L"tbzrcache", L"create_new_console", L"", buf, sizeof(buf)/sizeof(buf[0]), cfg_name.c_str());
		// empty or explicitly '0' means false, otherwise true.
		create_new_console = buf[0] != 0 && buf[0] != '0';
	}
	if (cmdline.empty()) {
		// not specified in the INI - look for 'tbzrcache[w].exe' next to our DLL
		WCHAR buf[MAX_PATH] = {0};
		::GetModuleFileNameW(g_hModule, buf, MAX_PATH);
		cmdline = buf;
		cmdline = cmdline.substr(0, cmdline.rfind(L'\\'));
		cmdline += L"\\tbzrcache.exe";
		if (::GetFileAttributesW(cmdline.c_str())==0xFFFFFFFF) {
            DWORD err = GetLastError();
			TBZR_ERROR("tbzrcache executable '%S' does not exist (%d)\n", cmdline.c_str(), err);
            throw TBzrWin32Exception(err);
		}
	}
	// Save the old PYTHONPATH
	if (!new_py_path.empty()) {
		temp_env_path[0] = 0;
		::GetEnvironmentVariableW(L"PYTHONPATH", temp_env_path, sizeof(temp_env_path)/sizeof(temp_env_path[0]));
		old_py_path = temp_env_path;
		TBZR_NOTE("Set PYTHONPATH to '%S' (was '%S')\n", new_py_path.c_str(), old_py_path.c_str());
		::SetEnvironmentVariableW(L"PYTHONPATH", new_py_path.c_str());
	}
	// So, theoretically we have a cmdline to execute.
	if (!cmdline_extra.empty()) {
		cmdline += L" ";
		cmdline += cmdline_extra;
	}
	TBZR_NOTE("Starting cache process by executing: %S\n", cmdline.c_str());
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);
	DWORD create_flags = create_new_console ? CREATE_NEW_CONSOLE : DETACHED_PROCESS;
	BOOL ok = ::CreateProcess(NULL, (LPWSTR)cmdline.c_str(), NULL, NULL, FALSE, create_flags, NULL, NULL, &si, &pi);
	if (!old_py_path.empty())
		::SetEnvironmentVariableW(L"PYTHONPATH", old_py_path.c_str());
	if (!ok) {
        DWORD err = GetLastError();
		TBZR_ERROR("CreateProcess failed with error %d\n", err);
        throw TBzrWin32Exception(err, "CreateProcess failed.");
	}
	TBZR_MUTTER("CreatedProcess with PID %d\n", pi.dwProcessId);
	::CloseHandle(pi.hProcess);
	::CloseHandle(pi.hThread);
}

// This function should be called in the critical section.
void CRemoteLink::MakeRequest()
{
	// Makes a request, but catches pipe errors which imply a retry will succeed.
	try {
		RawRequest();
    }
    catch (TBzrWin32Exception &e) {
        if (e.m_win32error == ERROR_BAD_PIPE) {
            try {
                TBZR_NOTE("Remote request has ERROR_BAD_PIPE - retrying...\n");
                RawRequest();
            }
            catch (TBzrException &e) {
                TBZR_ERROR("Some error has occured - %s\n", e.what());
				throw e;
            }
        }
        else {
            TBZR_ERROR("Some error has occured - error=0x%x\n", e.m_win32error);
			throw e;
        }
    }
    catch (TBzrException &e) {
        TBZR_ERROR("Some error has occured - %s\n", e.what());
		throw e;
    }
}

// Don't use this function directly. Use MakeRequest().
void CRemoteLink::RawRequest()
{
	// If this fails, an Atl exception is thrown with the win32 error code 
	// (either due to being unable to open it, or  due to a recent error 
	// meaning we don't even try!)
	// This fixed buffer size is a bit of a problem - but removing
	// it would mean not using TransactNamedPipe() and its likely we see
	// better perf by sticking with it.
	try {
		EnsurePipeOpen();
    }
    catch (TBzrException &) {
		// We've failed to open the pipe - try and start the cache
		// but only if the last try to start the cache was a certain time
		// ago. If we just try over and over again without a small pause
		// in between, the explorer is rendered unusable!
		// Failing to start the cache can have different reasons: missing exe,
		// missing registry key, corrupt exe, ...
		DWORD start = ::GetTickCount();
		if (start < m_last_timeout) { // last_timeout is when it *expires*
			TBZR_MUTTER("Remote request last timeout was too recent - skipping\r\n");
            throw TBzrWin32Exception(ERROR_TIMEOUT,
                                     "Skipped because remote request last timeout was too recent.");
		}
		StartRemoteProcess();
		// Wait for the cache to open
		DWORD end_time = ::GetTickCount() + 500;
		for (;;) {
			try {
				EnsurePipeOpen();
				break;
			}
            catch (TBzrException &) {
				DWORD now = GetTickCount();
				if (now > end_time) {
					m_last_timeout = now + m_timeout_ms;
					TBZR_ERROR("Failed to wait for remote process to create the pipe (took %d ms)\r\n",
							   now-start);
                    throw TBzrWin32Exception(ERROR_TIMEOUT,
                        "Failed to wait for remote process to create the pipe.");
				}
				TBZR_MUTTER("sleeping to give the server a chance to create the pipe\r\n");
				Sleep(200);
			}
		}
	}
	// pipe is open - prepare for the transaction.
	OVERLAPPED overlapped;
	memset(&overlapped, 0, sizeof(overlapped));
	overlapped.hEvent = m_hevent;
	// Do the transaction in overlapped mode.
	// That way, if anything happens which might block this call
	// we still can get out of it. We NEVER MUST BLOCK THE SHELL!
	// A blocked shell is a very bad user impression, because users
	// who don't know why it's blocked might find the only solution
	// to such a problem is a reboot and therefore they might loose
	// valuable data.
	// One particular situation where the shell could hang is when
	// the [TSVN crash report dialog comes up]
	// But really, the number of ways things could potentially hang is
	// limited only by your imagination :)
	// Sure, it would be better to have no situations where the shell
	// even can get blocked, but the timeout of 10 seconds is long enough
	// so that users still recognize that something might be wrong and
	// report back to us so we can investigate further.

	// hrmph: ideally, we would have a shorter timeout (2 seconds?), upon
	// which we show better UI feedback - eg, cursor, window with crawling
	// tortoise animation ;) etc.  That would better reflect when it is
	// us causing any "slowdowns" reported.
	DWORD bytesReceived;
	ReqBuffer &b = m_buffer->m_buffer;
	if (!TransactNamedPipe(m_hpipe,
						   b.buffer,  // data to write
						   b.req_header.total_bytes, // number to write
						   b.buffer, // out buffer to fill
						   sizeof(b.buffer), // size of out buffer
						   &bytesReceived, // number actually read
						   &overlapped) && ::GetLastError() != ERROR_IO_PENDING) {
			// XXX - Note that some errors here will simply mean the server
			// process has gone away, so could be retried.
			// Particularly: Failed using the pipe: (230, 'TransactNamedPipe', 'The pipe state is invalid.')
        DWORD err = GetLastError();
        TBZR_ERROR("Failed to TransactNamedPipe - %d\r\n", err);
        ClosePipe();
        throw TBzrWin32Exception(err, "Failed to TransactNamedPipe.");
	}
	if (WAIT_OBJECT_0 == WaitForSingleObject(m_hevent, m_timeout_ms) &&
		GetOverlappedResult(m_hpipe, &overlapped, &bytesReceived, TRUE)) {
		TBZR_MUTTER("Successfully transacted the pipe\r\n");
		// worked - veryify the result.
		if (bytesReceived != (DWORD)b.resp_header.total_bytes) {
			TBZR_ERROR("Mismatch in number of bytes - pipe got %d, header says %d\r\n",
					   bytesReceived, b.resp_header.total_bytes);
			// This implies the pipe-state is screwey,
			ClosePipe();
            throw TBzrWin32Exception(ERROR_INVALID_DATA, "Mismatch in number of bytes.");
		}
		if (b.resp_header.error_code) {
			TBZR_ERROR("Remote server returned error code %d\r\n",
					   b.resp_header.error_code);
			// no need to close the pipe here - the server just barfed for a command
            throw TBzrException("Remote server returned error code.");
		}
		// all worked!!
		return;
	}
	// Failed
	DWORD err = ::GetLastError();
	// closing it is important - we will not retry to open it for a short
	// while, meaning one timeout is all Explorer will see for a while.
	ClosePipe();
	TBZR_ERROR("Failed to read the data from the pipe - %d\r\n", err);
    throw TBzrWin32Exception(err, "Failed to read the data from the pipe.");
}

std::wstring CRemoteLink::GetPipeName()
{
	static std::wstring base(L"\\\\.\\pipe\\TortoiseBZR");
	HANDLE tok;
	if (!::OpenProcessToken(::GetCurrentProcess(), TOKEN_QUERY, &tok)) {
		TBZR_ERROR("Failed to fetch the process token\r\n");
        throw TBzrWin32Exception(GetLastError(), "Failed to fetch the process token.");
	}
	TOKEN_STATISTICS stats;
	DWORD ignoreLen;
	if (!::GetTokenInformation(tok, TokenStatistics, &stats, sizeof(stats), &ignoreLen)) {
		TBZR_ERROR("Failed to GetTokenInformation\r\n");
        throw TBzrWin32Exception(GetLastError(), "Failed to GetTokenInformation.");
	}

	std::wstringstream ret;
	ret << base;
	ret << L"-";
	// iostream doesn't know about a "LUID" but it does about a ULARGE_INTEGER
	ULARGE_INTEGER aid;
	aid.HighPart = stats.AuthenticationId.HighPart;
	aid.LowPart =  stats.AuthenticationId.LowPart;
	ret << aid.QuadPart;
	return ret.str();
}

void CRemoteLink::RequestReleaseHandles(const WCHAR *path)
{
    TBZR_MUTTER("RequestReleaseHandle\n");
    AutoCritical _lock(m_lock);
	m_buffer->ResetForRequest(REQUEST_RELEASE_HANDLES);
	m_buffer->PushString(path);
	MakeRequest();
	// nothing to unpack!
}

DWORD CRemoteLink::RequestFileStatus(const WCHAR *path, DWORD flags)
{
    TBZR_MUTTER("RequestFileStatus\n");
    AutoCritical _lock(m_lock);
	m_buffer->ResetForRequest(REQUEST_FILE_STATUS);
	m_buffer->PushString(path);
	m_buffer->PushInt(flags);
	MakeRequest();
	return (DWORD)m_buffer->PopInt();
	// nothing to unpack!
}

std::vector<TBZRMenuInfo> CRemoteLink::RequestMenuCommands(std::vector<std::wstring> names, DWORD flags)
{
	TBZR_MUTTER("RequestMenuCommands\r\n");
	AutoCritical _autolock(m_lock);
	m_buffer->ResetForRequest(REQUEST_MENU_COMMANDS);
	m_buffer->PushSize(names.size());
	for (size_t i = 0; i < names.size(); ++i)
		m_buffer->PushString(names[i]);
	m_buffer->PushInt(flags);
	MakeRequest();
	std::vector<TBZRMenuInfo> ret;
	// unpack the complicated menu info.
	int num_items = (int)m_buffer->PopSize();
	for (int itemnum = 0; itemnum < num_items; ++itemnum) {
		TBZRMenuInfo info;
		info.flags = (DWORD)m_buffer->PopInt();
		info.level = (int)m_buffer->PopInt();
		info.text = m_buffer->PopString();
		info.long_text = m_buffer->PopString();
		info.ico = m_buffer->PopString();
		// Now the list of 'commands'
		int num_commands = (int)m_buffer->PopSize();
		for (int cmdnum = 0; cmdnum < num_commands; ++cmdnum) {
			TBZRMenuCommand cmd;
			cmd.keystate = (DWORD)m_buffer->PopInt();
			cmd.exec_type = (DWORD)m_buffer->PopInt();
			cmd.cmdline = m_buffer->PopString();
			info.commands.push_back(cmd);
		}
		ret.push_back(info);
	}
	return ret;
}

std::wstring CRemoteLink::RequestGetParam(const WCHAR *param_name)
{
    TBZR_MUTTER("RequestGetParam\n");
    AutoCritical _autolock(m_lock);
	m_buffer->ResetForRequest(REQUEST_GET_PARAM);
	m_buffer->PushString(param_name);
	MakeRequest();
	return m_buffer->PopString();
}

/* vim: set ts=4 sw=4 noexpandtab : */
