// stdafx.h : include file for standard system include files,
// or project specific include files that are used frequently,
// but are changed infrequently

#pragma once

#include <windows.h>
#include <tchar.h>
#include "ShTypes.h"
#include "ShlObj.h"
#include "shlwapi.h"
#include "targetver.h"

#include "resource.h"

#include <stdio.h>
#include <stdarg.h>
#include <exception>

namespace tbzr {

// Some debugging/tracing related functions.
// note that we use "char *", not TCHAR *, so we don't need
// to use _T() wrappers everywhere...
inline void __cdecl TBzrLogNote(LPCSTR pszFormat, ...) throw()
{
	const int LOG_BUFFER_SIZE = 256;
	char chMsg[LOG_BUFFER_SIZE ];
	va_list pArg;

	va_start(pArg, pszFormat);
	_vsnprintf_s(chMsg, LOG_BUFFER_SIZE, LOG_BUFFER_SIZE-1, pszFormat, pArg);
	va_end(pArg);
	chMsg[LOG_BUFFER_SIZE - 1] = 0;
	OutputDebugStringA(chMsg);
	// Maybe use the ATL Module::LogEvent function too?
}

inline void __cdecl TBzrLogNote(LPWSTR pszFormat, ...) throw()
{
	const int LOG_BUFFER_SIZE = 256;
	wchar_t chMsg[LOG_BUFFER_SIZE ];
	va_list pArg;

	va_start(pArg, pszFormat);
	_vsnwprintf_s(chMsg, LOG_BUFFER_SIZE, LOG_BUFFER_SIZE-1, pszFormat, pArg);
	va_end(pArg);
	chMsg[LOG_BUFFER_SIZE - 1] = 0;
	OutputDebugStringW(chMsg);
}

// for now, errors still only go to OutputDebugString
#define TBzrLogError TBzrLogNote

// "mutter" is verbose, and only ever in debug builds.
#ifdef NDEBUG
#define TBZR_MUTTER
#else
#define TBZR_MUTTER TBzrLogNote
#endif

// "note" is enabled in release builds, but only to OutputDebugString()
#define TBZR_NOTE TBzrLogNote // todo - make this better

// "error" always happens, and ideally would complain louder than just
// OutputDebugString - but that is all it does now...
#define TBZR_ERROR TBzrLogError


class CriticalSection {
    CRITICAL_SECTION cs;
    bool locked;
public:
    const char *name;

    CriticalSection(const char *name) : locked(false), name(name) {
        InitializeCriticalSection(&cs);
    }

    ~CriticalSection() {
        if (locked) {
            TBZR_ERROR("CriticalSection::~CriticalSection() %s is deleted without unlock\r\n", name);
            this->Unlock();
        }
        DeleteCriticalSection(&cs);
    }

    void Lock() {
        EnterCriticalSection(&cs);
        locked = true;
    }

    void Unlock() {
        if (!locked) {
            TBZR_ERROR("CriticalSection::Unock() %s unlocked without lock\r\n", name);
            return;
        }
        locked = false;
        LeaveCriticalSection(&cs);
    }
};

struct AutoCritical {
    CriticalSection *cs_;

    AutoCritical(CriticalSection &cs) : cs_(&cs) {
        TBZR_MUTTER("Locking critical section %s ...", cs_->name);
        cs_->Lock();
        TBZR_MUTTER("Locked\r\n");
    }
    ~AutoCritical() {
        TBZR_MUTTER("Unlocking critical section %s\r\n", cs_->name);
        cs_->Unlock();
    }
};

template<typename T>
class AutoComPtr {
    T *ptr;
public:
    AutoComPtr(T* pt) : ptr(pt) {}
    ~AutoComPtr() { ptr->Release(); }
};


struct TBzrException : public std::exception {
    TBzrException() {}
    TBzrException(const char *what) : std::exception(what) {}
};

struct TBzrHresultException : public TBzrException {
    HRESULT m_hr;
    TBzrHresultException(HRESULT hr) : m_hr(hr) {}
    TBzrHresultException(HRESULT hr, const char *what) :
        m_hr(hr), TBzrException(what) {}
};

struct TBzrWin32Exception : public TBzrException {
    DWORD m_win32error;
    TBzrWin32Exception(DWORD dwError) : m_win32error(dwError) {
    }
    TBzrWin32Exception(DWORD dwError, const char *what) :
        m_win32error(dwError), TBzrException(what) {
    }
};

extern HMODULE g_hModule;
extern BOOL g_isExplorer;
extern LONG g_numObject;

} // namespace tbzr;

// The stl
#include <string>
#include <sstream>
#include <vector>
#include <map>


// Our CLSIDs
#include "INITGUID.h"

DEFINE_GUID(CLSID_CopyHook, 0x77450df1, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);

DEFINE_GUID(CLSID_IconOverlayAdded, 0x77450df2, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(CLSID_IconOverlayNormal, 0x77450df3, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(CLSID_IconOverlayModified, 0x77450df4, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(CLSID_IconOverlayConflict, 0x77450df5, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
//DEFINE_GUID(CLSID_IconOverlayLocked, 0x77450df6, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
//DEFINE_GUID(CLSID_IconOverlayReadOnly, 0x77450df7, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(CLSID_IconOverlayDeleted, 0x77450df8, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(CLSID_IconOverlayIgnored, 0x77450df9, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
//DEFINE_GUID(CLSID_IconOverlayInvalid, 0x77450dfa, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(CLSID_IconOverlayUnversioned, 0x77450dfb, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);

DEFINE_GUID(CLSID_ContextMenu, 0x77450dfc, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);

/***
Here is a list of GUIDs available for use:  Please DELETE the UUID from the
list when you use it.  It was generated using 'uuidgen -x -s -n20' then edited

DEFINE_GUID(, 0x77450dfd, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(, 0x77450dfe, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(, 0x77450dff, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(, 0x77450e00, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(, 0x77450e01, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(, 0x77450e02, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
DEFINE_GUID(, 0x77450e03, 0xb5e7, 0x11dd, 0xbc, 0xc2, 0x00, 0x50, 0x56, 0xc0, 0x00, 0x08);
***/

#define TBZR_IUNKNOWN_IMPLEMENT() \
    STDMETHODIMP_(ULONG) AddRef() {\
        InterlockedIncrement(&tbzr::g_numObject);\
        return InterlockedIncrement(&m_refcnt);\
    }\
    STDMETHODIMP_(ULONG) Release() {\
        InterlockedDecrement(&tbzr::g_numObject);\
        ULONG res = InterlockedDecrement(&m_refcnt);\
        if (!res) delete this;\
        return res;\
    }\

using namespace tbzr;
