# Actions - primarily on menus.
import sys
import os
import logging

from tbzrlib.vcspath import VCSPath
from tbzrlib.i18n import (gettext, N_)
from app import get_app


logger = logging.getLogger("bzr.tbzr.commands")

# 'condition states' borrowed from TSVN - the state is used to dictate if
# a given command is enabled or not.
ITEMIS_ONLYONE =                               1<<0
ITEMIS_TWO =                                   1<<1
ITEMIS_FOLDER =                                1<<2
ITEMIS_VERSIONED =                             1<<3
ITEMIS_VERSIONEDFOLDER =                       1<<4
ITEMIS_INVERSIONEDFOLDER =                     1<<5
ITEMIS_UNCHANGED =                             1<<6
ITEMIS_IGNORED =                               1<<7
ITEMIS_ADDED =                                 1<<8
ITEMIS_DELETED =                               1<<9
ITEMIS_CONFLICT =                              1<<10
ITEMIS_PATCHFILE =                             1<<11
ITEMIS_PATCHINCLIPBOARD =                      1<<12
ITEMIS_EXTENDED =                              1<<13
ITEMIS_UNVERSIONED =                           1<<14


# A couple of containers for action definitions and their commands.
class ActionExec(object):
    KEYSTATE_ALL = 0 # or something... :)
    # execution options - most common is via tbzrcommand.exe
    # 0 is reserved for "call back to the server with the cmdline and selection"
    EXEC_TBZRCOMMAND_STD = 1 # std tbzrcommand - --hwnd
    EXEC_TBZRCOMMAND_STD_SELECTION = 2 # std tbzrcommand, plus --file or --filelist.
    def __init__(self, cmdline, exec_options=EXEC_TBZRCOMMAND_STD,
                 keystate=KEYSTATE_ALL):
        self.cmdline = cmdline
        self.exec_options = exec_options
        # only used when the keystate matches.
        self.keystate = keystate

class Action(object):
    def __init__(self, id, commands, text, long_text, ico_file=None,
                 conditions=None):
        self.id = id
        # for convenience, allow a single command to be passed.
        if isinstance(commands, ActionExec):
            commands = [commands]
        self.commands = commands
        self.text = text
        self.long_text = long_text
        self.ico_file = ico_file
        # Conditions are a tuple of (yes, no) bitmasks.  Each of the 'yes'
        # states must be set and none of the 'no' states can be set. Each
        # condition is checked seperately and first to return True wins.
        self.conditions = conditions

    def __repr__(self):
        return "Action(id=%r)" % self.id

def get_tbzr_exec(command, want_sel=True):
    # fist workout how to execute tbzrcommand itself.
    bindir = os.path.dirname(os.path.abspath(sys.executable.decode("mbcs")))
    if hasattr(sys, "frozen"):
        # py2exe
        starter = '"%s"' % os.path.join(bindir, "tbzrcommand.exe")
    else:
        # not frozen - hosted in Python source.
        pythonw = os.path.join(bindir, 'pythonw.exe')
        script_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
        script = os.path.join(script_dir, 'tbzrcommand.py')
        starter = '"%s" "%s"' % (pythonw, script)
    # and put on the rest of the cmdline
    cmdline = '%s --command=%s' % (starter, command)
    if want_sel:
        exec_options = ActionExec.EXEC_TBZRCOMMAND_STD_SELECTION
    else:
        exec_options = ActionExec.EXEC_TBZRCOMMAND_STD
    return ActionExec(cmdline, exec_options)

# all actions, keyed by ID
_actions = {}
# default order for actions
_action_order = []

def get_default_actions():
    return [
        Action("tbzr.newtree", get_tbzr_exec("getnew"),
               N_("Chec&kout/Branch..."), N_("Checkout or Branch a Bazaar branch"),
               "tsvn/menucheckout.ico",
               [(ITEMIS_FOLDER, ITEMIS_VERSIONED|ITEMIS_VERSIONEDFOLDER),
                ]),

        Action("tbzr.commit", get_tbzr_exec("commit"),
               N_("&Commit..."), N_("Commit local changes"),
               "tsvn/menucommit.ico",
               [(ITEMIS_VERSIONED, ITEMIS_CONFLICT),
                (ITEMIS_VERSIONEDFOLDER, ITEMIS_CONFLICT),
                ]),

        Action("tbzr.init", get_tbzr_exec("init"),
               N_("&Init..."), N_("Initialize a Bazaar branch"),
               None,
               [(ITEMIS_FOLDER|ITEMIS_ONLYONE, ITEMIS_VERSIONED|ITEMIS_VERSIONEDFOLDER),
               ]),

        Action("tbzr.explorer", get_tbzr_exec('explorer'),
               N_("&Explorer"), N_("Start Bazaar Explorer"),
               None,
               [(ITEMIS_FOLDER|ITEMIS_ONLYONE, 0),
               ]),

        None, # a sep

        Action("tbzr.browse", get_tbzr_exec("browse"),
               N_("Bro&wse inventory"),
               N_("Opens the inventory browser for the current branch"),
               "tsvn/menurepobrowse.ico",
               [(ITEMIS_VERSIONED|ITEMIS_ONLYONE|ITEMIS_FOLDER, 0),
               ]),

        Action("tbzr.diff", get_tbzr_exec("diff"),
               N_("&Diff"), N_("Show local changes"),
               "tsvn/menucompare.ico",
               # Note: TSVN also excludes folders for some reason - we don't.
               [(ITEMIS_VERSIONED|ITEMIS_ONLYONE, ITEMIS_UNCHANGED),
                (ITEMIS_TWO, None),
                ]),

        # TSVN has "&Diff with previous version", "Diffs the working copy file with the one before the last commit."
        # TSVN: "&Diff with URL", "Compares the selected file with a file in the repository"

        Action("tbzr.log", get_tbzr_exec("log"),
               N_("&Log"), N_("Show a log for the selection"),
               "tsvn/menulog.ico",
               [(ITEMIS_VERSIONED|ITEMIS_ONLYONE, ITEMIS_ADDED),
                (ITEMIS_FOLDER|ITEMIS_VERSIONEDFOLDER, ITEMIS_ADDED),
                (ITEMIS_VERSIONEDFOLDER|ITEMIS_ONLYONE, ITEMIS_ADDED),
               ]),

        Action("tbzr.cat", get_tbzr_exec("cat"),
               N_("Show in &viewer"),
               N_("Show the file in the builtin file viewer"),
               None,
               [(ITEMIS_VERSIONED|ITEMIS_ONLYONE, ITEMIS_FOLDER|ITEMIS_ADDED)
                ]),

        Action("tbzr.annotate", get_tbzr_exec("annotate"),
               N_("A&nnotate"),
               N_("Show the history of each line in the file"),
               "tsvn/menublame.ico",
               [(ITEMIS_VERSIONED|ITEMIS_ONLYONE, ITEMIS_FOLDER|ITEMIS_ADDED)
                ]),

        # TSVN: "Check for modi&fications", "Shows all files which were changed since the last update, locally and in the repository"
        # TSVN: "Revision &graph", "Shows a graphical representation of copies/tags/branches"

        None, # a sep

        # TSVN: "&Edit conflicts", "Launches the external diff / merge program to solve the conflicts"
        # TSVN: "Res&olved...", "Resolves conflicted files"
        # TSVN: "&Update to revision...", "Updates the working copy to a specific revision"
        # TSVN: "Re&name...", "Renames files / folders inside version control"

        Action("tbzr.add", get_tbzr_exec("add"),
               N_("&Add..."), N_("Add item(s) to Bazaar"),
               "tsvn/menuadd.ico",
               [(ITEMIS_INVERSIONEDFOLDER|ITEMIS_UNVERSIONED, 0),
                (ITEMIS_VERSIONED|ITEMIS_FOLDER, 0),
                (ITEMIS_IGNORED, 0),
                (ITEMIS_DELETED, ITEMIS_FOLDER|ITEMIS_ONLYONE),
               ]),

        # TSVN: "&Delete (keep local)", "Deletes files / folders from version control but keeps the file"
        Action("tbzr.remove", get_tbzr_exec("remove"),
               N_("&Delete (keep local)"),
               N_("Deletes files / folders from version control but keeps the file"),
               "tsvn/menudelete.ico",
               [(ITEMIS_VERSIONED | ITEMIS_INVERSIONEDFOLDER, ITEMIS_ADDED),
               ]),

        Action("tbzr.conflicts", get_tbzr_exec("conflicts"),
               N_("&Conflicts"),
               N_("Show conflicts"),
               "tsvn/menuresolve.ico",
               [(ITEMIS_INVERSIONEDFOLDER|ITEMIS_FOLDER, 0),
                (ITEMIS_INVERSIONEDFOLDER|ITEMIS_CONFLICT, 0),
               ]),

        Action("tbzr.revert", get_tbzr_exec("revert"),
               N_("Re&vert..."), N_("Revert local changes"),
               "tsvn/menurevert.ico",
               [(ITEMIS_VERSIONED, ITEMIS_UNCHANGED),
               ]),

        # TSVN: "Undo Add..", "Reverts an addition to version control"
        # TSVN: "Delete unversioned items...", "Removes all unversioned and ignored items"
        # TSVN: "&Clean up", "Cleanup interrupted operations, locked files, ..."
        # TSVN: "Get loc&k...", "Locks a file for other users and makes it editable by you."
        # TSVN: "Release loc&k", "Releases locks on files so other users can edit them again."
        # TSVN: "Break loc&k", "Releases locks even if you're not the owner of the locks (breaks them)."

        None, # a sep

        Action("tbzr.update", get_tbzr_exec("getupdates"),
               N_("&Update..."), N_("Get updates to the working tree"),
               "tsvn/menuupdate.ico",
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
                ]),

        Action("tbzr.uncommit", get_tbzr_exec("uncommit"),
               N_("&Uncommit..."), N_("Move the tip of a branch to an earlier revision"),
               None,
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
               ]),

        Action("tbzr.tag", get_tbzr_exec("tag"),
               N_("&Tag..."),
               N_("Edit tags"),
               None,
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
               ]),

        Action("tbzr.push", get_tbzr_exec("push"),
               N_("&Push..."), N_("Push local revisions to other branch"),
               "tsvn/menucommit.ico",
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
                ]),
        Action("tbzr.pull", get_tbzr_exec("pull"),
               N_("P&ull..."), N_("Pull revisions from other branch"),
               "tsvn/menuupdate.ico",
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
                ]),
        Action("tbzr.send", get_tbzr_exec("send"),
               N_("&Send..."), N_("Mail or create a merge-directive for submitting changes"),
               "tsvn/menucommit.ico",
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
                ]),

        Action("tbzr.merge", get_tbzr_exec("merge"),
               N_("&Merge..."), N_("Merge changes from other branch"),
               "tsvn/menumerge.ico",
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
                ]),
        Action("tbzr.export", get_tbzr_exec("export"),
               N_("&Export..."), N_("Export current or past revision"),
               "tsvn/menuexport.ico",
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
                ]),


        # TSVN: "Branch/&tag...", "Creates a 'cheap' copy inside the repository used for branches or tagging"
        # TSVN: "S&witch...", "Switch working copy to another branch / tag"
        # TSVN: "&Merge...", "Merges a branch into the main trunk"
        # TSVN: "Merge all...", "Merges all missing revisions back into the working copy using merge-tracking information"
        # TSVN: "E&xport...", "Exports a repository to a clean working copy without the svn administrative folders"
        # TSVN: "Relo&cate...", "Use this if the URL of the repository has changed"

        None, # a sep

        # TSVN: "Create repositor&y here", "Creates a repository database at the current location"

        # TSVN: "Add (as replacement)...", "Add (as replacement)..."
        # TSVN: "&Import...", "Imports the directory to a repository"

        # TSVN: "Add to &ignore list", "Adds the selected file(s) or the filemask to the 'ignore' list"
        # TSVN: "Remove from &ignore list", "Removes the file or filemask from the list of ignored items"

        Action("tbzr.bind", get_tbzr_exec("bind"),
               N_("Bind..."),
               N_("Convert the current branch into a checkout of the supplied branch."),
               None,
               [(ITEMIS_INVERSIONEDFOLDER|ITEMIS_FOLDER|ITEMIS_ONLYONE, 0),
               ]),

        Action("tbzr.unbind", get_tbzr_exec("unbind"),
               N_("Unbind..."),
               N_("Convert the current checkout into a regular branch"),
               None,
               [(ITEMIS_INVERSIONEDFOLDER|ITEMIS_FOLDER|ITEMIS_ONLYONE, 0),
               ]),

        Action("tbzr.switch", get_tbzr_exec("switch"),
               N_("&Switch..."),
               N_("Set the branch of a checkout and update."),
               'tsvn/menuswitch.ico',
               [(ITEMIS_INVERSIONEDFOLDER|ITEMIS_FOLDER|ITEMIS_ONLYONE, 0),
               ]),

        Action("tbzr.info", get_tbzr_exec("info"),
               N_("&Info"),
               N_("Shows information about current location."),
               None,
               [(ITEMIS_FOLDER|ITEMIS_INVERSIONEDFOLDER|ITEMIS_ONLYONE, 0),
               ]),

        None, # a sep

        # TSVN: "Create &patch...", "Creates a unified diff file with all changes you made"
        # TSVN: "Apply &patch...", "Applies a unified diff file to the working copy"

        Action("tbzr.refresh", get_tbzr_exec("refresh"),
               N_("&Refresh"), N_("Refresh status icons for this tree"),
               'thg/refresh_overlays.ico',
               [(ITEMIS_FOLDER|ITEMIS_ONLYONE, 0),
               ]),
        Action("tbzr.run", get_tbzr_exec("run"),
               N_("&Run command"), N_("Run Bazaar command"),
               None,
               [(ITEMIS_FOLDER|ITEMIS_ONLYONE, 0),
               ]),
        Action("tbzr.help", get_tbzr_exec("help", want_sel=False),
               N_("&Help"), N_("Tortoise Bazaar Help"), "tsvn/menuhelp.ico"),
        Action("tbzr.settings", get_tbzr_exec("settings", want_sel=False),
               N_("&Settings"), N_("Tortoise Bazaar Settings"),
               "tsvn/menusettings.ico"),
        Action("tbzr.about", get_tbzr_exec("about", want_sel=False),
               N_("A&bout"), N_("About Tortoise Bazaar"),
               "tsvn/menuabout.ico"),
    ]

def _build_actions():
    assert not _actions and not _action_order
    for a in get_default_actions():
        if a is not None:
            _actions[a.id] = a
        _action_order.append(a) # even Nones get in here!

def get_action(action_id):
    if not _actions:
        _build_actions()
    try:
        return _actions[action_id]
    except KeyError:
        raise ValueError(repr(action_id))

def _filtered_actions(flags):
    ret = []
    for act in _action_order:
        if act is None: # a sep
            result = True
        elif act.conditions:
            result = False
            for yes, no in act.conditions:
                if yes and no:
                    if yes & flags == yes and (no & ~flags) == no:
                        result = True
                        break
                elif yes:
                    if yes & flags == yes:
                        result = True
                        break
                elif no:
                    if (no & ~flags) == no:
                        result = True
                        break
                else:
                    assert False, "can't both be False!"
        else:
            result = True
        if result:
            ret.append(act)
    return ret

def _get_flag(filenames):
    app = get_app()
    dir_cache = app.dir_cache

    flags = 0
    if len(filenames)==1:
        flags |= ITEMIS_ONLYONE
    elif len(filenames)==2:
        flags |= ITEMIS_TWO
    for raw_path in filenames:
        path = VCSPath(raw_path)
        if not app.is_path_allowed(path):
            return -1 # show all flags

        status = dir_cache.get_status_for_path(path, 1)
        if status.is_progress():
            return -1 # show all flags

        is_versioned = status.is_versioned()
        if not status.is_nothing():
            flags |= ITEMIS_INVERSIONEDFOLDER
        if is_versioned:
            flags |= ITEMIS_VERSIONED
            # must usual case us unchanged - test that first.
            if status.is_changed():
                if status.is_modified():
                    pass # hrm - no flag for this??
                elif status.is_added():
                    flags |= ITEMIS_ADDED
                elif status.is_deleted():
                    flags |= ITEMIS_DELETED
                elif status.is_conflicted():
                    flags |= ITEMIS_CONFLICT
                else:
                    assert False, status # missed a state!
            else:
                flags |= ITEMIS_UNCHANGED
        elif status.is_unversioned():
            flags |= ITEMIS_UNVERSIONED
        elif status.is_ignored():
            flags |= ITEMIS_IGNORED
        else:
            # not in a versioned folder.
            pass

        # ack - the below is wrong in some cases?  Are we sure its a
        # VERSIONEDFOLDER?  What about ITEMIS_INVERSIONEDFOLDER?
        if path.is_directory:
            flags |= ITEMIS_FOLDER
            if flags & is_versioned:
                flags |= ITEMIS_VERSIONEDFOLDER
    return flags

def get_actions_for_files(filenames):
    if not _actions:
        _build_actions()

    if get_app().sleeping:
        ret = list(_action_order)
    else:
        # flags that define the specified items
        flags = _get_flag(filenames)

        # if success to get flag, filter actions by the flag.
        if flags < 0:
            ret = list(_action_order)
        else:
            ret = _filtered_actions(flags)

    # We still need to implement a user-preference for which commands go
    # on which menu, but for now we stick with TSVN's defaults, which
    # boil down to all items before the first sep go on the top-level.
    # We also probably don't need arbitrary depths - so just return "toplevel,
    # sub" for now.
    actions = [None] # always start with a sep
    sub_actions = []
    filling = actions
    for action in ret:
        # if first sep, just swap
        if action is None and filling is actions: # first sep!
            filling = sub_actions
        else:
            filling.append(action)
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("actions for %s are %s/%s",
                     filenames,
                     [a.id for a in actions if a],
                     [a.id for a in sub_actions if a])
    return actions, sub_actions
