# This need an abstract base class with these methods.  This class captures
# both singletons and bzr specific code.
import os
import time
import logging

import bzrlib.workingtree
from bzrlib.trace import note
from bzrlib.config import config_dir

import tbzrlib.bzrstatus as bzrstatus
import tbzrlib.i18n as i18n

from vcspath import VCSPath

import win32api
import win32event
import win32con

from pipe.server import PipeServer
from config import Config

logger = logging.getLogger("bzr.tbzr.app")

class StartupError(Exception):
    pass


# The core application object.  An App object is a singleton within a process
# and can be fetched with get_app().  This app will be "initialized", but not
# "running".
# A "running App object is a singleton for the entire system (for the user
# anyway).  It is protected by a global mutex.  Trying to set a start a
# second App object will fail.
# An App object that is not running doesn't have much - no taskbar, no
# command dispatcher, etc.
class App(object):
    def __init__(self):
        self._uid = None
        self.watcher = None
        self.dir_cache = None
        self.hmutex = None
        self.config = None
        self.command_dispatcher = None
        self.config_notifiers = [] # list of callables.
        # Note this must be a "manual reset" event (the second arg) as more
        # than 1 thing waits on it, but when signalled we want them all to
        # wake.  Initial state (3rd arg) is unset.
        self.stop_event = win32event.CreateEvent(None, 1, 0, None)
        self.config_event_name = "tbzr-config-change-%s" % (self._get_uid(),)
        self.config_event = None
        self.refresh_event = None
        self._sleeping = False

        import bzrlib.plugin
        import bzrlib.plugins
        bzrlib.plugin.set_plugins_path()

        self.loaded_plugins = set()
        wanted_plugins = "symlink loom pipeline explorer".split()
        for p in wanted_plugins:
            try:
                __import__('bzrlib.plugins.' + p, globals(), locals(), [])
                self.loaded_plugins.add(p)
            except Exception:
                pass

    def get_sleeping(self):
        return self._sleeping

    def set_sleeping(self, next_sleeping):
        now_sleeping = self._sleeping
        self._sleeping = next_sleeping
        if next_sleeping and not now_sleeping:
            self.dir_cache.stop()
            self._kill_watcher()
        elif not next_sleeping and now_sleeping:
            self.dir_cache.restart()
            self._start_watcher()

    sleeping = property(get_sleeping, set_sleeping)

    def _get_uid(self):
        # XXX - duplicated in pipe/__init__.py: might end up needing
        # refactoring if terminal services 'Global\\' etc prefixes become
        # an issue - in which case we should remove this method and replace
        # with something which returns a 'fully-qualified' name given a
        # base name.
        if self._uid is None:
            import win32process, win32security
            tok = win32security.OpenProcessToken(win32process.GetCurrentProcess(),
                                                 win32security.TOKEN_QUERY)
            info = win32security.GetTokenInformation(tok,
                                                     win32security.TokenStatistics)
            self._uid = info['AuthenticationId']
        return self._uid

    def get_config_filename(self):
        # get the location of the config file - same as bzr itself.
        return os.path.join(config_dir(), "tbzr.conf")

    def _create_windows_event(self, event_name):
        import win32event, win32api, winerror
        try:
            # you would think we only need EVENT_MODIFY_STATE here, but it
            # seems (on Vista x64 at least) that should the original creating
            # process terminate, the handle that remains open due to us can't
            # be synchronized. To verify: 1) change to EVENT_MODIFY_STATE, 2)
            # execute 'settings' from the taskbar and keep dialog open, 3)
            # terminate tbzrcache, 4) attempt to start tbzrcache - fails with
            # ACCESS_DENIED in WaitForMultipleObjects in the main run() loop.
            # 5) close dialog, app again starts normally
            event = win32event.OpenEvent(win32event.EVENT_ALL_ACCESS, False,
                                         event_name)
        except win32event.error, exc:
            if exc[0] == winerror.ERROR_FILE_NOT_FOUND:
                event = win32event.CreateEvent(None, True, False, event_name)
            else:
                event = None
                logger.error("Error opening event %r: %s", event_name, exc)
        return event

    def create_config_event(self):
        """A named Windows event is used to communicate a config change.

        This event is used because the application changing the config (eg,
        'tbzrcommand') may not be the same application using them (ie, the
        'tbzrcache' process.)
        """
        import win32event, win32api, winerror
        self.config_event = self._create_windows_event(self.config_event_name)

    def create_refresh_event(self):
        """A named Windows event is used to notify cache reliability is broken.

        This event is used because tbzr's cache is not correct in some situation.
        For example, 'init' in directory cached as "not in workingtree"
        """
        import win32event, win32api, winerror
        event_name = "tbzr-cache-refresh-%s" % (self._get_uid(),)
        self.refresh_event = self._create_windows_event(event_name)

    def create_app_mutex(self):
        # avoid polluting the global namespace.
        import win32event, win32api, winerror
        uid = self._get_uid()
        mutex_name = "tbzr-server-%s" % (uid,)
        try:
            hmutex = win32event.CreateMutex(None, True, mutex_name)
            if win32api.GetLastError() != winerror.ERROR_ALREADY_EXISTS:
                self.hmutex = hmutex
                return True
            logger.debug("Mutex '%s' already exists", mutex_name)
        except win32event.error, exc:
            logger.debug("Error opening app mutex '%s': %s", mutex_name, exc)
            if exc[0] != winerror.ERROR_ALREADY_EXISTS:
                raise
        return False

    def is_path_allowed(self, path):
        """\
        Indicates if a path is allowed in the cache. TSVN has a user-pref
        indicating of removable disks are allowed, and other handling for
        UNC paths etc, and also checks for general validity.  So we do too :)
        """

        if path.is_matched(self.config.get("icon_overlays.exclude_path")):
            return False

        config_name = "queried_drives." + path.get_device_type()
        return (self.config.get(config_name) or
                path.is_matched(self.config.get("icon_overlays.include_path")))

    def initialize(self):
        i18n.install()
        self.on_activity()
        self.command_dispatcher = None # created if we run()

        self.config = Config(self.get_config_filename(), self.on_config_change)
        self.create_config_event()
        self.create_refresh_event()
        # we create a dircache here for the sake of the test suite etc - it's
        # safe as we don't persist anything.  If we move to a persistant cache
        # we would probably want to move this to 'run'.
        #from cache.statuscache import StatusCache
        # hrm - it looks like we may as well pass the app itself!
        from wtcache import WtCache
        self.dir_cache = WtCache(self)

    # A kind of "event notification" system: we are called back with an
    # 'on_{event}' method when something of interest happens. We may then
    # broadcast the event of interest to any number of things which have
    # registered via 'register_notify_{event}'
    def on_activity(self):
        """A callback for workers to note they saw some activity.  Used for
        the 'exit after n minutes of inactivity' feature.
        """
        self.last_activity_tick = win32api.GetTickCount()

    def on_config_change(self):
        """Called by code which changes the config.  Indirectly causes
        all the config_notifiers to be called.
        """
        win32event.PulseEvent(self.config_event)
        # the event will cause our run loop to wake and do the actual
        # reload then call the notifications...

    def do_refresh_cache(self):
        win32event.PulseEvent(self.refresh_event)

    def register_notify_config_change(self, callback):
        """Register a function to be called when the config changes"""
        self.config_notifiers.append(callback)

    # Support for stopping the app.
    def is_stopping(self):
        rc = win32event.WaitForSingleObject(self.stop_event, 0)
        return rc != win32event.WAIT_TIMEOUT

    def stop(self):
        """The canonical way of shutting down the server.
        """
        # NOTE: might be called somewhat "reentrantly" - ie, setting stop_event
        # might trigger other things to stop (such as the pipe thread), which
        # itself calls this stop() method to ensure the app is dieing.
        self._kill_watcher()
        win32event.SetEvent(self.stop_event)

    def calc_idle_shutdown_ms(self):
        # calculate how many ms until we should shutdown due to idle
        # inactivity.  Returns -1 (win32event.INFINITE) if idle shutdown is
        # disabled, None if the shutdown time has passed, or a +ve number of
        # ms.
        idle_timeout = self.config.get('tbzrcache.idle_timeout')
        if idle_timeout <= 0:
            return win32event.INFINITE
        else:
            # idle_timeout is actually a float for fine control
            stop_tick = int(self.last_activity_tick + idle_timeout * 60000)
            now_tick = win32api.GetTickCount()
            if now_tick >= stop_tick:
                return None
            return stop_tick - now_tick

    def _start_watcher(self):
        self._kill_watcher()

        from watcher import DirectoryWatcher
        self.watcher = DirectoryWatcher(self.dir_cache,
                                        self.taskbar.get_hwnd())
        self.watcher.daemon = True
        self.watcher.start()

    def _kill_watcher(self):
        if self.watcher is not None:
            self.watcher.stop()
            self.watcher = None

    # the main app run loop.
    def run(self):
        # Only one app can run at a time.
        if not self.create_app_mutex():
            raise StartupError("Another instance of this server is running - please "
                               "close that instance and try again")
        # now create our more-serious objects.
        # avoid circular imports...
        from dispatcher import CommandDispatcher
        self.command_dispatcher = CommandDispatcher()

        # TODO: follow the below comment.
        #  All of our workers are threads, and each of them monitors our 'stop'
        #  event and auto closes.  Also, if any of them terminates or dies, they
        #  often *set* the stop event, allowing the app to be auto-restarted and
        #  everything (hopefully) restored to goodness.

        from ui.taskbar import Taskbar
        taskbar = Taskbar(self)
        taskbar.daemon = True
        taskbar.start()
        self.taskbar = taskbar

        self._start_watcher()

        pipeserver = PipeServer(self)
        pipeserver.daemon = True
        pipeserver.start()

        # using 'note' here ensures it goes to stdout and bzr.log in all cases.
        note("tbzrcache running...")
        events = [self.stop_event]

        config_index = refresh_index = None
        if self.config_event is not None:
            config_index = len(events)
            events.append(self.config_event)
        if self.refresh_event is not None:
            refresh_index = len(events)
            events.append(self.refresh_event)

        while 1:
            # calculate the time to sleep until we need to shutdown.  Its
            # likely that when we wake we will find we must remain alive due
            # to activity in the meantime...
            sleep_time = self.calc_idle_shutdown_ms()
            if sleep_time is None:
                logger.debug("application shutting down due to inactivity")
                break

            logger.debug("application run loop sleeping for %d ms.", sleep_time)
            rc = win32event.WaitForMultipleObjects(events, False, sleep_time)
            if rc == win32event.WAIT_OBJECT_0:
                note("tbzrcache terminating")
                break
            elif config_index is not None and rc == win32event.WAIT_OBJECT_0+config_index:
                note("config change detected")
                self.config.reload()
                self.dir_cache.clear_cache()
                for fn in self.config_notifiers:
                    try:
                        fn()
                    except Exception:
                        logger.exception("Failed to call config notification %s", fn)
            elif refresh_index is not None and rc == win32event.WAIT_OBJECT_0+refresh_index:
                note("clearing cache")
                self.dir_cache.clear_cache()
            # and continue around the loop with the updated config changes.
        self.stop() # incase our stop-event isn't already set...

        # wait for everything to shut down
        workers = [pipeserver, taskbar, self.watcher]
        for w in workers:
            if w is None:
                continue
            w.join(5)
            if w.isAlive():
                logger.warning("Stopping while %s workers are alive", w)
        logger.debug("application terminating")

    # Factories for creating concrete VCS states.  These need rethinking (eg,
    # maybe they could become singletons if the status object was immutable?)
    def vcs_status_modified_factory(self):
        # cacheddirectory sometimes needs to set a directory as explicitly
        # 'modified' if, eg, a child has been added.
        return bzrstatus.Modified
    def vcs_status_unchanged_factory(self):
        # cacheddirectory sometimes needs to set a directory as explicitly
        # 'modified' if, eg, a child has been added.
        return bzrstatus.Unchanged
    def vcs_status_unversioned_factory(self):
        # Sometimes cacheddirectory optimizes the unversioned state...
        return bzrstatus.Unversioned
    def vcs_status_ignored_factory(self):
        # Sometimes cacheddirectory optimizes the unversioned state...
        return bzrstatus.Ignored
    def vcs_status_nothing_factory(self):
        return bzrstatus.Nothing
    def vcs_status_progress_factory(self):
        return bzrstatus.Progress
    def vcs_status_conflict_factory(self):
        return bzrstatus.Conflict

_app = None

def get_app():
    global _app
    if _app is None:
        _app = App()
        _app.initialize()
    return _app
