# coding: utf-8

__all__ = ('BZRStatus', 'from_bzr_change_tuple')

(_STATE_NOTHING, 
 _STATE_UNVERSIONED,
 _STATE_IGNORED,
 _STATE_UNCHANGED,
 _STATE_PROGRESS,
 _STATE_MODIFIED,
 _STATE_ADDED,
 _STATE_DELETED,
 _STATE_MISSING,
 _STATE_CONFLICT,
 ) = range(10)

_state_names = { # for debugging...
    _STATE_NOTHING: 'nothing',
    _STATE_PROGRESS: 'progress',
    _STATE_IGNORED: 'ignored',
    _STATE_UNCHANGED: 'unchanged',
    _STATE_MODIFIED: 'modified',
    _STATE_ADDED: 'added',
    _STATE_DELETED: 'deleted',
    _STATE_MISSING: 'missing',
    _STATE_CONFLICT: 'conflict',
    _STATE_UNVERSIONED: 'unversioned',
}


class BZRStatus(object):

    __slots__ = ('_state')

    # These states are mutually exclusive - the UI can only show 1 such state at
    # a time. A state of 0 means "nothing" - ie, not of interest to us - ie,
    # an unversioned folder not under the root of a working tree.
    # Our impl of most_important() assumes the higher the number, the "more
    # important" the state.
    def __init__(self, state=0):
        self._state = state

    def __repr__(self):
        return "%s(state=%r)" % (self.__class__.__name__,
                                 _state_names.get(self._state, self._state))

    @property
    def state(self):
        return self._state

    def __eq__(self, other):
        return self._state == other.state

    def most_important(self, other):
        if other is None or self._state >= other._state:
            return self
        return other

    def is_nothing(self):
        return self._state == _STATE_NOTHING

    def is_versioned(self):
        """Return True if the file is being managed by Bzr, or False if the
        file is ignored, unversioned(unknown) or outside a Bzr tree.
        """
        return self._state not in (_STATE_NOTHING,
                                   _STATE_UNVERSIONED,
                                   _STATE_IGNORED)

    def is_progress(self):
        return self._state == _STATE_PROGRESS

    def is_unversioned(self):
        """Return True if the file file specifically has a state of
           UNVERSIONED - ie, what Bzr would call 'unknown', or False if
           ignored, added, etc"""
        return self._state == _STATE_UNVERSIONED

    def is_modified(self):
        return self._state == _STATE_MODIFIED

    def is_added(self):
        return self._state == _STATE_ADDED

    def is_deleted(self):
        return self._state == _STATE_DELETED

    def is_ignored(self):
        return self._state == _STATE_IGNORED

    def is_missing(self):
        return self._state == _STATE_MISSING

    def is_conflicted(self):
        return self._state == _STATE_CONFLICT

    def is_changed(self):
        """Returns True if the item can be considered 'changed' in any way.

        An item is considered 'changed' if it is modified, added, missing,
        etc
        """
        return (self._state >= _STATE_MODIFIED)

(Nothing,
 Unversioned,
 Ignored,
 Unchanged,
 Progress,
 Modified,
 Added,
 Deleted,
 Missing,
 Conflict,
 ) = [BZRStatus(i) for i in xrange(10)]

def from_bzr_change_tuple(tree, info):
    conflicts = set(c.path for c in tree.conflicts())
    (file_id, path, content_change, versioned, parent_id, name,
     kind, executable) = info

    if path[1] == u'':
        return Unchanged

    if path[1] in conflicts:
        return Conflict

    if versioned == (False, False):
        # unknown
        if tree.is_ignored(path[1]):
            return Ignored
        else:
            return Unversioned
    elif versioned == (False, True):
        return Added
    elif versioned == (True, False) or (kind[0] is not None and kind[1] is None):
        return Deleted
    elif content_change and (kind[1] is not None):
        # Versioned and changed, but not deleted
        return Modified
    # XXX - any more?
    else:
        return Unchanged
