import sys
from bzrlib.config import ConfigObj
from tbzrlib.i18n import (N_, gettext)
import logging

logger = logging.getLogger('bzr.tbzr')

# Its difficult to reuse bzrlib.config, as the base 'Config' class defines
# a number of options (eg, get_editor).
# So we do our best to stay within the spirit of it :)  Kinda ;)
# Note that the 'settings' dialog uses the Opt class - doc appears
# as the text for a checkbox, help_text appears as tooltip text, for example.
class Opt(object):
    def __init__(self, default, doc, help_text=None):
        self.default = default
        # NOTE: consumer of these expected to gettext() them.
        self.doc = doc
        self.help_text = help_text
    def encode(self, val):
        raise NotImplementedError
    def decode(self, val):
        raise NotImplementedError

class BoolOpt(Opt):
    def decode(self, val):
        if val.lower() in ["1", "true", "yes"]:
            return True
        if val.lower() in ["0", "false", "no"]:
            return False
        raise ValueError(repr(val))

    def encode(self, val):
        return str(not not val)

class FloatOpt(Opt):
    def decode(self, val):
        return float(val) # raises ValueError all by itself.

    def encode(self, val):
        return str(val)

class StringOpt(Opt):
    def decode(self, val):
        return unicode(val)

    def encode(self, val):
        return unicode(val)

# A "data dictionary"
config_config = {
    'tbzrcache' : {
        # NOTE: The C++ shell extension also uses the [tbzrcache] section to
        # specify the command-line and params used to execute tbzrcache
        # itself. tbzrcache itself obviously doesn't need to make use of these
        # values, so they don't exist here, but it uses values of
        # 'command_prefix', 'command_options', 'create_new_console' and
        # 'pythonpath'
        'idle_timeout' :
            #NOTE: config doc for idle_timeout. (shown as checkbox text)
            FloatOpt(60.0, N_("Idle timeout to tbzrcache terminates"),
                     # NOTE: config help for idle_timeout. (shown as popup help).
                     N_("Number of minutes of inactivity after which the tbzrcache process will terminate, or zero to never automatically terminate."))
    },
    'icon_overlays' : {
        'show_as_mod_if_unversioned' :
            BoolOpt(False, N_("Unversioned files mark parent folder as modified"),
                    N_("If set, then unversioned files will mark the status of the parent directory as modified - ie, they will get the modified icon.")),
        'show_ignored' :
            BoolOpt(True, N_("Show overlay for ignored items"),
                    N_("If set, ignored items get a special icon overlay")),
        'show_only_in_explorer' :
            BoolOpt(False, N_("Show overlay and context menues only in explorer"),
                    N_("If set, overlay and context menues are shown only in explorer")),
        'include_path' :
            StringOpt("", N_("Include path"),
                    N_("Specify folder paths (separated by newline) in which tbzr will show icon overlay.\n"
                       "You can use wildcard here.(*, ?, [abc], [!abc])")),
        'exclude_path' :
            StringOpt("", N_("Exclude path"),
                    N_("Specify folder paths (separated by newline) in which tbzr will avoid showing icon overlay.\n"
                       "You can use wildcard here.(*, ?, [abc], [!abc])")),
    },
    'queried_drives' : {
        'removable' :
            BoolOpt(False, N_("Removable devices")),
        'floppy' :
            BoolOpt(False, N_("Floppy Drives")),
        'cdrom' :
            BoolOpt(False, N_("CD and DVD Drives")),
        'fixed' :
            BoolOpt(True, N_("Fixed disks")),
        'remote' :
            BoolOpt(False, N_("Network Drives")),
        'unknown' :
            BoolOpt(False, N_("Unknown Drives")),
    },
    'task_bar' : {
        'always_show_icon' :
            BoolOpt(True, N_("Show an icon in the taskbar when tbzrcache is running"),
                    N_("If set, then an icon will be created in the taskbar, providing a menu for shutting down the application, displaying preferences, etc.")),
    },
}

class Config(object):
    def __init__(self, filename, config_change_callback):
        logger.debug("Config loading from file %r", filename)
        self._parser = ConfigObj(filename, encoding='utf-8')
        # ensure all sections are setup.
        for sect_name in config_config.iterkeys():
            if self._parser.get(sect_name) is None:
                self._parser[sect_name] = {}
        self._value_cache = {} # cache of already read and validated config options.
        # xxx - now what?

        # XXX - move this option to a real option....
        # values must come from VCSPath - default is only 'fixed' devices
        self.allowed_device_types = set(("fixed",))
        self.config_change_callback = config_change_callback

    def write(self):
        self._parser.write()
        # tell the app the change was made.
        self.config_change_callback()

    def reload(self):
        logger.info("reloading config values")
        self._value_cache.clear()
        self._parser.reload()

    def dump(self, tofile):
        # XXX - this will need fixing if/when we have true nested sections etc
        for key, items in config_config.iteritems():
            print >> tofile, "[%s]" % key
            for item_name, item_def in items.iteritems():
                full_name = key + "." + item_name
                cur = self.get(full_name)
                if cur == item_def.default:
                    def_str = "(default)"
                else:
                    def_str = "(customized)"
                print >> tofile, "%s = %s # %s %s" % (item_name, cur, def_str,
                                                      item_def.doc)

    def get(self, config_name):
        try:
            return self._value_cache[config_name]
        except KeyError:
            pass
        val = self.read_value(config_name)
        self._value_cache[config_name] = val
        return val

    def get_opt(self, config_name):
        # Get the option object (ie, metadata) for the option name.
        # Raises ValueError for an invalid option.

        # split on dotted names.
        bits = config_name.split('.')
        assert len(bits) > 1, "config value must have at least 1 section name"
        item_def = config_config
        for bit in bits:
            try:
                item_def = item_def[bit]
            except KeyError:
                raise ValueError(config_name)
        assert hasattr(item_def, "default"), "not a valid leaf option name"
        return item_def

    def _get_sect_for_opt(self, config_name):
        bits = config_name.split('.')
        val = self._parser
        for bit in bits[:-1]:
            val = val[bit]
        return val, bits[-1]

    def set(self, name, value):
        item_def = self.get_opt(name)
        sect, tail = self._get_sect_for_opt(name)
        write_value = item_def.encode(value)
        sect[tail] = write_value
        # and set in our cache.
        self._value_cache[name] = value

    def read_value(self, config_name):
        item_def = self.get_opt(config_name)
        sect, tail = self._get_sect_for_opt(config_name)
        try:
            val = sect[tail]
        except KeyError:
            # config file doesn't have that option - no prob
            ret = item_def.default
            logger.debug("config option '%s' is using default of '%s'",
                         config_name, ret)
            return ret

        try:
            ret = item_def.decode(val)
            logger.debug("config option '%s' read value '%s'", config_name, ret)
            return ret
        except ValueError, why:
            logger.warning("config option '%s' is invalid: %s", config_name, why)
            return item_def.default
