# This is the coder and decoder for the wire data used by TBZR.
import sys
import os
import logging
import struct
import tbzrlib
from tbzrlib.vcspath import VCSPath
from tbzrlib.i18n import gettext

from app import get_app
from actions import get_actions_for_files

logger = logging.getLogger("bzr.tbzr.commands")

is_frozen = getattr(sys, "frozen", None) is not None

(
REQUEST_HELLO,
REQUEST_GOODBYE,
REQUEST_RESTART,
REQUEST_FILE_STATUS,
REQUEST_MENU_COMMANDS,
REQUEST_RELEASE_HANDLES,
REQUEST_REFRESH_STATUS,
REQUEST_GET_PARAM,
) = range(8)

# Exceptions related to request/response failures.
class RequestFailed(Exception):
    pass

# A few trivial commands - todo - move somewhere else...
def hello(sval):
    logger.info("Hello request: %s", sval)
    return u"I'm a little teapot"

def goodbye(sval):
    logger.info("Shutdown request: %s", sval)
    get_app().stop()
    return u"thanks for all the fish"

def get_file_status(sval, flags):
    # todo: more flags than recursive??
    if get_app().sleeping:
        logger.info('skip get_file_status msg because of sleeping')
        return 0
    logger.debug("get_file_status fetching status for %r (flags=%d)", sval, flags)
    path = VCSPath(sval)
    state = get_app().dir_cache.get_status_for_path(path, flags != 0)
    ret = state.state
    logger.debug("status for %s is %s", path, ret)
    # TSVN has user prefs which allow some states to not be
    # shown (eg, TSVN has a special overlay for 'unversioned',
    # but defaults to *not* showing it!)

    # use "unversioned" for out of cache object.
    if ret == 2: # ignored
        if not get_app().config.get("icon_overlays.show_ignored"):
            ret = 0
    return ret

def release_handles(path):
    logger.debug("release_handles for %r", path)
    app = get_app()
    watcher = app.watcher
    path = VCSPath(path).long_path()
    ret = watcher.remove_path_and_children(path)
    watcher.block_path(path)
    app.dir_cache.block_path(path) #TODO: move block_path to app.
    logger.debug("result for release_handles(%s) is %r", path, ret)
    return ret

def refresh_status(path):
    logger.debug(u"refresh_status for %r", path)
    cache = get_app().dir_cache
    path = VCSPath(path).directory
    cache.remove_cache_for_path(path)
    if not get_app().sleeping:
        cache.get_status_for_path(path)
    return 0

sys_params = { "config_event_name" : get_app().config_event_name }

def get_param(name):

    logger.debug(u"get_param for %r", name)
    try:
        param_type, param_name = name.split(':', 1)
        if param_type == "system":
            return unicode(sys_params.get(param_name, ""))
        elif param_type == "config":
            return unicode(get_app().config.get(param_name)) or u""
        else:
            return u""
    except ValueError:
        return u""
        
_icon_map = {} # used only in non-frozen builds.
_icon_module = None

def fixup_ico(ico_name):
    global _icon_module
    if not ico_name:
        return u""

    if is_frozen and not _icon_map:
        # We (hopefully!) have a special resource in our executable.
        import win32api
        import cPickle as pickle
        _icon_module = win32api.GetModuleFileNameW(0)
        try:
            data = win32api.LoadResource(None, "ICON_MAP", 1)
            _icon_map.update(pickle.loads(data))
        except (win32api.error, pickle.PickleError), exc:
            logger.warning("Failed to load icon map from executable: %s", exc)
            # make sure the map has something; bzr is always 0.
            _icon_map['bzr.ico'] = 0
    try:
        ico_index = _icon_map[ico_name]
    except KeyError:
        pass
    else:
        return _icon_module + "#" + str(ico_index)

    # Either in source-code, or an icon not packaged with the binaries.
    if ico_name == 'bzr.ico':
        # running from source-code - see what we can find.
        import bzrlib
        _bzr_ico = os.path.abspath(
                     os.path.join(bzrlib.__path__[0], "..", u"bzr.ico"))
        return _bzr_ico

    # It must be a filename - in source code builds it is relative to
    # tbzrlib.
    if not is_frozen and not os.path.isabs(ico_name):
        ico_name = os.path.join(os.path.abspath(tbzrlib.__path__[0]),
                                "resources", ico_name)
    assert os.path.isfile(ico_name), ico_name # invalid icon!
    return ico_name.decode("mbcs")


# bitmask of states =  disabled, checked, etc.
# XXX - DISABLED etc not used: disabled menu items are "bad form" and there
# are no use cases for "checked" items yet - kill this?
MENU_FLAG_NORMAL = 0
MENU_FLAG_DISABLED = 1

def get_menu_commands(filenames, flags):
    #ignore special shellextension objects like 'My Computer', 'Trash'.
    filenames = [f for f in filenames if not f.startswith(u'::')]

    actions, subs = get_actions_for_files(filenames)
    # now convert this into the menu information.
    depth = 0
    items = []
    for a in actions:
        # Anything on the top-level gets a "BZR" prefix auto added.
        if a is None:
            item = (0, depth, u"", u"", u"", [])
        else:
            text = u"Bazaar " + gettext(a.text)
            commands = [(c.keystate, c.exec_options, c.cmdline) for c in a.commands]
            item = (MENU_FLAG_NORMAL, depth, text, gettext(a.long_text),
                    fixup_ico(a.ico_file), commands)
        items.append(item)
    # and children
    depth += 1
    last = None
    for a in subs:
        if a is None:
            if last is None:
                continue # avoid dupe seps
            item = (0, depth, u"", u"", u"", [])
        else:
            commands = [(c.keystate, c.exec_options, c.cmdline) for c in a.commands]
            item = (MENU_FLAG_NORMAL, depth, gettext(a.text), gettext(a.long_text),
                    fixup_ico(a.ico_file), commands)
        items.append(item)
        last = a
    # and finally the "parent" of the children, which goes last
    depth -= 1
    item = (MENU_FLAG_NORMAL, depth, u"&Tortoise Bazaar",
            gettext("More Tortoise Bazaar options"),
            fixup_ico("bzr.ico"), [])
    items.append(item)
    return items

class CommandDispatcher:
    """A combination encoder/decoder and dispatcher.
    
    Tied to our byte wire protocol but not to the transport of those bytes.
    """
    # Allow the exceptions to be accessed via an instance.
    RequestFailed = RequestFailed

    # This is the "public" interface to the command-channel.
    commands = {
        # Cmd ID       callable, in_args,
        REQUEST_HELLO: (hello, (unicode,)), 
        REQUEST_GOODBYE: (goodbye, (unicode,)),
        REQUEST_FILE_STATUS: (get_file_status, (unicode, int)),
        REQUEST_MENU_COMMANDS: (get_menu_commands, ([(unicode,)], int)),
        REQUEST_RELEASE_HANDLES: (release_handles, (unicode,)),
        REQUEST_REFRESH_STATUS: (refresh_status, (unicode,)),
        REQUEST_GET_PARAM: (get_param, (unicode,)),
    }

    def dispatch(self, raw_buffer):
        # First 2 bytes are length of request as a sanity check, next 2
        # bytes are client version (hi/low), next 2 bytes the "ID" of the
        # request.
        hdr_fmt = "HBBH"
        hdr_size = struct.calcsize(hdr_fmt)
        if len(raw_buffer) < hdr_size:
            raise RequestFailed("Remote request didn't even have a full header")
    
        # First 2 bytes are length of request as a sanity check, next 2
        # bytes are client version (hi/low), next 2 bytes the "ID" of the
        # request.
        req_len, cver_hi, cver_low, req_id = struct.unpack(
                                            hdr_fmt, raw_buffer[:hdr_size])
        logger.debug("incoming request - request type=%d, client version=%d.%d, bytes=%d",
                     req_id, cver_hi, cver_low, req_len)
        if req_len != len(raw_buffer):
            raise RequestFailed("Byte lengths don't match (read %d, remote said was %d)" %
                                (len(raw_buffer), req_len))
        if cver_hi != 0 or cver_low != 1:
            raise RequestFailed("Expecting version 0.1 clients only (got %d.%d)" %
                                (cver_hi, cver_low))
    
        req_buf = raw_buffer[hdr_size:]
    
        try:
            func, in_desc = self.commands[req_id]
        except KeyError:
            raise RequestFailed("Unknown request ID %d", req_id)
    
        args = unpack_request_data(req_buf, *in_desc)
        ret = func(*args)
        result_buf = pack_request_data(ret)
        return result_buf

# Simple utilities for packing and unpacking simple types into bytes suitable
# for moving across the wire.
def pack_request_data(*args):
    """Pack data into a buffer suitable for sending across the pipe.

    eg:
    buffer = pack_request_data(u"foo", 10)
    """
    bits = []
    for a in args:
        if isinstance(a, list):
            # push the length of the list
            bits.append(struct.pack("H", len(a)))
            # and recursively add the list itself
            for sub in a:
                # each item must be a tuple
                assert isinstance(sub, tuple), sub
                bits.append(pack_request_data(*sub))
        elif isinstance(a, unicode):
            # Don't bother encoding - everything is native unicode.
            this = a.encode('unicode-internal')
            # push the length of the string.
            bits.append(struct.pack("H", len(this)))
            # and the string itself
            fmt = "%ds" % (len(this),)
            bits.append(struct.pack(fmt, this))
        elif isinstance(a, int):
            # use long-long as a nod to 64bits (eg, it may carry a hwnd)
            bits.append(struct.pack("Q", a))
        else:
            raise ValueError("Can't pack %r (type=%s) objects" % (a, type(a)))
    return "".join(bits)

def _unpack_request_data(buf, ndx, types):
    ret = []
    size_size = struct.calcsize("H")
    for t in types:
        if isinstance(t, list):
            (size,) = struct.unpack("H", buf[ndx:ndx+size_size])
            ndx += size_size
            # must be exactly 1 tuple describing each list item
            assert len(t)==1 and type(t[0]) is tuple, t
            litems = []
            for i in range(size):
                sub, ndx = _unpack_request_data(buf, ndx, t[0])
                litems.extend(sub)
            ret.append(litems)

        elif t is unicode:
            (size,) = struct.unpack("H", buf[ndx:ndx+size_size])
            ndx += size_size
            str_fmt = "%ds" % (size,)
            (data,) = struct.unpack(str_fmt, buf[ndx:ndx+size])
            ndx += size
            ret.append(data.decode('unicode-internal'))
        elif t is int:
            # use long-long as a nod to 64bits (eg, it may carry a hwnd)
            llsize = struct.calcsize("Q")
            (val,) = struct.unpack("Q", buf[ndx:ndx+llsize])
            ndx += llsize
            ret.append(val)
        else:
            raise ValueError("Can't unpack objects of type '%s'" % (t,))

    return ret, ndx

def unpack_request_data(buf, *types):
    """Unpack a string into the specified types"""
    ret, end_ndx = _unpack_request_data(buf, 0, types)
    assert end_ndx == len(buf), \
           "Request buffer %d bytes, but consumed %d" % (len(buf), end_ndx)
    return tuple(ret)
