# This is the client side of the pipe communication.  It is primarily for the
# test suite (although it is also used by the Python implemented shell
# extension while the C++ version is pending.)

# Note that this is intentionally not Pythonic - its closer to how the C++
# code will talk to the pipe.

import sys
import os
import time
import thread
import threading
import struct
import subprocess

import pywintypes
import win32api
import win32event
import win32pipe
import win32file
import win32con
import winerror

from tbzrlib.pipe import get_pipe_name
from tbzrlib.dispatcher import pack_request_data, unpack_request_data
from tbzrlib.dispatcher import (REQUEST_HELLO, REQUEST_GOODBYE, REQUEST_RESTART,
                                REQUEST_FILE_STATUS, REQUEST_MENU_COMMANDS,
                                REQUEST_RELEASE_HANDLES, REQUEST_REFRESH_STATUS,
                                )

# 2 diag functions, a-la typical C++ code.
def trace(msg, *args):
    prefix = thread.get_ident()
    try:
        print str(prefix) + ":" + (msg % args)
    except IOError:
        pass

def error(debug_break, msg, *args):
    try:
        if debug_break:
            print "**** Break into debugger!!!"
        print "ERROR:", (msg % args)
    except IOError:
        pass

# A few exceptions for errors communicating over the pipe.
class RemoteRequestException(Exception):
    pass

class RemoteRequestCantSend(RemoteRequestException):
    """Thrown when we can't even send the request to the server (due to the
    pipe not being up, or a recent error, etc)
    """

class RemoteRequestFailed(RemoteRequestException):
    """Thrown when a remote request was made, but the response from the server
    indicated an error occurred at the server.
    """
    def __init__(self, rc, msg=None):
        if msg is None:
            try:
                msg = win32api.FormatMessage(rc).strip()
            except win32api.error:
                msg = ""
        RemoteRequestException.__init__(self, rc, msg)

class RemoteRequestProtocolError(RemoteRequestException):
    """Thrown when the response from the remote request wasn't formatted as
    we expect (eg, premature EOF, etc)
    """

class RemoteCacheLink:
    version = (0,1) # 0.1
    timeout_ms = 10000
    def __init__(self):
        # // create an unnamed (=local) manual reset event for use in the overlapped structure
        # self.event lives longer than one instance of the pipe
        self.event = win32event.CreateEvent(None, 0, 0, None)
        self.hpipe = None
        # 'last_timeout' is unfortunately named - it really means 'when does
        # the last timeout expire" - ie, at what time do I ertry the next
        # attempt at connecting/starting the process...
        self.last_timeout = 0
        self.lock = threading.Lock()

    def close_pipe(self):
        trace("Closing pipe")
        # XXX AutoLocker lock(m_critSec);
        if self.hpipe is not None:
            self.hpipe.Close()
            self.hpipe = None

    def ensure_pipe_open(self):
        if self.hpipe is not None:
            trace("ensure_pipe_open - its already open")
            return True
        pipe_name = get_pipe_name()
        open_mode = win32con.GENERIC_READ | win32con.GENERIC_WRITE
        trace("ensure_pipe_open - opening %s", pipe_name)
        # Connecting is complicated:
        # * CreateFile may return ERROR_PIPE_BUSY meaning you need to wait
        #   for an instance to become available.
        # * WaitNamedPipe does this, but even after it returns success, there
        #   is no guarantee another process still doesn't beat us to getting
        #   that instance.
        # * WaitNamedPipe can return ERROR_FILE_NOT_FOUND if zero instances
        #   exist - presumably this will be due to the one existing instance
        #   that previously existed no longer does.
        # So: we are forced to do all these operations in a large "retry" loop
        # while still being careful not to block for too long (note that in
        # all cases we are just waiting for the server thread to get back
        # around to opening a new connection, so it shouldn't take long.
        # [Note: TSVN doesn't handle the failure of WaitNamedPipe nor that
        # a subsequent ConnectNamedPipe may still fail with "pipe busy - so
        # this now more resembles transport_pipes.cc from Gdb than it does
        # TSVN]
        pipe_wait_retries_left = 64
        pipe_wait_timeout = 50 # wait 50ms for a pipe to become available.
        for i in range(5):
            try:
                hpipe = win32file.CreateFile(pipe_name,
                                             open_mode,
                                             0, # no shareing
                                             None, # default security
                                             win32con.OPEN_EXISTING,
                                             win32con.FILE_FLAG_OVERLAPPED,
                                             None)
                # worked!
                break
            except win32file.error, exc:
                if exc[0] == winerror.ERROR_FILE_NOT_FOUND:
                    # server might still be starting - BUT - its also possible
                    # the server is failing to start - this would mean we
                    # *always* take a full timeout hit, which isn't what we
                    # want.
                    trace("The pipe does not exist")
                    return False
                if exc[0] != winerror.ERROR_PIPE_BUSY:
                    error(True, "Non pipe-busy error: %s", exc)
                    return False
                # // TSVNCache is running but is busy connecting a different client.
                # // Do not give up immediately but wait for a few milliseconds until
                # // the server has created the next pipe instance
                # BUT - see above - ERROR_FILE_NOT_FOUND is possible here.
                # transport_pipes.cc does the "busy-ish" loop here; it seems
                # we could possibly avoid this and just retry from the outer
                # loop again, but it must be done that way for a reason...
                while pipe_wait_retries_left > 0:
                    try:
                        win32pipe.WaitNamedPipe(pipe_name, pipe_wait_timeout)
                        break
                    except win32pipe.error, exc:
                        if exc[0] == winerror.ERROR_FILE_NOT_FOUND:
                            win32api.Sleep(0) # give the server a chance.
                    pipe_retries_left -= 1
                else:
                    # too many retries
                    error(False, "Couldn't wait for pipe to become available")
                    return False
                # OK - theoretically a pipe instance is available - let's
                # go back around the loop and see if we can get it before
                # someone beats us to it!

        # The pipe connected; change to message-read mode.
        win32pipe.SetNamedPipeHandleState(
                        hpipe, win32pipe.PIPE_READMODE_MESSAGE, None, None)
        self.hpipe = hpipe
        return True

    def start_remote_process(self):
        argv = None
        if not hasattr(sys, "frozen"):
            # running from source code - just locate and fire up the process.
            runner = os.path.join(sys.prefix, "pythonw.exe")
            if not os.path.isfile(runner):
                # XXX - 64bits???
                runner = os.path.join(sys.prefix, "PCBuild", "pythonw.exe")
            if not os.path.isfile(runner):
                error(False, "Can't find process runner pythonw.exe in '%s'"
                             % (sys.prefix,))
                return
            # got the .exe, now find the script file.
            import tbzrlib
            script_dir = os.path.join(os.path.dirname(os.path.dirname(tbzrlib.__file__)), "scripts")
            script = os.path.join(script_dir, "tbzrcache.py")
            # no other args for now.
            argv = [runner, script]
        elif hasattr(sys, "frozendllhandle"):
            # probably means we are embedded in a tortoise client!
            mod = win32api.GetModuleFileName(sys.frozendllhandle)
            bindir = os.path.dirname(mod)
            argv = [os.path.join(bindir, "tbzrcachew.exe")]

        if not argv:
            raise RuntimeError, "Don't know how to start the remote process"

        p = subprocess.Popen(argv)
        # don't wait
        trace("Started remote process %s -> %d" % (argv, p.pid))
        return

    def make_raw_remote_request(self, request, nbytes_response):
        # our self.handle management isn't smart enough to deal with 2 threads
        # running, so we use a simple lock.  A smarter way would be a queue
        # of 'available' handles or some such...
        self.lock.acquire()
        try:
            return self._make_raw_remote_request(request, nbytes_response)
        finally:
            self.lock.release()

    def _make_raw_remote_request(self, request, nbytes_response):
        # Make a remote request - 'request' is a buffer already encoded, while
        # nbytes_response is the number of bytes to allocate for the response.
        # The result is a buffer of less than nbytes_response length, or None
        # if we could not make a call on the pipe (either due to being unable
        # to open it, or due to a recent error meaning we don't even try!)
        # XXX - this fixed buffer size is a bit of a problem - but removing
        # it would mean not using TransactNamedPipe() and its likely we see
        # better perf by sticking with it.
        if not self.ensure_pipe_open():
            # // We've failed to open the pipe - try and start the cache
            # // but only if the last try to start the cache was a certain time
            # // ago. If we just try over and over again without a small pause
            # // in between, the explorer is rendered unusable!
            # // Failing to start the cache can have different reasons: missing exe,
            # // missing registry key, corrupt exe, ...
            start = time.time()
            if start - self.last_timeout < 0:
                trace("Remote request last timeout was too recent - skipping")
                raise RemoteRequestCantSend()

            self.start_remote_process()

            # // Wait for the cache to open
            end_time = time.time() + 2
            while not self.ensure_pipe_open():
                now = time.time()
                if now - end_time > 0:
                    self.last_timeout = now + self.timeout_ms / 1000
                    error(True, "Failed to wait for remote process to create the pipe (took %2f ms)",
                          ((now-start)*1000))
                    raise RemoteRequestCantSend()
                trace("sleeping to give the server a chance to create the pipe")
                time.sleep(0.2) # this sleep is not in tsvn!?

        overlapped = pywintypes.OVERLAPPED()
        assert self.event is not None
        overlapped.hEvent = self.event
        #// Do the transaction in overlapped mode.
        #// That way, if anything happens which might block this call
        #// we still can get out of it. We NEVER MUST BLOCK THE SHELL!
        #// A blocked shell is a very bad user impression, because users
        #// who don't know why it's blocked might find the only solution
        #// to such a problem is a reboot and therefore they might loose
        #// valuable data.
        #// One particular situation where the shell could hang is when
        #// the [TSVN crash report dialog comes up]
        # But really, the number of ways things could potentially hang is
        # limited only by your imagination :)
        #// Sure, it would be better to have no situations where the shell
        #// even can get blocked, but the timeout of 10 seconds is long enough
        #// so that users still recognize that something might be wrong and
        #// report back to us so we can investigate further.
        # hrmph: ideally, we would have a shorter timeout (2 seconds?), upon
        # which we show better UI feedback - eg, cursor, window with crawling
        # tortoise animation ;) etc.  That would better reflect when it is
        # us causing any "slowdowns" reported.
        response = win32file.AllocateReadBuffer(nbytes_response)

        try: # catch IO exceptions writing to the pipe
            # Bugger - might not have TransactNamedPipe (needs pywin32-212)
            if hasattr(win32pipe, "TransactNamedPipe"):
                win32pipe.TransactNamedPipe(self.hpipe, request, response,
                                            overlapped)
                rc = win32event.WaitForSingleObject(self.event, self.timeout_ms)
                success = rc == win32event.WAIT_OBJECT_0
                trace("TransactNamedPipe result was %d (success=%s)", rc, success)
            else:
                #trace("no win32pipe.TransactNamedPipe - consider upgrading "
                #      "pywin32to build 212")
                # do it in 2 calls, plus a slight complication for timeouts.
                end_time = win32api.GetTickCount() + self.timeout_ms
                rc, nbytes = win32file.WriteFile(self.hpipe, request, overlapped)
                rc = win32event.WaitForSingleObject(self.event, self.timeout_ms)
                success = rc == win32event.WAIT_OBJECT_0
                trace("TransactNamedPipe-ish write result was %s", success)
                if success:
                    rc, _ = win32file.ReadFile(self.hpipe, response, overlapped)
                    rc = win32event.WaitForSingleObject(self.event,
                                                        end_time - win32api.GetTickCount())
                    success = rc == win32event.WAIT_OBJECT_0
                    trace("TransactNamedPipe-ish read result was %s", success)
            if success:
                nbytes = win32file.GetOverlappedResult(self.hpipe, overlapped, True)
                trace("remote response returned %d bytes", nbytes)
                ret = response[:nbytes]

        except win32file.error, exc:
            # XXX - Note that some errors here will simply mean the server
            # process has gone away, so could be retried.
            # Particularly: Failed using the pipe: (230, 'TransactNamedPipe', 'The pipe state is invalid.')
            trace("Failed using the pipe: %s" % (exc,))
            success = False

        if not success:
            # closing it is important - we will not retry to open it for a short
            # while, meaning one timeout is all Explorer will see for a while.
            trace("failed to issue remote request")
            self.close_pipe()
            raise RemoteRequestCantSend()

        return ret

    def make_remote_request(self, req_type, req_body):
        # pack up the request header
        fmt = "HBBH"
        ver_maj, ver_min = self.version
        total_size = struct.calcsize(fmt) + len(req_body)
        req_header = struct.pack(fmt, total_size, ver_maj, ver_min, req_type)
        # make the request
        response = self.make_raw_remote_request(req_header+req_body, 1024*32)
        # unpack and verify the result.
        resp_fmt = "HH"
        resp_fmt_size = struct.calcsize(resp_fmt)
        if len(response) < resp_fmt_size:
            raise RemoteRequestProtocolError()
        nbytes, errcode = struct.unpack(resp_fmt, response[:resp_fmt_size])
        if nbytes != len(response):
            error(False, "Remote response error: premature EOF")
            raise RemoteRequestProtocolError()
        
        if errcode != 0:
            error(False, "Remote response error: errcode=%d", errcode)
            raise RemoteRequestFailed(errcode)
        # return the raw response data.
        return response[resp_fmt_size:]

    def request_hello(self):
        req = pack_request_data(u"")
        return unpack_request_data(self.make_remote_request(REQUEST_HELLO, req), unicode)[0]

    def request_goodbye(self):
        req = pack_request_data(u"so long")
        return unpack_request_data(self.make_remote_request(REQUEST_GOODBYE, req), unicode)[0]

    def request_file_status(self, filename, flags):
        req = pack_request_data(filename, flags)
        return unpack_request_data(self.make_remote_request(REQUEST_FILE_STATUS, req), int)[0]

    def request_menu_commands(self, filenames, flags):
        req = pack_request_data(filenames, flags)
        command_desc = [(int, int, unicode)]
        ret_desc = [(int, int, unicode, unicode, unicode, command_desc)]
        return unpack_request_data(self.make_remote_request(REQUEST_MENU_COMMANDS, req),
                                   ret_desc)[0]

    def request_release_handles(self, path):
        req = pack_request_data(path)
        return unpack_request_data(self.make_remote_request(REQUEST_RELEASE_HANDLES, req), int)[0]

    def request_refresh_status(self, path):
        req = pack_request_data(path)
        return unpack_request_data(self.make_remote_request(REQUEST_REFRESH_STATUS, req), int)[0]


# a singleton
_cache_link = None
_cache_link_lock = threading.Lock()
def get_cache_link():
    global _cache_link
    if _cache_link is None:
        _cache_link_lock.acquire()
        try:
            if _cache_link is None:
                _cache_link = RemoteCacheLink()
        finally:
            _cache_link_lock.release()
    return _cache_link

if __name__=='__main__':
    def test(argv):
        if not argv:
            argv = ["hello"]
        link = RemoteCacheLink()
        cmd = argv[0]
        args = []
        # mangle up the types of the args in a brutal but effective way.
        for arg in argv[1:]:
            try:
                args.append(int(arg))
            except ValueError:
                args.append(unicode(arg))
        func = getattr(link, "request_" + cmd)
        try:
            ret = func(*tuple(args))
            print "%s->%s" % (cmd, ret)
        except RemoteRequestException, exc:
            import traceback
            traceback.print_exc()
            print "%s->ERROR (%s)" % (cmd, exc)

    test(sys.argv[1:])
