# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import tempfile
import _winreg
from StringIO import StringIO

from bzrlib.revisionspec import RevisionSpec
from bzrlib.tests import TestCaseWithTransport, TestCase

from tbzrlib.tests.shellext import get_clsid_for_handler, _ShellExtensionFeature

import pythoncom
import win32gui
from win32com.shell import shell, shellcon

# These states are from the TSVN generic Icon Overlay handler (enum FileState in
# TortoiseOverlays/ShellExt.h). These are states which share icon overlay slots
# between the various tortoise impls.
FileStateNormal = 0
FileStateModified = 1
FileStateConflict = 2
FileStateLocked = 3
FileStateReadOnly = 4
FileStateDeleted = 5
FileStateAdded = 6
FileStateIgnored = 7
FileStateUnversioned = 8
FileStateInvalid = 9

# This maps each state to a ProgID which implements that specific handler (ie,
# the object we create to test.
state_name_map = {
    FileStateNormal: 'Normal',
    #FileStateModified
    #FileStateConflict
    #FileStateLocked 
    #FileStateReadOnly
    #FileStateDeleted
    FileStateAdded : 'Added',
    #FileStateIgnored
    FileStateUnversioned : 'Unversioned',
    #FileStateInvalid
}

def get_overlay_id_for_state(state):
    # The C++ version of the shell extension doesn't bother registering
    # any ProgIDs - so we look up the TortoiseOverlays section to locate the
    # clsid we need to create.
    name = state_name_map[state]
    for hroot in (_winreg.HKEY_CURRENT_USER, _winreg.HKEY_LOCAL_MACHINE):
        subkey_keyname = r"SOFTWARE\TortoiseOverlays\%s" % name
        try:
            hk = _winreg.OpenKey(hroot, subkey_keyname)
            clsid, type = _winreg.QueryValueEx(hk, "BZR")
            if type == _winreg.REG_SZ:
                break
        except WindowsError:
            continue
    else:
        raise RuntimeError("No handler installed for %r" % name)
    
    iid = shell.IID_IShellIconOverlayIdentifier
    return pythoncom.CoCreateInstance(clsid, None, pythoncom.CLSCTX_ALL, iid)


class _IconOverlayFeature(_ShellExtensionFeature):
    def __init__(self):
        super(_IconOverlayFeature, self).__init__("IconOverlay")

    def _probe(self):
        try:
            # just pick any random handler
            get_overlay_id_for_state(FileStateNormal)
            return True
        except (RuntimeError, pythoncom.com_error):
            return False

    def feature_name(self):
        return '%s Shell Extension' % self.handler_type

IconOverlayFeature = _IconOverlayFeature()


class TestIcons(TestCase):
    _test_needs_features = [IconOverlayFeature]

    def test_icons_valid(self):
        for state, state_name in state_name_map.iteritems():
            handler = get_overlay_id_for_state(state)
            filename, index, flags = handler.GetOverlayInfo()
            # hrm - but our handler always returns emtpy string as
            # TortioseOverlays manages the icons.
            # use win32gui.LoadImage()???

class TestIsMemberOf(TestCaseWithTransport):
    _test_needs_features = [IconOverlayFeature]

    def _check_state_for_file(self, filename, expected_state):
        # convert the filename to a PIDL.
        # Only one handler should say "yes", so check them all to ensure the
        # others say "no".
        filename = os.path.abspath(filename)
        found = False
        attrib = 0 # should come from IShellFolder.GetAttributesOf?
        for state, state_name in state_name_map.iteritems():
            handler = get_overlay_id_for_state(state)
            is_member = handler.IsMemberOf(filename, attrib)==0
            if expected_state is not None:
                # its really an xor, but spell it out for clarity.
                self.failIf(is_member and state != expected_state,
                            "Handler for %s returned True (only %s was expected)" %
                            (state, expected_state))
                self.failIf(not is_member and state == expected_state,
                            "Handler for %s returned False (True expected)" %
                            (state,))
                # else the states are correct.
            if is_member:
                self.failIf(expected_state is None,
                            "Handler for %s returned True" % (state,))
                found = True
        if expected_state is not None:
            self.failIf(not found, "We didn't find the expected handler!")

    def test_non_vcs_dir(self):
        # We expect 'normal' - but what does 'normal' mean above - surely
        # that just means *no* handler?  So enumerate all handlers and ensure
        # none of them work.
        # XXX - this fails due to the user-pref saying "don't show unversioned"
        if 1:
            self._check_state_for_file(tempfile.tempdir, None)
        else:
            self._check_state_for_file(tempfile.tempdir, FileStateUnversioned)

    def test_unmodified(self):
        """Check the added handler detects an added file."""
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([('tree/a', 'content of a\n')])
        tree.add('a')
        tree.commit('one')
        self.failUnless(os.path.isfile('tree/a'))
        self._check_state_for_file('tree/a', FileStateNormal)

    def test_added(self):
        """Check the added handler detects an added file."""
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([('tree/a', 'content of a\n')])
        tree.add('a')
        self.failUnless(os.path.isfile('tree/a'))
        self._check_state_for_file('tree/a', FileStateAdded)

# run these via the tbzrcache selftests
