# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from bzrlib.tests import TestCaseInTempDir, UnavailableFeature
from tbzrlib.watcher import DirectoryWatcherBase
from tbzrlib.vcspath import VCSPath
from tbzrlib.ui import taskbar
from tbzrlib import app
import time
import os
import logging
import win32event
import win32api
import win32gui
import win32con

# we want a new 'application' instance for each test (BUT - this means you
# generally can't call get_app() twice from the same test)
# XXX - also in test_directorycache - its still not quite clear how to
# abstract this cleanly or create TestCase base classes we can reuse!
def get_app(prefs=None):
    if app._app is not None:
        app._app = None
    a = app.get_app()
    if prefs:
        for name, val in prefs.iteritems():
            # only set it if the current value isn't the new value - this
            # also ensures we exercise the 'fetch default value' code.
            if a.config.get(name) != val:
                a.config.set(name, val)
    return a

# We need a hwnd to register device notifications against.  We don't actually
# send or receive the notifications, but we do test the registration and
# unregistration for these events.
class TestWindow(taskbar._TaskbarWindow):
    def create_icon(self):
        # never want a taskbar icon.
        pass

class TestWindowRunner(taskbar.Taskbar):
    WindowClass = TestWindow

class DirectoryWatcher(DirectoryWatcherBase):
    def __init__(self, stop_event, hwnd):
        self.changes = []
        super(DirectoryWatcher, self).__init__(stop_event, hwnd)
    def on_change_detected(self, filename):
        self.changes.append(filename)

class TestDirectoryWatcher(TestCaseInTempDir):
    def wait_for_changes(self):
        for i in range(20):
            if self.watcher.changes:
                return self.watcher.changes
            time.sleep(0.1)
        raise AssertionError, "no changes!"

    def setUp(self):
        TestCaseInTempDir.setUp(self)
        self.app = get_app()
        self.window_runner = TestWindowRunner(self.app)
        self.window_runner.start()
        self.watcher = DirectoryWatcher(self.app.stop_event, self.window_runner.get_hwnd())
        self.watcher.start()
    
    def tearDown(self):
        self.failUnless(self.watcher.isAlive(), "watcher already stopped?")
        win32event.SetEvent(self.app.stop_event)
        self.app = None
        self.window_runner.join(5)
        self.failIf(self.window_runner.isAlive(), "window thread didn't die")
        self.watcher.join(5)
        self.failIf(self.watcher.isAlive(), "watcher didn't die")
        TestCaseInTempDir.tearDown(self)
        
    def test_add_path(self):
        logger = logging.getLogger('bzr')
        self.watcher.add_path(VCSPath(unicode(self.test_dir)))
        self.failUnlessEqual(self.watcher.watched_paths_count, 1)
        self.watcher._trigger_handle_rebuild(True)
        fname = os.path.abspath('test.txt')
        logger.info('modifying file %r', fname)
        f = file(fname, "w")
        f.close()
        os.remove("test.txt")
        changes = self.wait_for_changes()
        self.failUnless(fname in changes, changes)

    def test_expiry(self):
        self.watcher.add_path(VCSPath(unicode(self.test_dir)), 0.1)
        self.failUnlessEqual(self.watcher.watched_paths_count, 1)
        time.sleep(0.2) # wait for expiry to pass
        self.watcher._trigger_handle_rebuild(True)
        self.failUnlessEqual(self.watcher.watched_paths_count, 0)

    def test_common_path(self):
        os.mkdir("path1")
        os.mkdir("path2")
        self.watcher.add_path(VCSPath(unicode("path1")))
        self.failUnlessEqual(self.watcher.watched_paths_count, 1)
        # Add a second path with a shared parent - we expect only one watcher:
        # that of the shared parent.
        self.watcher.add_path(VCSPath(unicode("path2")))
        self.failUnlessEqual(self.watcher.watched_paths_count, 1)
        self.failUnlessEqual(self.watcher.watched_paths[0][0].path,
                             os.path.abspath(self.test_dir))

    def test_child_path(self):
        os.mkdir("path1")
        path1 = VCSPath(unicode("path1"))
        os.mkdir("path1\\sub")
        sub = VCSPath(unicode("path1\\sub"))
        self.watcher.add_path(path1)
        self.failUnlessEqual(self.watcher.watched_paths_count, 1)
        # Add a child directory - we should detect it is a child and continue
        # to watch the parent.
        self.watcher.add_path(sub)
        self.failUnlessEqual(self.watcher.watched_paths_count, 1)
        self.failUnless(self.watcher.is_path_watched(path1))
        self.failUnless(self.watcher.is_path_watched(sub))

    def test_invalid_path(self):
        path = VCSPath(unicode("not-a-file.txt"))
        self.watcher.add_path(path)
        self.failUnlessEqual(self.watcher.watched_paths_count, 1)
        self.watcher._trigger_handle_rebuild(True)
        self.failUnlessEqual(self.watcher.watched_paths_count, 0)

    def test_close_handle(self):
        path = VCSPath(unicode(self.test_dir))
        self.watcher.add_path(path)
        self.failUnlessEqual(self.watcher.watched_paths_count, 1)
        self.watcher._trigger_handle_rebuild(True)
        info = self.watcher._current_watch_map[path]
        if info.hDevNotify:
            self.watcher.close_watcher_handle(info.hDevNotify)
            self.failUnlessEqual(self.watcher.watched_paths_count, 0)
            self.failUnless(self.watcher.is_path_blocked(path))
        else:
            raise UnavailableFeature("pywin32-213 doesn't seem installed")

if __name__=='__main__':
    # XXX - how to hook into bzrlib selftest?
    import unittest
    unittest.main()
