# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
import os
import time
from StringIO import StringIO
import tempfile

from bzrlib.revisionspec import RevisionSpec
from bzrlib.tests import TestCaseWithTransport, TestCase
from bzrlib import osutils, ignores

from tbzrlib.vcspath import VCSPath
import tbzrlib.app
import tbzrlib.dispatcher
from tbzrlib.app import BZRStatus
from tbzrlib import wtcache

# we want a new 'application' instance for each test (BUT - this means you
# generally can't call get_app() twice from the same test)
def get_app(prefs=None):
    if tbzrlib.app._app is not None:
        tbzrlib.app._app = None
    app = tbzrlib.app.get_app()
    if prefs:
        for name, val in prefs.iteritems():
            # only set it if the current value isn't the new value - this
            # also ensures we exercise the 'fetch default value' code.
            if app.config.get(name) != val:
                app.config.set(name, val)
    return app

class TestVCSPath(TestCase):
    def test_hash(self):
        tempdir = tempfile.tempdir.decode(sys.getfilesystemencoding())
        d = {}
        d[VCSPath(tempdir)] = 1
        # create a new object to index with.
        self.failUnlessEqual(d.get(VCSPath(tempdir)), 1)

class TestSimple(TestCase):
    def test_clear(self):
        path = VCSPath(tempfile.tempdir.decode(sys.getfilesystemencoding()))
        dc = get_app().dir_cache
        # should not be in the cache.
        self.failUnlessEqual(dc._get_cache_for_path(path), None)
        entry = dc._get_cache_for_path(path, True)
        # its in the cache!
        self.failUnless(dc._get_cache_for_path(path) is entry)
        # clear the cache
        dc.clear_cache()
        # should again not be in the cache.
        self.failUnlessEqual(dc._get_cache_for_path(path), None)
        self.failUnlessEqual(len(dc.dircache), 0)

    def test_remove_cached_item(self):
        path = VCSPath(tempfile.tempdir.decode(sys.getfilesystemencoding()))
        dc = get_app().dir_cache
        # should not be in the cache.
        self.failUnlessEqual(dc._get_cache_for_path(path), None)

        entry = dc._get_cache_for_path(path, True)
        # its in the cache!
        self.failUnless(dc._get_cache_for_path(path) is entry)
        # remove it from the cache.
        dc.remove_cache_for_path(path)
        # should again not be in the cache.
        self.failUnlessEqual(dc._get_cache_for_path(path), None)

    def test_fails_for_file(self):
        path = VCSPath(__file__.decode(sys.getfilesystemencoding()))
        dc = get_app().dir_cache
        # should not be in the cache.
        self.failUnlessRaises(ValueError, dc._get_cache_for_path, path, True)

# A base class that has helper methods for giving back a tree in various
# states.
class TestWithSimpleTreeStates(TestCaseWithTransport):

    def get_tree_with_added(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/dir/'])
        self.build_tree_contents([('tree/dir/a', 'content of a\n')])
        tree.add('dir/')
        tree.add('dir/a')
        self.failUnless(os.path.isfile('tree/dir/a'))
        root_path = VCSPath(os.path.abspath(tree.abspath('')))
        return tree, root_path

    def get_tree_with_unchanged(self):
        tree, root = self.get_tree_with_added()
        tree.commit('one')
        return tree, root

    def get_tree_with_ignored(self):
        tree, root = self.get_tree_with_unchanged()
        ignores.tree_ignores_add_patterns(tree, ['ignored'])
        self.build_tree_contents([('tree/dir/ignored', 'ignored file\n')])
        self.failUnless(os.path.isfile('tree/dir/ignored'))
        return tree, root

    def get_tree_with_deleted(self):
        tree, root = self.get_tree_with_unchanged()
        self.failUnless(os.path.isfile('tree/dir/a'))
        tree.remove(['dir/a'], keep_files=False)
        self.failIf(os.path.isfile('tree/dir/a'))
        self.failUnless(os.path.isdir('tree/dir'))
        return tree, root

    def get_tree_with_modified(self):
        tree, root = self.get_tree_with_added()
        tree.commit('one')
        f = open('tree/dir/a', 'w')
        f.write('new content of a\n')
        f.close()
        return tree, root

# NOTE that most other tests have a *parent* repo above the test directory.
# For example, requesting the bzr status of the parent of the test tree
# generally succeeds.  This is by-design in the bzr test suite.
# This test class side-steps that by creating its own temp directory and doing
# its tests there, explicitly to check how things behave in that situation.
class TestNonVersionedParent(TestWithSimpleTreeStates):
    def setUp(self):
        super(TestNonVersionedParent, self).setUp()
        temp_dir = tempfile.mkdtemp(prefix='testbzr-tbzr-test-branch-')
        os.chdir(temp_dir)
        self.addCleanup(lambda:osutils.rmtree(temp_dir))
        self.addCleanup(lambda:os.chdir('..'))
        self.temp_dir = temp_dir

    def test_nonversioned_parent(self):
        # First request we make on this non-versioned directory is the root
        # of the directory itself.
        tree = self.make_branch_and_tree('tree')
        # fetch the status for the non-versioned parent of the tree.
        tree_parent = VCSPath(self.temp_dir.decode(sys.getfilesystemencoding()))
        app = get_app()
        status = app.dir_cache.get_status_for_path(tree_parent, False)
        self.failIf(status.is_versioned(), status)
        # not only should it not be versioned, it should be 'nothing'
        self.failUnlessEqual(status, BZRStatus(BZRStatus.STATE_NOTHING))

        # but asking for the tree itself should give us a state.
        status = app.dir_cache.get_status_for_path(tree_parent.join('tree'), False)
        ex = BZRStatus(BZRStatus.STATE_UNCHANGED)
        self.failUnlessEqual(status, ex)

    def test_nonversioned_parent_root(self):
        # First request we make on this non-versioned directory is versioned
        # sub-directory.
        tree = self.make_branch_and_tree('tree')
        # fetch the status for the non-versioned parent of the tree.
        tree_parent = VCSPath(self.temp_dir.decode(sys.getfilesystemencoding()))
        app = get_app()
        status = app.dir_cache.get_status_for_path(tree_parent, False)
        self.failUnlessEqual(status, BZRStatus(BZRStatus.STATE_NOTHING))
        # asking for the tree itself should give us a state.
        status = app.dir_cache.get_status_for_path(tree_parent.join('tree'), False)
        ex = BZRStatus(BZRStatus.STATE_UNCHANGED)
        self.failUnlessEqual(status, ex)

    def test_nonversioned_parent_child(self):
        # First entry object we get is the versioned child
        tree = self.make_branch_and_tree('tree')
        # fetch the status for the non-versioned parent of the tree.
        tree_parent = VCSPath(self.temp_dir.decode(sys.getfilesystemencoding()))
        child = tree_parent.join('tree')
        # asking for the tree itself should give us a state.
        app = get_app()
        status = app.dir_cache.get_status_for_path(child, False)
        ex = BZRStatus(BZRStatus.STATE_UNCHANGED)
        self.failUnlessEqual(status, ex)

class TestSimpleVersioned(TestWithSimpleTreeStates):
    def test_unchanged(self):
        tree, path = self.get_tree_with_unchanged()
        app = get_app()
        status = app.dir_cache.get_status_for_path(path, False)
        # Nothing is modified
        ex = BZRStatus(BZRStatus.STATE_UNCHANGED)
        self.failUnlessEqual(status, ex)

    def test_unchanged_cached(self):
        tree, path = self.get_tree_with_unchanged()
        app = get_app()
        status = app.dir_cache.get_status_for_path(path, False)
        # Nothing is modified
        self.failUnlessEqual(status, BZRStatus(BZRStatus.STATE_UNCHANGED))
        # should be the same result if we ask again
        status = app.dir_cache.get_status_for_path(path, False)
        self.failUnlessEqual(status, BZRStatus(BZRStatus.STATE_UNCHANGED))

    def test_ignored_file(self):
        tree, path = self.get_tree_with_ignored()
        dir = path.join('dir')
        app = get_app()
        status = app.dir_cache.get_status_for_path(dir.join('ignored'), False)
        ex = BZRStatus(BZRStatus.STATE_IGNORED)
        self.failUnlessEqual(status, ex)

    def test_added_file(self):
        tree, root_path = self.get_tree_with_added()
        dir = root_path.join('dir')
        app = get_app()
        status = app.dir_cache.get_status_for_path(dir.join('a'), False)
        # The entry for the child should be 'added'
        ex = BZRStatus(BZRStatus.STATE_ADDED)
        self.failUnlessEqual(status, ex)

    def test_added_child(self):
        tree, path = self.get_tree_with_added()
        app = get_app()
        status = app.dir_cache.get_status_for_path(path, False)
        # The entry for the dir itself should be 'modified'
        ex = BZRStatus(BZRStatus.STATE_MODIFIED)
        self.failUnlessEqual(status, ex)

    def test_modified_file(self):
        tree, root_path = self.get_tree_with_modified()
        dir = root_path.join('dir')
        app = get_app()
        status = app.dir_cache.get_status_for_path(dir.join('a'), False)
        # The entry for the child should be 'modified'
        ex = BZRStatus(BZRStatus.STATE_MODIFIED)
        self.failUnlessEqual(status, ex)

    def test_deleted_file(self):
        tree, root_path = self.get_tree_with_deleted()
        dir = root_path.join('dir')
        app = get_app()
        status = app.dir_cache.get_status_for_path(dir.join('a'), False)
        # The entry for the child should be 'deleted'
        ex = BZRStatus(BZRStatus.STATE_DELETED)
        self.failUnlessEqual(status, ex)

    def _test_unversioned_file(self, prefs, expected_parent_state):
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/dir/'])
        self.build_tree_contents([('tree/dir/a', 'content of a\n')])
        self.failUnless(os.path.isfile('tree/dir/a'))
        tree.add('dir/')
        tree.commit("one")
        app = get_app(prefs)
        dir = VCSPath(os.path.abspath(tree.abspath(''))).join("dir")
        status = app.dir_cache.get_status_for_path(dir.join('a'), False)
        # The entry for the child should be 'unversioned'
        ex = BZRStatus(BZRStatus.STATE_UNVERSIONED)
        self.failUnlessEqual(status, ex)
        # check the status of the parent directory itself
        status = app.dir_cache.get_status_for_path(dir, False)
        self.failUnlessEqual(status, expected_parent_state)

    def test_unversioned_file_1(self):
        prefs = {'icon_overlays.show_as_mod_if_unversioned': False}
        self._test_unversioned_file(prefs,
                                    BZRStatus(BZRStatus.STATE_UNCHANGED))

    def test_unversioned_file_2(self):
        prefs = {'icon_overlays.show_as_mod_if_unversioned': True}
        self._test_unversioned_file(prefs,
                                    BZRStatus(BZRStatus.STATE_MODIFIED))

    def _test_unversioned_subdir(self, prefs, expected_parent_state):
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/dir/'])
        self.failUnless(os.path.isdir('tree/dir'))
        dir = VCSPath(tree.abspath(''))

        # check the status of the parent directory itself
        app = get_app(prefs)
        status = app.dir_cache.get_status_for_path(dir, False)
        self.failUnlessEqual(status, expected_parent_state)

    def test_unversioned_subdir_1(self):
        prefs = {'icon_overlays.show_as_mod_if_unversioned': False}
        self._test_unversioned_subdir(prefs,
                                    BZRStatus(BZRStatus.STATE_UNCHANGED))

    def test_unversioned_subdir_2(self):
        prefs = {'icon_overlays.show_as_mod_if_unversioned': True}
        self._test_unversioned_subdir(prefs,
                                    BZRStatus(BZRStatus.STATE_MODIFIED))

    def _test_ignored_file(self, prefs, expected_parent_state):
        tree, path = self.get_tree_with_ignored()
        dir = path.join('dir')
        app = get_app()
        status = app.dir_cache.get_status_for_path(dir.join('ignored'), False)
        # The entry for the child should be 'ignored'
        ex = BZRStatus(BZRStatus.STATE_IGNORED)
        self.failUnlessEqual(status, ex)
        # check the status of the parent directory itself
        status = app.dir_cache.get_status_for_path(dir, False)
        self.failUnlessEqual(status, expected_parent_state)

    def test_ignored_file_1(self):
        prefs = {'icon_overlays.show_as_mod_if_unversioned': False}
        self._test_ignored_file(prefs, BZRStatus(BZRStatus.STATE_UNCHANGED))

    def test_ignored_file_2(self):
        # Ignored files do *NOT* cause the 'show_as_mod_if_unversioned'
        # preference to kick in...
        prefs = {'icon_overlays.show_as_mod_if_unversioned': True}
        self._test_ignored_file(prefs, BZRStatus(BZRStatus.STATE_UNCHANGED))
    
    def test_refresh_child(self):
        tree, path = self.get_tree_with_unchanged()
        dir = path.join('dir')
        child = dir.join('a')
        app = get_app()
        status = app.dir_cache.get_status_for_path(child, False)
        ex = BZRStatus(BZRStatus.STATE_UNCHANGED)
        self.failUnlessEqual(status, ex)
        # Now, modify the file on the filesystem and call 'refresh' - it should
        # magically become 'modified'
        n = child.last_write_time
        # *sob* - this sucks - ensure we get a new timestamp
        time.sleep(2) # needs >1 seconds to discard most_recent_status.
        open('tree/dir/a', 'w').close() # enough to make it modified.
        # it was different - should have hit vcs again.
        #self.failUnlessEqual(cacheddirectory.n_vcs_requests, 2)
        status = app.dir_cache.get_status_for_path(VCSPath(child.path), False)
        ex = BZRStatus(BZRStatus.STATE_MODIFIED)
        self.failUnlessEqual(status, ex)
        # should not have hit vcs re-fetching the status.
        #self.failUnlessEqual(cacheddirectory.n_vcs_requests, 2)

# This is testing the main entry point for the cache.
class TestGetStatusForPath(TestWithSimpleTreeStates):
    def test_unchanged(self):
        tree, path = self.get_tree_with_unchanged()
        app = get_app()
        status = app.dir_cache.get_status_for_path(path, True)
        # Nothing is modified
        ex = BZRStatus(BZRStatus.STATE_UNCHANGED)
        self.failUnlessEqual(status, ex)

    def test_added_file(self):
        tree, root_path = self.get_tree_with_added()
        dir = root_path.join('dir')
        app = get_app()
        status = app.dir_cache.get_status_for_path(dir.join('a'), False)
        # The entry for the child should be 'added'
        ex = BZRStatus(BZRStatus.STATE_ADDED)
        self.failUnlessEqual(status, ex)

    def test_added_file_parent(self):
        tree, root_path = self.get_tree_with_added()
        dir = root_path.join('dir')
        app = get_app()
        status = app.dir_cache.get_status_for_path(dir, True)
        # The status for the parent dir is 'added'
        ex = BZRStatus(BZRStatus.STATE_ADDED)
        self.failUnlessEqual(status, ex)
        # but the status of the root is simply 'modified'
        status = app.dir_cache.get_status_for_path(root_path, True)
        ex = BZRStatus(BZRStatus.STATE_MODIFIED)
        self.failUnlessEqual(status, ex)
        # we made 2 vcs requests - the child then the parent.
        #self.failUnlessEqual(cacheddirectory.n_vcs_requests, 2)

    def test_added_file_root(self):
        tree, root_path = self.get_tree_with_added()
        app = get_app()
        status = app.dir_cache.get_status_for_path(root_path, True)
        # ALSO: this test will fail when we get non-recursive status updates -
        # in which case we will need to wait for the crawler.
        ex = BZRStatus(BZRStatus.STATE_MODIFIED)
        self.failUnlessEqual(status, ex)

    def test_added_child(self):
        tree, path = self.get_tree_with_added()
        app = get_app()
        status = app.dir_cache.get_status_for_path(path, False)
        # The entry for the dir itself should be 'modified' (well, its not clear
        # it *should* be added, but it is now :)
        ex = BZRStatus(BZRStatus.STATE_MODIFIED)
        self.failUnlessEqual(status, ex)

    def test_modified_file(self):
        tree, root_path = self.get_tree_with_modified()
        child = root_path.join('dir/a')
        app = get_app()
        status = app.dir_cache.get_status_for_path(child, False)
        # The entry for the child should be 'modified'
        ex = BZRStatus(BZRStatus.STATE_MODIFIED)
        self.failUnlessEqual(status, ex)

    def test_deleted_file(self):
        tree, root_path = self.get_tree_with_deleted()
        child = root_path.join('dir/a')
        app = get_app()
        status = app.dir_cache.get_status_for_path(child, False)
        # The entry for the child should be 'deleted'
        ex = BZRStatus(BZRStatus.STATE_DELETED)
        self.failUnlessEqual(status, ex)

    def test_nowatch_path(self):
        app = get_app()
        tree, root_path = self.get_tree_with_modified()
        # tell it not to watch this dir.
        app.dir_cache.nowatch_paths.add(root_path)
        child = root_path.join('dir/a')
        status = app.dir_cache.get_status_for_path(child, False)
        # The entry for the child should be 'unversioned'
        self.failIf(status.is_versioned(), status)
        # should not have hit the vcs at all
        #self.failUnlessEqual(cacheddirectory.n_vcs_requests, 0)


# When we go via the dispatcher we get some additional semantics - test them
# here.
class TestViaDispatcher(TestWithSimpleTreeStates):
    def test_long_short_path(self):
        import win32api
        tree, path = self.get_tree_with_unchanged()
        short_path = win32api.GetShortPathName(path.path)
        long_path = win32api.GetLongPathNameW(path.path)
        status = tbzrlib.dispatcher.get_file_status(short_path, 1)
        # Nothing is modified
        self.failUnlessEqual(status, BZRStatus.STATE_UNCHANGED)
        # and a request for the long version of the path should get the exact
        # same result.
        status = tbzrlib.dispatcher.get_file_status(long_path, 1)
        self.failUnlessEqual(status, BZRStatus.STATE_UNCHANGED)
        #self.failUnlessEqual(cacheddirectory.n_vcs_requests, 1)
    

# This is black-box testing for the directory watcher
class TestGetStatusForChangingPath(TestWithSimpleTreeStates):
    def test_modified(self):
        tree, path = self.get_tree_with_unchanged()
        app = get_app()
        status = app.dir_cache.get_status_for_path(path, True)
        # Nothing is modified
        self.failUnlessEqual(status, BZRStatus(BZRStatus.STATE_UNCHANGED))
        # should only have made 1 VCS request
        #self.failUnlessEqual(cacheddirectory.n_vcs_requests, 1)
        # modify the file.
        self.build_tree_contents([('tree/dir/a', 'new content of a\n')])
        self.failUnlessEqual(status, BZRStatus(BZRStatus.STATE_MODIFIED))
        # one more request to get the new status.
        #self.failUnlessEqual(cacheddirectory.n_vcs_requests, 2)
    

class TestGetStatusForPathUnversioned(TestCase):
    def test_nonversioned_dir(self):
        path = VCSPath(tempfile.tempdir.decode(sys.getfilesystemencoding()))
        app = get_app()
        status = app.dir_cache.get_status_for_path(path, False)
        # The entry for the dir should be 'nothing'
        self.failUnless(status.is_nothing(), status)
        # which we still see as 'not versioned'
        self.failIf(status.is_versioned(), status)
        # but not 'unversioned'
        self.failIf(status.is_unversioned(), status)

    def test_nonversioned_file(self):
        temp_name = os.path.join(tempfile.tempdir, "tbzr_test_file")
        open(temp_name, "w").close()
        try:
            path = VCSPath(temp_name.decode(sys.getfilesystemencoding()))
            app = get_app()
            status = app.dir_cache.get_status_for_path(path, False)
            # The entry for the child should be 'nothing'
            self.failUnless(status.is_nothing(), status)
            # which we still see as 'not versioned'
            self.failIf(status.is_versioned(), status)
            # but not 'unversioned'
            self.failIf(status.is_unversioned(), status)
        finally:
            os.unlink(temp_name)

# This is black-box testing for the directory watcher's integration with the
# cache.
class TestGetStatusForChangingPath(TestWithSimpleTreeStates):
    def setUp(self):
        super(TestGetStatusForChangingPath, self).setUp()
        from tbzrlib.watcher import DirectoryWatcher
        self.app = get_app()
        self.dir_watcher = DirectoryWatcher(self.app.stop_event, self.app.dir_cache)
        self.dir_watcher.start()

    def tearDown(self):
        self.dir_watcher.stop()
        super(TestGetStatusForChangingPath, self).tearDown()

    def wait_for_status_change(self, path, expected):
        for i in range(20):
            status = self.app.dir_cache.get_status_for_path(path, True)
            if status == expected:
                break
            time.sleep(0.1)
        self.failUnlessEqual(status, expected)

    def test_modified(self):
        tree, path = self.get_tree_with_unchanged()
        status = self.app.dir_cache.get_status_for_path(path, True)
        # Nothing is modified
        self.failUnlessEqual(status, BZRStatus(BZRStatus.STATE_UNCHANGED))
        # should only have made 1 VCS request
        #self.failUnlessEqual(cacheddirectory.n_vcs_requests, 1)
        # modify the file - the dir watcher should pickup the change.
        self.build_tree_contents([('tree/dir/a', 'new content of a\n')])
        self.wait_for_status_change(path, BZRStatus(BZRStatus.STATE_MODIFIED))

if __name__=='__main__':
    # XXX - how to hook into bzrlib selftest?
    import unittest
    unittest.main()
