# Manages the task-bar icon.

# This is currently no more than a "stub" implementation.  Currently the
# icon is always shown, and only offers to exit the program.
# Later we might want to add tooltips showing the status, options to
# disable icons, etc - see TSVNCache for inspiration.

import sys
import os
import threading
import logging

import win32api
import win32event
import win32con
import win32process
import winerror
if sys.getwindowsversion() >= (5,1):
    import winxpgui as win32gui
else:
    import win32gui
import win32gui_struct
import struct

logger = logging.getLogger("bzr.tbzr.taskbar")
from tbzrlib.i18n import gettext
from tbzrlib import actions
from tbzrlib.dispatcher import fixup_ico # this is misplaced :(

# GUIDs for device notifications we care about from MSSDK's Ioevent.h
GUID_IO_VOLUME_DISMOUNT = "{d16a55e8-1059-11d2-8ffd-00a0c9a06d32}"
GUID_IO_VOLUME_LOCK = "{50708874-c9af-11d1-8fef-00a0c9a06d32}"

# Loads an icon given a string in one of the formats we support.
# NOTE: you must call win32gui.DestroyIcon on the result if not None
def load_icon(ico_name):
    ico = fixup_ico(ico_name)
    ico_flags = win32con.LR_DEFAULTSIZE
    try:
        if '#' in ico:
            mod_name, res = ico.split('#', 1)
            res = int(res)
            logger.debug("Loading icon resource %r from %r", res, mod_name)
            mod = win32api.LoadLibraryEx(mod_name, None,
                                         win32con.LOAD_LIBRARY_AS_DATAFILE)
        else:
            mod = None
            res = ico
            ico_flags |= win32con.LR_LOADFROMFILE
            logger.debug("Loading icon file %r", ico)

        hico = win32gui.LoadImage(mod, res, win32con.IMAGE_ICON, 0, 0,
                                  ico_flags)
    except win32gui.error, exc:
        logger.error("Failed to load icon %r: %s", ico, exc)
        hico = None
    if mod is not None:
        win32api.FreeLibrary(mod)
    return hico

class _TaskbarWindow:
    """A hidden window used for the icon"""
    CMD_EXIT = 1025
    CMD_SLEEP = 1026
    WM_USER_CONFIG_CHANGE = win32con.WM_USER
    WM_USER_TASKBAR = win32con.WM_USER + 1

    def __init__(self, app):
        self.app = app
        self.menu_items = {}
        # If explorer dies and restarts, it sends this message to us.
        msg_TaskbarCreated = win32gui.RegisterWindowMessage("TaskbarCreated")
        message_map = {
                msg_TaskbarCreated: self.OnTaskbarCreated,
                win32con.WM_DESTROY: self.OnDestroy,
                win32con.WM_COMMAND: self.OnCommand,
                self.WM_USER_TASKBAR : self.OnTaskbarNotify,
                # owner-draw related handlers.
                win32con.WM_MEASUREITEM: self.OnMeasureItem,
                win32con.WM_DRAWITEM: self.OnDrawItem,
                self.WM_USER_CONFIG_CHANGE: self.OnTBZRSettingChange,
                # Device notifications.
                win32con.WM_DEVICECHANGE: self.OnDeviceChange,
        }
        # Register the Window class.
        wc = win32gui.WNDCLASS()
        hinst = wc.hInstance = win32api.GetModuleHandle(None)
        wc.lpszClassName = "TBZRCache_Taskbar"
        wc.style = win32con.CS_VREDRAW | win32con.CS_HREDRAW;
        wc.hCursor = win32gui.LoadCursor( 0, win32con.IDC_ARROW )
        wc.hbrBackground = win32con.COLOR_WINDOW
        wc.lpfnWndProc = message_map # could also specify a wndproc.
        try:
            classAtom = win32gui.RegisterClass(wc)
        except win32gui.error, exc:
            # might be running the test suite etc.
            if exc[0] != winerror.ERROR_CLASS_ALREADY_EXISTS:
                raise
            classAtom = wc.lpszClassName
        # Create the Window.
        style = win32con.WS_OVERLAPPED | win32con.WS_SYSMENU
        self.hwnd = win32gui.CreateWindow( classAtom, "Tortoise BZR", style, \
                0, 0, win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT, \
                0, 0, hinst, None)
        win32gui.UpdateWindow(self.hwnd)
        self.have_icon = False
        self.create_icon()

    def create_icon(self):
        if not self.app.config.get("task_bar.always_show_icon"):
            self.destroy_icon() # incase our config changed.
            return

        hico = load_icon('bzr.ico')
        if hico is not None:
            try:
                # add the icon to the taskbar.
                flags = win32gui.NIF_ICON | win32gui.NIF_MESSAGE | win32gui.NIF_TIP
                nid = (self.hwnd, 0, flags, self.WM_USER_TASKBAR, hico,
                       u"Tortoise Bazaar")
                try:
                    win32gui.Shell_NotifyIcon(win32gui.NIM_ADD, nid)
                    self.have_icon = True
                except win32gui.error:
                    # This is common when windows is starting, and this code is hit
                    # before the taskbar has been created.  When explorer starts, we
                    # get the TaskbarCreated message, so it all works out in the end :)
                    logger.debug("It looks like we were too fast for explorer...")
            finally:
                win32gui.DestroyIcon(hico)
        else:
            logger.error("Failed to load the bzr icon for the taskbar")

    def destroy_icon(self):
        if not self.have_icon:
            return
        nid = (self.hwnd, 0)
        try:
            win32gui.Shell_NotifyIcon(win32gui.NIM_DELETE, nid)
            logger.debug("Removed taskbar icon")
        except win32gui.error:
            # can't remove it - maybe explorer is dead? Whatever, we mustn't
            # fail
            logger.info("Failed to remove the taskbar icon")
        self.have_icon = False

    def OnTBZRSettingChange(self, hwnd, msg, wparam, lparam):
        # We have changed configuration, and the app has already loaded
        # the new config - all we need do is update the state of our icon
        # incase that user-pref is the one that changed.
        logger.info("Reqeust to reload our config")
        self.create_icon()

    def OnTaskbarCreated(self, hwnd, msg, wparam, lparam):
        logger.debug("Explorer just created a taskbar...")
        self.create_icon()

    def OnDestroy(self, hwnd, msg, wparam, lparam):
        # The window is on its way down - terminate our event loop if it
        # isn't already.
        self.destroy_icon()
        win32gui.PostQuitMessage(1) # Terminate the app.

    def _fillMenu(self, hmenu):
        # Adding our menus with a bitmap of HBMMENU_CALLBACK allows us to
        # draw it (converting directly to a bitmap makes transparency hard)
        self.menu_items.clear()
        PackMENUITEMINFO = win32gui_struct.PackMENUITEMINFO
        cmd_id = self.CMD_SLEEP

        for act_id in ['tbzr.help', 'tbzr.settings', 'tbzr.about']:
            cmd_id += 1
            action = actions.get_action(act_id)
            self.menu_items[cmd_id] = action
            hbmpItem = None
            if action.ico_file:
                hbmpItem = win32con.HBMMENU_CALLBACK

            item, extras = PackMENUITEMINFO(text=gettext(action.text),
                                            hbmpItem=hbmpItem,
                                            wID=cmd_id)
            win32gui.InsertMenuItem(hmenu, -1, 1, item)

        # a sep
        item, extras = win32gui_struct.PackMENUITEMINFO(
                       fType=win32con.MFT_SEPARATOR)
        win32gui.InsertMenuItem(hmenu, -1, 1, item)

        # 'Sleep' command.
        fState = win32con.MFS_UNCHECKED
        if self.app.sleeping:
            fState = win32con.MFS_CHECKED
        item, extras = PackMENUITEMINFO(text=gettext("&Sleep TBZR"),
                                        wID=self.CMD_SLEEP,
                                        fState=fState,
                                        )
        win32gui.InsertMenuItem(hmenu, -1, 1, item)

        # 'Exit' command.
        hbmpItem = win32con.HBMMENU_POPUP_CLOSE # no icon for 'sleep'
        item, extras = PackMENUITEMINFO(text=gettext("Exit program"),
                                        hbmpItem=hbmpItem,
                                        wID=self.CMD_EXIT)
        win32gui.InsertMenuItem(hmenu, -1, 1, item)


    def OnTaskbarNotify(self, hwnd, msg, wparam, lparam):
        # For now, we treat left and right clicks identically - display a
        # menu.
        # these happen *lots* - only see them at max log level
        logger.log(1, "Taskbar notify - wparam=%x, lparam=%x", wparam, lparam)
        if lparam in (win32con.WM_LBUTTONUP, win32con.WM_RBUTTONUP):
            self.hmenu = menu = win32gui.CreatePopupMenu()
            self._fillMenu(menu)
            pos = win32gui.GetCursorPos()
            # See http://msdn.microsoft.com/library/default.asp?url=/library/en-us/winui/menus_0hdi.asp
            # (*sigh* - link is stale - search MSDN for 'TrackPopupMenu' examples)
            win32gui.SetForegroundWindow(self.hwnd)
            win32gui.TrackPopupMenu(menu, win32con.TPM_LEFTALIGN,
                                    pos[0], pos[1], 0, self.hwnd, None)
            win32gui.PostMessage(self.hwnd, win32con.WM_NULL, 0, 0)
        return 1

    def OnCommand(self, hwnd, msg, wparam, lparam):
        id = win32api.LOWORD(wparam)
        logger.debug("Taskbar command - id=%d", id)

        if id == self.CMD_EXIT:
            # There are oh-so-many ways to terminate our app - killing the
            # window, which terminates the event-loop is but one :)
            logger.info("Closing application in response to menu command")
            win32gui.DestroyWindow(self.hwnd)
            return

        if id == self.CMD_SLEEP:
            logger.info("Sleeping application in response to menu command")
            item, extras = win32gui_struct.EmptyMENUITEMINFO()
            win32gui.GetMenuItemInfo(self.hmenu, id, False, item)
            state = win32gui.GetMenuState(self.hmenu, id, win32con.MF_BYCOMMAND)
            if state >= 0 and state & win32con.MFS_CHECKED:
                self.app.sleeping = False
            else:
                self.app.sleeping = True
            return

        # must be an action to execute.
        action = self.menu_items[id]
        assert len(action.commands)==1, "can't handle more than 1"
        cmd = action.commands[0]
        assert cmd.exec_options == cmd.EXEC_TBZRCOMMAND_STD, \
               "can only execute simple commands"

        # execute it - do it the same as our context menu does it
        si = win32process.STARTUPINFO()
        logger.info("Executing taskbar command %r", cmd.cmdline)
        rc = win32process.CreateProcess(None, cmd.cmdline, None, None, 0,
                                        win32process.DETACHED_PROCESS, None, None, si)
        logger.debug("Created process with pid %s", rc[2])

    # Owner-draw related functions.  We only draw the icons
    def OnMeasureItem(self, hwnd, msg, wparam, lparam):
        fmt = "5iP"
        buf = win32gui.PyMakeBuffer(struct.calcsize(fmt), lparam)
        data = struct.unpack(fmt, buf)
        ctlType, ctlID, itemID, itemWidth, itemHeight, itemData = data

        # is there something better than '4'???
        cy = win32api.GetSystemMetrics(win32con.SM_CYMENU) - 4
        cx = cy
        new_data = struct.pack(fmt, ctlType, ctlID, itemID, cx, cy, itemData)
        win32gui.PySetMemory(lparam, new_data)
        return True 

    def OnDrawItem(self, hwnd, msg, wparam, lparam):
        ## lparam is a DRAWITEMSTRUCT
        fmt = "5i2P4iP"
        data = struct.unpack(fmt, win32gui.PyGetString(lparam, struct.calcsize(fmt)))
        ctlType, ctlID, itemID, itemAction, itemState, hwndItem, \
                hDC, left, top, right, bot, itemData = data

        rect = left, top, right, bot
        action = self.menu_items[itemID]

        hicon = load_icon(action.ico_file)
        if hicon is not None:
            try:
                try:
                    # draw the icon.  rect is the entire area we should use.
                    win32gui.DrawIconEx(hDC, left, top, hicon, right-left, bot-top,
                               0, 0, win32con.DI_NORMAL)
                except win32gui.error, exc:
                    logger.error("Failed to draw icon %r: %s", ico, exc)
            finally:
                win32gui.DestroyIcon(hicon)
        else:
            logger.error("Failed to locate an icon for %r", action.ico_file)
        return True

    # Device notifications sent for handles our watcher has open.
    def OnDeviceChange(self, hwnd, msg, wparam, lparam):
        info = win32gui_struct.UnpackDEV_BROADCAST(lparam)
        watcher = self.app.watcher
        if watcher is None:
            logger.debug("Ignoring DEVICECHANGE event - no watcher")
            return
        logger.debug("DeviceChange message (%d: %s)", wparam, info)
        if wparam==win32con.DBT_CUSTOMEVENT:
            if info.devicetype==DBT_DEVTYP_HANDLE:
                logger.debug("DeviceChange custom event %s", info.eventguid)
                if info.eventguid in (GUID_IO_VOLUME_DISMOUNT, GUID_IO_VOLUME_LOCK):
                    watcher.close_watcher_handle(info.hdevnotify)
        elif wparam==win32con.DBT_DEVICEREMOVEPENDING:
            logger.debug("DeviceRemovePending event (%d)", info.devicetype)
            if info.devicetype==win32con.DBT_DEVTYP_HANDLE:
                watcher.close_watcher_handle(info.hdevnotify)
        elif wparam==win32con.DBT_DEVICEQUERYREMOVE:
            logger.debug("DeviceQueryRemove event (%d)", info.devicetype)
            if info.devicetype==win32con.DBT_DEVTYP_HANDLE:
                watcher.close_watcher_handle(info.hdevnotify)
        elif wparam==win32con.DBT_DEVICEREMOVECOMPLETE:
            logger.debug("DeviceRemoveComplete event (%d)", info.devicetype)
            if info.devicetype==win32con.DBT_DEVTYP_HANDLE:
                watcher.close_watcher_handle(info.hdevnotify)


class Taskbar(threading.Thread):
    WindowClass = _TaskbarWindow
    def __init__(self, app):
        self.hwnd = None
        self.app = app
        threading.Thread.__init__(self)
        self.window_created_event = threading.Event()
        app.register_notify_config_change(self.on_config_change)

    def on_config_change(self):
        # called back by the app when its config changed.
        hwnd = self.get_hwnd()
        if self.hwnd is None:
            logger.error("config change callback, but no hwnd available")
            return
        win32gui.PostMessage(hwnd, _TaskbarWindow.WM_USER_CONFIG_CHANGE, 0, 0)

    def get_hwnd(self):
        if self.hwnd is not None:
            return self.hwnd
        self.window_created_event.wait(5)
        if self.hwnd is None:
            logger.error("Failed to wait for the window to be created"
                         " (is the thread running?)")
        return self.hwnd

    def run(self):
        logger.debug("Taskbar thread starting")
        try: # finally to log a message
            try: # finally to close our app
                try: # write exceptions to the log.
                    window = self.WindowClass(self.app)
                    self.hwnd = window.hwnd
                    self.window_created_event.set()
                    self.run_event_loop()
                    # app may have been stopped externally, so our window may
                    # still be alive - kill it.
                    try:
                        win32gui.DestroyWindow(window.hwnd)
                    except win32gui.error:
                        pass
                except:
                    logger.exception("Taskbar thread failed")
            finally:
                # regardless of why our window died, once it has our entire app
                # must come down.
                self.app.stop()
        finally:
            logger.debug("Taskbar thread stopped")

    def run_event_loop(self):
        """Runs a message loop for the window until the app is stopped"""
        while 1:
            rc = win32event.MsgWaitForMultipleObjects([self.app.stop_event],
                                                    False, win32event.INFINITE,
                                                    win32event.QS_ALLINPUT)
            if rc==win32event.WAIT_OBJECT_0:
                # app.stop_event is set.
                break
            # messages are waiting to be processed - do 'em all.
            if win32gui.PumpWaitingMessages():
                # PostQuitMessage() delivered a WM_QUIT message, so exit.
                break
