# A VCSPath object is a TSVN inherited abstraction around a file-system path.
# Its quite possible this class can ultimately die, but it has been ported
# from TSVN to ensure that important semantics are inherited from the TSVN
# code in the short term.

import sys
import os
import stat
import time
import win32api, win32file, win32con # *sob* - should be platform agnostic
import fnmatch

# The following important differences exist between TSVNPath and VCSPath:
# * TSVNPath stores many versions of the same path (forward slash, windows, etc),
#   but we keep just 1, which has always been through abspath()
# * TSVNPath allows URLs etc to be stored as a "path", and has methods to
#   indicate its not a real path etc.  We don't.

class VCSPath(object):
    # possible return values for get_device_type
    DEVICE_TYPE_REMOVABLE = "removable"
    DEVICE_TYPE_FLOPPY = "floppy"
    DEVICE_TYPE_CDROM = "cdrom"
    DEVICE_TYPE_FIXED = "fixed"
    DEVICE_TYPE_REMOTE = "remote"
    # 'unknown' means the device does exist, but its type isn't known.
    DEVICE_TYPE_UNKNOWN = "unknown"

    __slots__ = ["path", "_need_info", "_mtime", "_attr"]

    def __init__(self, path):
        # BZR uses utf8, but the file-system may not (does not on Windows),
        # so to avoid any ambiguity, we insist that the caller decode bytes
        # into a unicode string.
        if not isinstance(path, unicode):
            raise TypeError("paths must be unicode (got %s)" % (type(path,)))
        self.path = os.path.abspath(path)
        self._need_info = True
        self._mtime = None
        self._attr = None

    def long_path(self):
        # Return a "long" version of the object.
        path = self.path
        try:
            path = win32api.GetLongPathNameW(path)
        except NotImplementedError:
            # old windows versions - not likely these days!
            pass
        except win32api.error:
            # most likely is a "device not ready" or similar - but regardless,
            # if we can't make a long version, pretend it already is.
            pass
        if path == self.path:
            return self
        return VCSPath(path)

    # On Windows we wants paths to be case insensitive.
    if sys.platform.startswith('win'):
        # XXX - use an attribute?
        def __hash__(self):
            return hash(self.path.lower())
    
        def __cmp__(self, other):
            return cmp(self.path.lower(), other.path.lower())
    else:
        def __hash__(self):
            return hash(self.path)
    
        def __cmp__(self, other):
            return cmp(self.path, other.path)

    def reset(self):
        self._need_info = True

    def __str__(self):
        # XXX - the intent is that this be used more for diagnostic purposes
        # than anything else - you should never use str() to get a filename
        # to pass to the file-system!
        # XXX Is this dangerous?  What if someone accidently passes the object
        # itself?
        return self.path.encode("utf-8")

    def __unicode__(self):
        return self.path

    def __repr__(self):
        return "VCSPath(%r)" % (self.path,)

    def _update_info(self):
        self._need_info = False # set even if we fail!
        try:
            # *sob* - I can't make stat() give me a float on Windows, even
            # after calling os.stat_float_times(True)
            self._attr, ctime, atime, self._mtime, size = \
                                    win32file.GetFileAttributesExW(self.path)
        except (win32file.error, os.error):
            self._attr = None
            self._mtime = None

    def is_admin_dir(self):
        return self.path.endswith(os.path.sep + ".bzr")

    # XXX - its not yet clear why TSVN differentiates between equal with case
    # and equal without when it is Windows only - presumably so it can compare
    # a path fetched from svn.  However, see our __cmp__ and __hash__ above,
    # as on Windows, we don't want case to ever matter.
    def is_equal_without_case(self, other):
        return self.path.lower()==other.path.lower()

    def is_matched(self, wildcards):
        # Two or more wildcard expression can specified by separated by \n.
        # Sub directories of matched directory are also treated matched.
        if not wildcards:
            return False

        path = self.path + '\\'
        for wildcard in wildcards.split('\n'):
            wildcard = wildcard.strip()
            if wildcard and fnmatch.fnmatch(path, wildcard + u'\\*'):
                return True
        return False

    @property
    def directory(self):
        # Its not yet clear exactly why this is needed!
        if self.is_directory:
            return self
        return self.containing_directory

    @property
    def containing_directory(self):
        dirname = os.path.dirname(self.path)
        if dirname == self.path: # at the root.
            return None
        return VCSPath(dirname)

    def get_root_path_string(self):
        return os.path.splitdrive(self.path)[0]

    def get_device_type(self):
        return get_device_type(self.path)

    def is_ancestor_of(self, maybe_child):
        # Note that the TSVN implementation returns True when self is passed
        # in (ie, a path is considered to be an ancestor of itself.)
        me = self.path.lower()
        look = maybe_child.path
        last = None
        while look != last:
            if me == look.lower():
                return True
            last = look
            look = os.path.dirname(look)
        return False

    def join(self, child):
        return VCSPath(os.path.join(self.path, child))

    @property
    def last_write_time(self):
        if self._need_info:
            self._update_info()
        return self._mtime

    @property
    def is_readonly(self):
        if self._need_info:
            self._update_info()
        return self._attr is not None and self._attr & win32con.FILE_ATTRIBUTE_READONLY

    @property
    def is_directory(self):
        if self._need_info:
            self._update_info()
        return self._attr is not None and self._attr & win32con.FILE_ATTRIBUTE_DIRECTORY != 0

    @property
    def is_file(self):
        if self._need_info:
            self._update_info()
        return self._attr is not None and self._attr & win32con.FILE_ATTRIBUTE_DIRECTORY==0

    @property
    def exists(self):
        if self._need_info:
            self._update_info()
        return self._attr is not None

    @property
    def basename(self):
        return os.path.basename(self.path)

    @property
    def extension(self):
        return os.path.splitext(self.path)[1]

    @classmethod
    def longest_shared_path(self, path1, path2):
        """Returns the longest path these two paths have in common"""
        path1_split = path1.path.split("\\")
        path2_split = path2.path.split("\\")
        split = 0
        while min(len(path1_split), len(path2_split)) > split \
          and path1_split[split] == path2_split[split]:
            split += 1
        if split == 0:
            return None
        return VCSPath(u"\\".join(path1_split[:split]))


# get_device_type support.
if sys.platform.startswith("win"):
    import win32file, win32con

    # Maps win32 status to our status values.
    _device_type_map = {
        win32con.DRIVE_UNKNOWN: VCSPath.DEVICE_TYPE_UNKNOWN,
        # DRIVE_NO_ROOT_DIR means "invalid drive"!!
        win32con.DRIVE_NO_ROOT_DIR: None,
        win32con.DRIVE_REMOVABLE: VCSPath.DEVICE_TYPE_REMOVABLE,
        win32con.DRIVE_FIXED: VCSPath.DEVICE_TYPE_FIXED,
        win32con.DRIVE_REMOTE: VCSPath.DEVICE_TYPE_REMOTE,
        win32con.DRIVE_CDROM: VCSPath.DEVICE_TYPE_CDROM,
        # RAMDISKs aren't used any more (are they?) - let it be UNKNOWN
        # win32con.DRIVE_RAMDISK
    }

    # keyed by the lower case result of splitdrive(), with values being
    # (fetch_result, time_of_fetch)
    _device_type_cache = {}

    DEVICE_CACHE_TIMEOUT = 120 # we re-call win32file.GetDriveType() this often

    def get_device_type(path):
        if path.startswith("\\\\"): # must be a UNC path.
            dev_type = VCSPath.DEVICE_TYPE_REMOTE
        else:
            dev_type = VCSPath.DEVICE_TYPE_UNKNOWN
            # lets keep an eye on speed here, so we don't use splitdriv
            if path[1:2] == ':':
                drv = path[0].lower()
                dev_type, last_time = _device_type_cache.get(drv, (None, 0))
                if time.time() - last_time > DEVICE_CACHE_TIMEOUT:
                    # According to the docs, the trailing slash must exist.
                    win_drv_type = win32file.GetDriveTypeW(drv + ":\\")
                    dev_type = _device_type_map.get(win_drv_type,
                                                    VCSPath.DEVICE_TYPE_UNKNOWN)
                    # Removable drive has drive name 'a:' or 'b:' is floppy, maybe.
                    if dev_type == VCSPath.DEVICE_TYPE_REMOVABLE and drv in 'ab':
                        dev_type = VCSPath.DEVICE_TYPE_FLOPPY
                    _device_type_cache[drv] = (dev_type, time.time())
        return dev_type

else:
    print "Using default implementation of get_device_type"
    def get_device_type(path):
        return VCSPath.DEVICE_TYPE_UNKNOWN
