# A win32 directory watcher, by Dave Anderson, after the TSVN implemenation.
from __future__ import with_statement

import logging
import os
import sys
import threading

import pywintypes
import ntsecuritycon
import win32api
import win32con
import win32event
import win32file
import win32gui
import win32gui_struct

from tbzrlib.vcspath import VCSPath

logger = logging.getLogger('bzr.tbzr.dirwatcher')

mstick = win32api.GetTickCount # a tick in milliseconds.

class DirectoryWatcher(threading.Thread):
    """Watches the file system for changes.
    
    When changes are detected, those changes are reported back to the FolderCrawler
    which then can update the status cache.
    
    When a DirectoryWatcher object is created, a new thread is started which
    waits for file system change notifications.
    To add folders to the list of watched folders, call add_path.
    
    The folders are watched recursively. To prevent having too many folders watched,
    children of already watched folders are automatically removed from watching.
    
    """
    
    def __init__(self, hwnd=None):
        threading.Thread.__init__(self)
        self.lock = threading.RLock()  # locks are for intra-thread data ops
        self.stop_event = win32event.CreateEvent(None, 1, 0, None)
        # An event to say "wake from your loop" due to paths being added or deleted.
        self.wake_event = win32event.CreateEvent(None, 0, 0, None)

        # An event set each time we start and finish a rebuild of the handle
        # map. This is useful if you need to wait for a configuration change to
        # be applied on the worker thread.
        self.handles_rebuilding_event = threading.Event()
        self.handles_rebuilt_event = threading.Event()

        self.watched_paths = []
        # a set of handles we should close next time we update the handles.
        self.handles_to_close = set()

        self.crawler = None
        # TSVN has only a single blocked path - but its quite easy to have a
        # single device removal cause a number of paths to be unmounted, so
        # we need to block them all.
        self.blocked_paths = {}

        # a read-only view of the handles we are waiting on.  changing this
        # will *not* change the behaviour - it is mainly for the test suite to
        # be able to reach in and find the DirWatchInfo objects.
        self._current_watch_map = None

        self.hwnd = hwnd

    def on_change_detected(self, filename, action):
        raise NotImplementedError

    def run(self):
        self._watch_paths()

    def stop(self):
        win32event.SetEvent(self.stop_event)
        self.join(5)
        if self.isAlive():
            logger.warning("dirwatcher thread failed to stop")

    def add_path(self, path, expiry_seconds=120):
        """Adds a new path to be watched. The path points to a directory.
        
        If the path is already watched because a parent of that path is already
        watched recursively, then the new path is just ignored and the method
        returns false.

        Unlike TSVN, we only watch directories for a number of seconds -
        typically the same number of seconds the cache uses to hold files;
        there is no advantage to us watching folders that we don't hold cached
        results for!  (Note that it would be tricky to have the cache notify us
        as it expires items - that still will not tell us if an entire sub-tree
        has been dropped.  We should reconsider this if the cache moves to a
        truly 'branch oriented' system.)
        
        Presumes that path is a directory.
        """

        logger.debug("adding path '%s', expiry=%d", path, expiry_seconds)
        # XXX - StatusCache uses directorywalker...
        # ...directorywalker calls an instance of statuscache??
        # all we need is a StatusCache.config.allowed_device_types
        # maybe require a config object  on initialization instead?
        # maybe have statuscache check before calling add_path?
        #if not statuscache.instance.is_path_allowed(path):
        #    return False
        if self.is_path_blocked(path):
            logger.info("Path %s blocked from being watched", path)
            return False
        with self.lock:
            # Only have one watched folder per directory
            # This is what I'm guessing TSVN's comparisons reduce
            # to when we are only dealing with directories
            # Modify code accordingly should we
            # begin to work with more than directories
            for i in range(len(self.watched_paths)):
                watched_path = self.watched_paths[i][0]
                if watched_path.is_ancestor_of(path):
                    # this path or an ancestor is already watched
                    logger.debug("skipping path '%s': parent '%s' already watched",
                                 path, watched_path)
                    return False
                lsp = VCSPath.longest_shared_path(watched_path, path)
                if lsp is not None:
                    del self.watched_paths[i] # watched_path to be replace by new root
                    new_root = lsp
                    logger.debug("already watching '%s' and this request is for '%s': using shared parent '%s'",
                                 watched_path, path, new_root)
                    break
            else:
                new_root = path

            self.watched_paths.append((new_root, mstick() + expiry_seconds*1000))
            logger.info("added path to watch '%s' - now watching %d locations",
                        new_root, len(self.watched_paths))
            self._trigger_handle_rebuild()
            return True

    def remove_path_and_children(self, path):
        """Removes a path and all its children from the watched list."""
        removed = False
        with self.lock:
            for i in xrange(len(self.watched_paths)-1, -1, -1):
                if path.is_ancestor_of(self.watched_paths[i][0]):
                    removed = True
                    del self.watched_paths[i]
        if removed:
            self._trigger_handle_rebuild()
        return removed
    
    def is_path_watched(self, path):
        """Checks if a path is watched"""
        for watched_path, expiry in self.watched_paths:
            if watched_path.is_ancestor_of(path):
                return True
        return False

    @property
    def watched_paths_count(self):
        """Returns the number of recursively watched paths."""
        return len(self.watched_paths)
    
    def set_crawler(self, crawler):
        """Sets the crawler object which the change notifications are sent to."""
        self.crawler = crawler

    def close_handles_for_path(self, path):
        """Removes Path and Any Children, then resets the watcher"""
        for i, (watched_path, expiry) in enumerate(self.watched_paths):
            if path.is_ancestor_of(watched_path):
                self.block_path(path)
                self.remove_path_and_children(path)
                # should only be 1 - so break.
                break

    def _asynch_read_changes(self, watch_info):
        try:
            win32file.ReadDirectoryChangesW(watch_info.hDir,
                                            watch_info.buffer,
                                            True,
                                            win32con.FILE_NOTIFY_CHANGE_FILE_NAME |
                                            win32con.FILE_NOTIFY_CHANGE_DIR_NAME |
                                            win32con.FILE_NOTIFY_CHANGE_LAST_WRITE,
                                            watch_info.overlapped)
        except win32file.error, exc:
            logger.error("failed to read directory changes for '%s'", watch_info.path)
            watch_info.error_code = exc[0]

    def _watch_paths(self):
        wake_time = 10000 # 10 seconds.  2 minutes is the default cache time,
        # but we can't wait that long each sleep.  XXX - we should calculate
        # the next expiry time.
        logger.debug("in _watch_paths")
        watch_map = {}
        stopping = False
        while not stopping:
            with self.lock:
                self.handles_rebuilding_event.set()
                watch_map = self._rebuild_watch_map(watch_map)
                self.handles_rebuilt_event.set()
            watch_objects = list(watch_map.values())
            watch_handles = [self.wake_event, self.stop_event] + \
                            [o.overlapped.hEvent for o in watch_objects]
            self._current_watch_map = watch_map # for the test suite!!
            while not stopping:
                logger.debug("watcher waiting for %d handles", len(watch_handles))
                rc = win32event.WaitForMultipleObjects(watch_handles, False, wake_time)
                logger.debug("watcher awoke with rc=%d", rc)
                if rc in (win32event.WAIT_OBJECT_0, win32event.WAIT_TIMEOUT):
                    # take the outer loop again, rebuilding the map for expired/new/deleted paths.
                    break
                if rc == win32event.WAIT_OBJECT_0 + 1:
                    # stop event set.
                    logger.debug("watcher stopping")
                    stopping = True
                    break
                # must have been a change event for one of our objects.
                index = rc - win32event.WAIT_OBJECT_0 - 2
                # / NOTE: the longer this code takes to execute until ReadDirectoryChangesW
                # / is called again, the higher the chance that we miss some
                # / changes in the file system!
                watch_info = watch_objects[index]
                # reset the event nice and early
                win32event.ResetEvent(watch_info.overlapped.hEvent)
                logger.debug("queued completion status object %s" % watch_info)
                try:
                    bytes = win32file.GetOverlappedResult(watch_info.hDir, watch_info.overlapped, True)
                except win32file.error, exc:
                    logger.warning("Failed to get directory changes for '%s': %s", watch_info.path, exc)
                    watch_info.error_code = exc[0]
                    # and break out of this loop, forcing this handle to be closed
                    break

                changed_info = win32file.FILE_NOTIFY_INFORMATION(watch_info.buffer, bytes)
                for action, file_name in [result for result in changed_info]:
                    full_path = os.path.join(watch_info.path.path, file_name)
                    self.on_change_detected(full_path, action)
                    if self.crawler:
                        # XXX - move below to base class?
                        logger.debug("ready to request crawl for: %s" % full_path)
                        full_path = full_path.decode(sys.getfilesystemencoding())
                        # XXX - see http://blogs.msdn.com/oldnewthing/archive/2008/09/19/8957958.aspx
                        # We should simply check system+hidden attributes to decide if something
                        # should be ignored - that will pick up the recycle bin and
                        # other esoteric stuff.
                        if full_path.contains("\\tmp\\") or file_name.endswith("\\tmp"):
                            continue
                        if full_path.contains(":\\RECYCLER\\"):
                            continue
                        if full_path.contains(":\\$Recycle.Bin\\"):
                            continue
                        if file_name.endswith(".tmp"):
                            continue
                        logger.debug("change notification: %r", full_path)
                        self.crawler.AddPathForUpdate(VCSPath(full_path))
                # and prepare to read the next set of changes for this handle.
                self._asynch_read_changes(watch_info)
                if watch_info.error_code:
                    # there was an error - re-build the watch list.
                    break
        # clear all our handles.
        self._current_watch_map = None
        logger.debug("dirwatcher stopping - closing watch handles.")
        # reset watched paths to empty and ask to rebuild the map - it should
        # end up empty!
        self.watched_paths = []
        temp = self._rebuild_watch_map(watch_map)
        assert not temp, temp # expect empty watch map from empty list!
        logger.info("dirwatcher stopped")

    def _rebuild_watch_map(self, map):
        """Rebuild a watch map.


        Uses self.watched_paths and self.handles_to_close to build a new 'watch
        map' of paths to DirWatchInfo objects. Any objects in the old map that
        are not returned in the new map are closed.
        
        self.watched_paths may be purged of expired watch_paths.
        self.handles_to_close is always emptied.
        """
        logger.debug("Rebuilding watch map - starting with %d entries", len(map))
        new_map = {}
        # we discard watched_paths we can't watch - this holds the new entries.
        new_watched_paths = []
        now = mstick()
        for (path, expiry) in self.watched_paths:
            if now > expiry:
                logger.info("Discarding watch for %s - it has expired", path)
                continue
            try:
                watch_info = map[path]
                if watch_info.hDevNotify and long(watch_info.hDevNotify) in self.handles_to_close:
                    logger.info("Discarding watch for '%s' - handle close requested",
                                 path)
                    self.block_path(path)
                    continue
            except KeyError:
                # This is a new path - open it.
                try:
                    hDir = win32file.CreateFile(path.path,
                                                ntsecuritycon.FILE_LIST_DIRECTORY,
                                                win32file.FILE_SHARE_READ |
                                                win32file.FILE_SHARE_WRITE |
                                                win32file.FILE_SHARE_DELETE,
                                                None,
                                                win32file.OPEN_EXISTING,
                                                win32file.FILE_FLAG_BACKUP_SEMANTICS |
                                                win32file.FILE_FLAG_OVERLAPPED,
                                                None)
                except win32file.error, exc:
                    logger.warning("Failed to watch path '%s': %s", path, exc)
                    continue
                watch_info = DirWatchInfo(hDir, path, self.hwnd)
                self._asynch_read_changes(watch_info)
                if watch_info.error_code:
                    # eeek - couldn't watch for changes and it is a new
                    # object that will not be auto-closed below.  Kill it.
                    logger.debug("Failed to watch for initial changes of new item - ignoring")
                    watch_info.close_dir_handle()
                    watch_info = None
                    continue
                logger.debug("preparing to watch new path %s", path)
            # worked - either have an old one or a new one.
            if watch_info.error_code:
                logger.info("Discarding watch for '%s' - error code %d",
                            path, watch_info.error_code)
                continue

            new_watched_paths.append((path, expiry))
            new_map[path] = watch_info
        # done - update the new data.
        # Any in the old map but no in the new one must die.
        for path, info in map.iteritems():
            if path not in new_map:
                info.close_dir_handle()
                continue
        self.handles_to_close.clear()
        self.watched_paths = new_watched_paths
        logger.debug("Rebuilt watch map has %d entries", len(new_map))
        return new_map

    def _trigger_handle_rebuild(self, wait=False):
        if wait:
            self.handles_rebuilding_event.clear()
            self.handles_rebuilt_event.clear()
        win32event.SetEvent(self.wake_event)
        if wait:
            # we should get notified almost immediately!
            logger.debug('waiting for rebuild start event')
            self.handles_rebuilding_event.wait(10)
            assert self.handles_rebuilding_event.isSet()

            logger.debug('waiting for rebuild done event')
            self.handles_rebuilt_event.wait(10)
            assert self.handles_rebuilt_event.isSet()
            logger.debug('rebuilt events all set')

    def close_watcher_handle(self, hdir):
        # called when we need to close a directory handle we are watching.
        # likely to be called in response to a windows notification about the
        # device being removed etc.
        with self.lock:
            logger.debug("queuing removal of handle %s", hdir)
            self.handles_to_close.add(long(hdir))
        # no need to have the lock while signalling the event.
        self._trigger_handle_rebuild(True)
        logger.debug("completed removal of handle %s", hdir)

    def is_path_blocked(self, path):
        key = path
        expiry = self.blocked_paths.get(key)
        if expiry is None:
            # hrmph - search for an ancestor.
            # Must take lock while iterating
            with self.lock:
                for path_look, expiry_look in self.blocked_paths.iteritems():
                    if path_look.is_ancestor_of(path):
                        key = path_look
                        expiry = expiry_look
                        break
                else:
                    # not found iterating either!
                    return False

        if mstick() >= expiry:
            # the found entry has expired and is no longer being blocked.
            try:
                del self.blocked_paths[key]
            except KeyError:
                # another thread may have done this - doesn't matter.
                pass
            return False
        return True

    def block_path(self, path, ms=4000):
        # block the path from being watched for 4 seconds
        self.blocked_paths[path] = mstick() + ms
        logger.debug("blocking path %s", path)


class DirWatchInfo(object):

    def __init__(self, hDir, path, hwnd = None):
        assert(hDir)
        assert(path)
        self.hDir = hDir
        self.path = path
        self.buffer = win32file.AllocateReadBuffer(8192)
        self.overlapped = pywintypes.OVERLAPPED()
        # As per docs for GetOverlappedResult, we muse use a manual reset event.
        self.overlapped.hEvent = win32event.CreateEvent(None, True, 0, None)
        self.overlapped.object = path # index for the map
        self.hDevNotify = None
        self.error_code = None # set by the watcher on error.
        self.closed = False
        # *sob* - pywin32-212 doesn't release the GIL when calling
        # RegisterDeviceNotification() - we check that by seeing if
        # win32api.GetNativeSystemInfo exists - that too was introduced
        # in 213.
        pywin32_ok = hasattr(win32api, "GetNativeSystemInfo")
        if hwnd and not pywin32_ok:
            logger.info("skipping RegisterDeviceNotification due to pywin32 bug")
        if hwnd and pywin32_ok:
            filter = win32gui_struct.PackDEV_BROADCAST_HANDLE(hDir)
            try:
                self.hDevNotify = win32gui.RegisterDeviceNotification\
                                  (hwnd, filter, win32con.DEVICE_NOTIFY_WINDOW_HANDLE)
            except win32gui.error, exc:
                logger.warning("Failed to watch '%s' for changes: %s", path, exc)

    def __del__(self):
        # relying on GC is evil for things like handles etc.
        if not self.closed:
            logger.warning("GC collected a DirWatchInfo object for '%s' - handles are temporarily leaking!",
                           self.path)
            self.close_dir_handle()

    def close_dir_handle(self):
        assert not self.closed
        logger.debug("Closing directory handle to '%s'", self.path)
        self.closed = True
        if self.hDir:
            self.hDir.Close()
            self.hDir = None
        if self.hDevNotify:
            self.hDevNotify.Close()
            self.hDevNotify = None
