"""status cache system based on workingtree."""
import os
from time import time

from tbzrlib.vcspath import VCSPath
from tbzrlib.app import get_app
import tbzrlib.bzrstatus as bzrstatus

from win32com.shell import shell, shellcon

import logging
logger = logging.getLogger('bzr.tbzr.cache')
mutter = logger.debug # for now...

import bzrlib.workingtree as workingtree
import bzrlib.errors as errors
from tbzrlib.crawler import Crawler
from tbzrlib.pathset import PathSet

__all__ = ["WtCache"]

CACHE_TIMEOUT = 10 * 60 # seconds

class PathBlocker(object):
    def __init__(self):
        self.init()

    def init(self):
        self._blocked_paths = PathSet()
        self._blocked_expire = {}

    def block_path(self, path, expire_sec=5):
        directory = path.directory
        self._blocked_paths.add(directory)
        self._blocked_expire[directory] = time() + expire_sec

    def is_blocked(self, maybe_blocked):
        blocked_container = self._blocked_paths.get_container(maybe_blocked)
        for block_path in blocked_container:
            if self._blocked_expire[block_path] > time():
                return True
            del self._blocked_paths[block_path]
            del self._blocked_expire[block_path]


class WtCacheEntry(object):

    def __init__(self, tree):
        self.path = VCSPath(os.path.abspath(tree.abspath(u'')))
        self.members = {}
        t = time()
        self.updated_at = t
        self.discard_at = t + CACHE_TIMEOUT

        self.in_progress = True
        self.tree = tree

    def check_state(self):
        config = get_app().config
        show_as_modified = config.get("icon_overlays.show_as_mod_if_unversioned")
        show_ignored = config.get("icon_overlays.show_ignored")

        tree = self.tree
        del self.tree
        members = self.members

        tree.lock_read()
        try:
            for ch in tree.iter_changes(tree.basis_tree(),
                                        include_unchanged=True,
                                        want_unversioned=show_as_modified or show_ignored):
                status = bzrstatus.from_bzr_change_tuple(tree, ch)

                if status.is_deleted():
                    name = ch[1][0]
                else:
                    name = ch[1][1]

                path = VCSPath(os.path.join(self.path.path, name))
                if status.is_versioned() and not status.is_deleted():
                    lwt = path.last_write_time
                else:
                    lwt = None
                members[path] = (status, lwt)
        finally:
            tree.unlock()
        self._update_dirstate(show_as_modified)
        self.in_progress = False

        t = time()
        self.updated_at = t
        self.discard_at = t + CACHE_TIMEOUT

    def _update_dirstate(self, show_as_modified):
        app = get_app()
        members = self.members

        for path, (status, _) in self.members.items():
            if status.is_conflicted():
                new_status = app.vcs_status_conflict_factory()
            elif status.is_changed() or (show_as_modified and status.is_unversioned()):
                new_status = app.vcs_status_modified_factory()
            else:
                continue
            while True:
                parent_directory = path.containing_directory
                parent_status = members.get(parent_directory)
                if parent_status is None:
                    break
                if new_status.is_modified():
                    if parent_status[0].is_changed():
                        break
                else:
                    if parent_status[0].is_conflicted():
                        break
                
                members[parent_directory] = (
                        new_status, parent_status[1]
                        )
                path = parent_directory

    def get_status_for_member(self, path, recursive):
        if self.in_progress:
            return get_app().vcs_status_progress_factory(), None
        try:
            return self.members[path]
        except KeyError:
            return get_app().vcs_status_unversioned_factory(), None

class WtCache(object):
    def __init__(self, app):
        self.app = app

        self.dircache = {} # {workingtree root path: WtCacheEntry}
        self.blacklist = {} # {directory path: expired_time}
        self.blocked = PathBlocker()
        self.initialize()

    def initialize(self):
        self.most_recent_expires_at = time() - 1
        self.most_recent_status = 0
        self.dircache.clear()
        # this is a hard-coded list of paths to not watch
        self.nowatch_paths = set()

        # its not clear why TSVN chose these particular paths.
        csidls = [shellcon.CSIDL_COOKIES,
                  shellcon.CSIDL_HISTORY,
                  shellcon.CSIDL_INTERNET_CACHE,
                  shellcon.CSIDL_SYSTEM,
                  shellcon.CSIDL_WINDOWS,
                 ]
        for csidl in csidls:
            path = shell.SHGetFolderPath(0, csidl, None, 0)
            self.nowatch_paths.add(VCSPath(path))
        # it also appears tsvn allows the user to specify their own?

        self.crawler = Crawler(self)

    def stop(self):
        self.crawler.stop()
        self.crawler = None

    def restart(self):
        self.crawler = Crawler(self)

    def refresh(self):
        ## self.shell_cache.force_refresh() - force refresh of user prefs.
        self.clear_cache()

    def _is_path_good(self, path):
        for p in self.nowatch_paths:
            if p.is_ancestor_of(path):
                return False
        return True

    def clear_cache(self):
        self.dircache.clear()
        self.blacklist.clear()
        # TODO: clear crawler queue.

    def remove_cache_for_path(self, path):
        mutter(u"removing cache for path %s", path)
        if path in self.blacklist:
            del self.blacklist[path]
        else:
            wt_entry = self._get_cache_for_path(path)
            if wt_entry is not None:
                del self.dircache[wt_entry.path]

    def update_cache_for_path(self, path):
        mutter(u"Updating cache for path %s", path)
        if path in self.blacklist:
            del self.blacklist[path]
        else:
            wt_entry = self._get_cache_for_path(path)
            if wt_entry is not None:
                if wt_entry.in_progress:
                    return
                if wt_entry.updated_at + 1 < time():
                    return
                del self.dircache[wt_entry.path]
                # don't create a cache.
                # Cheking status will be caused from icon overlay.

    def _create_cache_for_path(self, path):
        if not path.exists:
            raise ValueError(u"_create_cache_for_path for non-existing directory: %s" % (path,))
        if not self.app.is_path_allowed(path):
            raise ValueError(u"_create_cache_for_path for not-allowed directory: %s" % (path,))
        mutter(u"adding %s to our cache", path)

        try:
            tree, rel_path = workingtree.WorkingTree.open_containing(path.path)
            wt_entry = WtCacheEntry(tree)
            self.dircache[wt_entry.path] = wt_entry
            self.crawler.add_entry(wt_entry)
            return wt_entry
        except (errors.NotBranchError, errors.NoWorkingTree):
            expired_at = time() + CACHE_TIMEOUT
            while path:
                self.blacklist[path] = expired_at
                path = path.containing_directory

    def _get_cache_for_path(self, path, create=False):
        if path.is_file:
            raise ValueError(u"%r is a file, not a directory" % (path,))
        dircache = self.dircache
        for wt_root, wt_entry in dircache.items():
            if wt_root.is_ancestor_of(path):
                if time() < wt_entry.discard_at:
                    return wt_entry
                else:
                    mutter("Cache for %s is outdated.", path)
                    del dircache[wt_root]
                    break
        if create:
            return self._create_cache_for_path(path)
        else:
            return None

    def get_status_for_path(self, path, recursive=True):
        """return status for path."""
        now = time()
        app = self.app
        if now < self.most_recent_expires_at:
            if path.is_equal_without_case(self.most_recent_path):
                return self.most_recent_status

        self.most_recent_path = path
        self.most_recent_expires_at = now+1
        self.most_recent_status = app.vcs_status_nothing_factory()

        if not self._is_path_good(path):
            mutter(u"path '%s' is no good!", path)
            return self.most_recent_status
        if not self.app.is_path_allowed(path):
            mutter(u"path '%s' is excluded due to user-preferences", path)
            return self.most_recent_status
        if self.blocked.is_blocked(path):
            mutter(u"path '%s' is blocked.", path)
            return self.most_recent_status

        # Now we know the path is allowed, convert it to a canonical 'long' version.
        path = path.long_path()
        try:
            status_fix = False
            watch_path = directory = path.directory

            # check blacklist cache.
            if directory in self.blacklist:
                if self.blacklist[directory] > time():
                    status_fix = True
                    self.most_recent_status = app.vcs_status_nothing_factory()
                else:
                    del self.blacklist[directory]

            # check workingtree cache.
            if not status_fix:
                wt_entry = self._get_cache_for_path(directory, False)
                if wt_entry is not None:
                    status, lwt = wt_entry.get_status_for_member(path, recursive)
                    if lwt is None or lwt == path.last_write_time:
                        status_fix = True
                        self.most_recent_status = status
                        watch_path = wt_entry.path

            # try to create workingtree cache
            if not status_fix:
                wt_entry = self._create_cache_for_path(directory)
                if wt_entry is not None:
                    status, lwt = wt_entry.get_status_for_member(path, recursive)
                    status_fix = True
                    self.most_recent_status = status
                    watch_path = wt_entry.path

            # path is not in workingtree when fail to create cache.
            if not status_fix:
                status_fix = True
                self.most_recent_status = app.vcs_status_nothing_factory()

            # If the result is versioned, we watch the directory
            # (watching unversioned directories would mean we
            # quickly end up watching the root of each drive!)
            if self.most_recent_status.is_versioned() and app.watcher is not None:
                app.watcher.add_path(watch_path, CACHE_TIMEOUT)
        except ValueError:
            # needs more thought - these are 2 cases known to
            # throw errors just above
            if path.is_admin_dir():
                mutter(u"Ignoring request a .bzr directory (%s)", path)
            elif not path.is_directory:
                mutter(u"Ignoring request for a path that is a file (%s)", path)
            else:
                raise
        return self.most_recent_status

    def block_path(self, path):
        self.blocked.block_path(path)
